unit Logger;

{$IFDEF FPC}
{$M+}
{$ENDIF}

interface

uses
  Classes,
  SysUtils;
  
type
  TLogger = class(TFileStream)
  private
    FApplicationName : string;
    FApplicationPath : string;
  protected

  public
    constructor Create;
    procedure LogError( ErrorMessage : string; Location : string ); virtual;
    procedure LogWarning( WarningMessage : string; Location : string ); virtual;
    procedure LogStatus( StatusMessage : string; Location : string ); virtual;
  published
    property ApplicationName : string read FApplicationName;
    property ApplicationPath : string read FApplicationPath;
  end;
  TDummyLogger = class(TLogger)
   public
    constructor Create;
    procedure LogError( ErrorMessage : string; Location : string ); override;
    procedure LogWarning( WarningMessage : string; Location : string ); override;
    procedure LogStatus( StatusMessage : string; Location : string ); override;

  end;

implementation

constructor TDummyLogger.Create;
begin
end;

constructor TLogger.Create;
begin
  FApplicationName := ExtractFileName( ParamStr(0) );
  FApplicationPath := ExtractFilePath( ParamStr(0) );
  inherited Create( FApplicationPath + Copy( FApplicationName, 0, Length( FApplicationName ){$IFDEF WIN32} - 4{$ENDIF} ) + '.log',
                    fmCreate {$IFNDEF FPC}or fmShareExclusive{$ENDIF} );
end;

procedure TDummyLogger.LogError(ErrorMessage, Location: string);
begin
end;

procedure TDummyLogger.LogStatus(StatusMessage, Location: string);
begin
end;

procedure TDummyLogger.LogWarning(WarningMessage, Location: string);
begin
end;

procedure TLogger.LogError(ErrorMessage, Location: string);
var
  S : string;
begin
  S := '*** ERROR *** : @ ' + TimeToStr(Time) + ' MSG : ' + ErrorMessage + ' IN : ' + Location + #13#10;
  Write(S[1], Length(S));
end;

procedure TLogger.LogStatus(StatusMessage, Location: string);
var
  S : string;
begin
  S := 'STATUS INFO : @ ' + TimeToStr(Time) + ' MSG : ' + StatusMessage + ' IN : ' + Location + #13#10;
  Write(S[1], Length(S));
end;

procedure TLogger.LogWarning(WarningMessage, Location: string);
var
  S : string;
begin
  S := '=== WARNING === : @ ' + TimeToStr(Time) + ' MSG : ' + WarningMessage + ' IN : ' + Location + #13#10;
  Write(S[1], Length(S));
end;

end.
