unit ac3d;

interface
{$DEFINE OZONE}
uses
  Classes, dglOpenGL;

const
   MaxUV = 10;
   MaxVertRef = 10;
type
  //----------------------------------------------------------------------------
  // Die verschiedenen Strukturen. Zuersteinmal ein einfacher Pukt.
  //----------------------------------------------------------------------------
  PACPoint = ^TACPoint;
  TACPoint = record
    x, y, z: Single;
  end;

  //----------------------------------------------------------------------------
  // Eine Normale (wird fr Objekte und Surfaces verwendet).
  //----------------------------------------------------------------------------
  PACNormal = ^TACNormal;
  TACNormal = record
    x, y, z: Single;
  end;

  //----------------------------------------------------------------------------
  // Ein Vertex beseht aus einem Punkt und einer Normalen.
  //----------------------------------------------------------------------------
  PACVertex = ^TACVertex;
  TACVertex = record
    Vertex : TACPoint;
    Normal : TACNormal;
  end;

  //----------------------------------------------------------------------------
  // Die UV-Koordinaten fr Texture-Mapping.
  //----------------------------------------------------------------------------
  PACUV = ^TACUV;
  TACUV = record
    u, v : Single;
  end;

  //----------------------------------------------------------------------------
  // Die Daten eines Surfaces (einer Oberflche).
  //----------------------------------------------------------------------------
  PACSurface = ^TACSurface;
  TACSurface = record
    Vertref     : array[0..MaxVertRef-1] of Integer;
    UVs         : array[0..MaxUV-1] of TACUV;
    NumVertRef  : Integer;
    Normal      : TACNormal;
    Flags       : Integer;
    Material    : Integer;
  end;

  //----------------------------------------------------------------------------
  // Die Daten eines Objektes.
  //----------------------------------------------------------------------------
  PACObject = ^TACObject;
  TACVertArray = array[0..1350] of TACVertex;
  TACSurfArray = array[0..3000] of TACSurface;
  TACObjCache = record
   Verts:TACVertArray;
   Surfs:TACSurfArray;
  end;
  TACObject = record
    Scale            : TVector3d;
    Rotate           : TVector3d;
    Trans            : TVector3d;
    Matr             : array [0..15] of Single;
    Loc              : TACPoint;
    Name             : string;
    Data             : string;
    URL              : string;
    Vertices         : TACVertArray;
    NumVert          : Integer;

    Surfaces         : TACSurfArray;
    NumSurf          : Integer;
    TextureRepeat_X,
    TextureRepeat_Y  : Single;
    TextureOffset_X,
    TextureOffset_Y  : Single;

    NumKids          : Integer;
    Kids             : array of PACObject;
    Matrix           : array[0..15] of single;
    ObjType          : Integer;
    Texture          : Cardinal;
    TextureFile      : string;
  end;

  //----------------------------------------------------------------------------
  // Die Farbe fr OpenGL.
  //----------------------------------------------------------------------------
  PACCol = ^TACCol;
  TACCol = record
    r, g, b, a : Single;
  end;

  //----------------------------------------------------------------------------
  // Daten fr Materialien.
  //----------------------------------------------------------------------------
  PACMaterial = ^TACMaterial;
  TACMaterial = record
    RGB          : TACCol;
    Ambient      : TACCol;
    Specular     : TACCol;
    Emissive     : TACCol;
    Shininess    : Single;
    Transparency : Single;
    Name         : String;
  end;

  //----------------------------------------------------------------------------
  // Die Events des Loaders.
  //----------------------------------------------------------------------------
//  TACLoadTexture = procedure(var GLTexID: GLuint; ATexFile: string);
//  TACFreeTexture = procedure(var GLTexID:GLuint);

  //----------------------------------------------------------------------------
  // Konstanten fr die Render-States.
  //----------------------------------------------------------------------------
  TACRenderStateEnum = (AC_USE_TEXTURE, AC_USE_LIGHT, AC_SOLIDDRAW,
                        AC_WIREFRAME, AC_POINTS, AC_SAVESTATES);

  //----------------------------------------------------------------------------
  // Die Loader-Klasse selber.
  //----------------------------------------------------------------------------
  TAC3DFile = class(TObject)
  public
    FObject : PACObject; // Zeiger auf die Geladenen Daten.
    FPalette : array of TACMaterial; // Die Material-Palette der Datei.
    FStream : TStream;

    FLastTexID  : Cardinal;
    FLastMatID  : Integer;
    FPalCount   : Integer;
    FDispListID : Cardinal;

//    FLoadTex : TACLoadTexture;
//    FFreeTex : TACFreeTexture;

    FUseTexture : Boolean;
    FUseLight   : Boolean;
    FSaveStates : Boolean;
    FPolygonMode : GLenum;
    FNoData : boolean;
    owner : pointer;
    procedure FreeObject(AObject : PACObject = nil);
    procedure InitSurface(var Surface: TACSurface);
    procedure ObjectCalcVertexNormals(Obj: PACObject);
    procedure SetMaterial(MatID: Integer);
    procedure DoRender(AObject: PACObject);

    function LoadAC3D: Boolean;
    function NewObject: PACObject;
    function ReadLine: string;
    function EndOfStream: Boolean;
    function TriCalcNormal(v1, v2, v3: TACPoint): TACNormal;
    function ReadSurface(var Surface: TACSurface; Obj: PACObject): Boolean;
    function StringToObjectType(s: string): Integer;
    function LoadObject(Parent: PACObject): PACObject;
    function GetPalEntry(ID: Integer): TACMaterial;
    //--------------------------------------------------------------------------
    // Konstructor und Destructor zum Initiliaiseren bzw. Freigeben der Daten
    // im Objekt bei dessen Erstellung bzw. Freigabe.
    //--------------------------------------------------------------------------
    constructor Create(own:pointer);
    destructor Destroy; override;

    //--------------------------------------------------------------------------
    // Die freigegebenen Prozeduren des Loaders
    //--------------------------------------------------------------------------
    procedure Render(x:integer = -1);
    procedure CreateDisplayList;
    procedure RenderDisplayList;
    procedure ReleaseDisplayList;
    procedure Release;
    procedure acEnable(FlagID: TACRenderStateEnum);
    procedure acDisable(FlagID: TACRenderStateEnum);

    //--------------------------------------------------------------------------
    // Die Funktionen zum Laden einer Datei. Dabei werden von die 3 Varianten
    // Resource, Datei und benutzerdefinierter Stream untersttzt.
    //--------------------------------------------------------------------------
    function LoadFromStream(AStream: TStream): Boolean;
    function LoadFromFile(FileName: string): Boolean;
    function LoadFromResource(Instance: THandle; ResName: string): Boolean;

    //--------------------------------------------------------------------------
    // Die Events.
    //--------------------------------------------------------------------------
//    property OnLoadTexture : TACLoadTexture read FLoadTex write FLoadTex;
//    property OnFreeTexture : TACFreeTexture read FFreeTex write FFreeTex;

    //--------------------------------------------------------------------------
    // Die Eigenschaften des Objektes.
    //--------------------------------------------------------------------------
    property ObjectData : PACObject read FObject write FObject;
    property PalCount : integer read FPalCount;
    property Palette[ID: Integer]: TACMaterial read GetPalEntry;
    property DisplayListID : GLuint read FDispListID;

    //--------------------------------------------------------------------------
    // Die Render-States knnen auch per Properties gesetzt werden, je nach Ge-
    // schmack.
    //--------------------------------------------------------------------------
    property UseTexture : Boolean read FUseTexture write FUseTexture;
    property UseLight : Boolean read FUseLight write FUseLight;
    property PolygonMode : GLenum read FPolygonMode write FPolygonMode;
    property SaveStates : Boolean read FSaveStates write FSaveStates;
    {$IFDEF OZONE}
    procedure CacheToFile(filename:string);
    procedure RestoreCache(filename:string);
    {$ENDIF}
  end;
const
  //----------------------------------------------------------------------------
  // Die Verschiedenen Objekt-Typen.
  //----------------------------------------------------------------------------
  AC_OBJECT_WORLD  = 999;
  AC_OBJECT_NORMAL = 0;
  AC_OBJECT_GROUP  = 1;
  AC_OBJECT_LIGHT  = 2;

  //----------------------------------------------------------------------------
  // Die verschiedenen Surface-Typen.
  //----------------------------------------------------------------------------
  AC_SURFACE_TYPE_POLYGON    = 0;
  AC_SURFACE_TYPE_CLOSEDLINE = 1;
  AC_SURFACE_TYPE_LINE       = 2;

  //----------------------------------------------------------------------------
  // Zusatz-Parameter fr ein Surface.
  //----------------------------------------------------------------------------
  AC_SURFACE_SHADED   = 1 shl 4;
  AC_SURFACE_TWOSIDED = 1 shl 5;

  //----------------------------------------------------------------------------
  // Sonstige Konstanten.
  //----------------------------------------------------------------------------
  AC_NO_TEXTURE = Cardinal(-1);

implementation

uses
  SysUtils{$IFDEF OZONE},o3game,o3utils{$ENDIF};

{$IFDEF OZONE}
procedure TAC3dFile.RestoreCache(filename:string);
var f:file of TACObjCache;
    t:TACObjCache;
 procedure readObj(o:PACObject);
 var x:integer;
 begin
  if o<>FObject then begin
   read(f,t);
   o.Vertices:=t.Verts;
   o.Surfaces:=t.Surfs;
  end;
  if o.NumKids>0 then
   for x:=0 to o.NumKids-1 do
    readObj(o.kids[x]);
 end;
begin
 Assign(f,filename);
 Reset(f);
 readObj(FObject);
 Close(f);
end;

procedure TAC3dFile.CacheToFile(filename:string);
var f:file of TACObjCache;
    t:TACObjCache;
 procedure writeObj(o:PACObject);
 var x:integer;
 begin
  t.Verts:=o.Vertices;
  t.Surfs:=o.Surfaces;
  if o<>FObject then write(f,t);
  if o.NumKids>0 then
   for x:=0 to o.NumKids-1 do
    writeObj(o.kids[x]);
 end;
begin
 Assign(f,filename);
 ReWrite(f);
 writeObj(FObject);
 Close(f);
end;

{$ENDIF}

////////////////////////////////////////////////////////////////////////////////
// GetNextToken
//------------------------------------------------------------------------------
// Diese Funktion durchsucht den in Str bergebenen string kopiert und lscht
// alles bis zum nchsten vorkommen von Seperator herraus.
//------------------------------------------------------------------------------
// Parameter: Seperator -> Das zeichen, durch welches die Token getrennt werden.
//            Str       -> Der zu durchsuchende String.
// Rckgabe : Result    -> Alles vom Anfang von Str bis zur nchsten Position
//                         von Seperator.
//            Str       -> Der Inhalt von Result wird aus Str gelscht.
function GetNextToken(Seperator: Char; var Str: string): string;
var
  P : integer;
begin
  //----------------------------------------------------------------------------
  // Result initialisieren und smtliche Leerzeichen am Anfang und Ende von Str
  // entfernen. Sollte Str dannach leer sein, dir Funktion verlassen. Anderen
  // Falls die erste Position von Seperator ermitteln.
  //----------------------------------------------------------------------------
  Result := '';
  Str    := Trim(Str);
  if Str = '' then Exit;
  P := Pos(Seperator, Str);

  //----------------------------------------------------------------------------
  // berprfen, ob der Seperator im String Str gefunden wurde. Wenn nicht,
  // (P ist dann 0), den Rest von Str zurckliefern und Str selber auf einen
  // Leerstring setzen. Anderen Falls alles bis zur ermitelten Position herraus-
  // kopieren und -lschen.
  //----------------------------------------------------------------------------
  if p = 0 then
  begin
    Result := Trim(Str);
    Str    := '';
  end else
  begin
    Result := Trim(Copy(Str, 1, p));
    Delete(Str, 1, p);
    Str := Trim(Str);
  end;
end;

{ TAC3DFile }

////////////////////////////////////////////////////////////////////////////////
// Create
//------------------------------------------------------------------------------
// Initialisiert das Objekt bei seiner Erstellung.
//------------------------------------------------------------------------------
// Parameter: --- Standard ---
// Rckgabe : --- Standard ---
constructor TAC3DFile.Create;
begin
  //----------------------------------------------------------------------------
  // Durch den Aufruf von ReleaseData werden die Default-Werte der Variablen ge-
  // setz.
  //----------------------------------------------------------------------------
  owner:=nil;
  {$IFDEF OZONE}owner:=own;{$ENDIF}
  Release;

  //----------------------------------------------------------------------------
  // Die Render-Flags initialisieren.
  //----------------------------------------------------------------------------
  FUseTexture := True;
  FUseLight   := True;
  FSaveStates := True;
  FPolygonMode := GL_FILL;
//  FLoadTex := L;
//  FFreeTex := F;
  FNoData := false;
end;


////////////////////////////////////////////////////////////////////////////////
// Destroy
//------------------------------------------------------------------------------
// Sorgt dafr, dass die vom Objekt belegten Resourcen beim Freigeben des
// Objektes ebenfalls freigegeben werden.
//------------------------------------------------------------------------------
// Parameter: --- Standard ---
// Rckgabe : --- Standard ---
destructor TAC3DFile.Destroy;
begin
  //----------------------------------------------------------------------------
  // Alle geladenen Daten freigeben und den Vorgnger der Funktion aufrufen.
  //----------------------------------------------------------------------------
  FreeObject;
  inherited;
end;


////////////////////////////////////////////////////////////////////////////////
// FreeObject
//------------------------------------------------------------------------------
// Gibt alle vom Objekt belegten Reourcen wieder frei.
//------------------------------------------------------------------------------
// Parameter: AObject -> Das freizugebende Objekt. Wenn NIL bergeben wird,
//                       wird FObject zum friegeben verwendet.
// Rckgabe : --- Keine ---
procedure TAC3DFile.FreeObject(AObject : PACObject = nil);
var
  i : Integer;
  Obj : PACObject;
begin
  //----------------------------------------------------------------------------
  // Sollte NIL bergeben worden sein, soll FObject freigegeben werden, anderen
  // Falls eben das bergebene Object.
  //----------------------------------------------------------------------------
  if Assigned(AObject) then
    Obj := AObject
  else
  begin

    //--------------------------------------------------------------------------
    // Es soll das Haupt-Objekt (FObject) freigegeben werden. Dazu muss aber
    // vorher berprft werden, ob FObject berhaupt definiert ist. Falls
    // nicht, die Funktion verlassen, da es nichts mehr zu tun gibt.
    //--------------------------------------------------------------------------
    if not Assigned(FObject) then Exit;
    Obj := FObject;
   end;

  //----------------------------------------------------------------------------
  // Sollte dem Objekt eine Textur zugewiesen worden sein, deise jetzt noch
  // freigeben.
  //----------------------------------------------------------------------------
  {$IFDEF OZONE}
  if Obj^.Texture <> AC_NO_TEXTURE then
    if Assigned(owner) then
      TOzoneMainClass(owner).gfxSys.FreeTexture(Obj^.Texture);
  {$ENDIF}
  //----------------------------------------------------------------------------
  // Nun alle Kinder des Objektes freigeben.
  //----------------------------------------------------------------------------
  for i := 0 to Obj^.NumKids - 1 do
    FreeObject(Obj^.Kids[i]);

  //----------------------------------------------------------------------------
  // Nun die Dynamischen Arrays des Objektes freigeben.
  //----------------------------------------------------------------------------
//  Finalize(Obj^.Surfaces);
//  Finalize(Obj^.Vertices);
  Finalize(Obj^.Kids);

  //----------------------------------------------------------------------------
  // Zu guter letzt das Object selber auch freigeben. Zum Schlu noch ber-
  // prfen, das bergebene Objekt NIL war. Wenn ja, wurde FObject freigegeben
  // und muss noch auf NIL gesetzt werden.
  //----------------------------------------------------------------------------
  Dispose(Obj);
  if AObject = nil then FObject := nil;
end;


////////////////////////////////////////////////////////////////////////////////
// Release                                                         
//------------------------------------------------------------------------------
// Gibt die bisher geladenene Datei weider frei.
//------------------------------------------------------------------------------
// Parameter: --- Keine ---
// Rckgabe : --- Keine ---
procedure TAC3DFile.Release;
begin
  //----------------------------------------------------------------------------
  // Die DisplayListe (falls eine Existiert), das Objekt und das Array mit der
  // Palette freigeben um den reservierten Speicher aufzurumen.
  //----------------------------------------------------------------------------
  // ReleaseDisplayList;
  FreeObject;
  Finalize(FPalette);

  //----------------------------------------------------------------------------
  // Nun alle Variablen auf die Default-Werte zurcksetzen.
  //----------------------------------------------------------------------------
  FObject := nil;
  FLastTexID := AC_NO_TEXTURE;
  FLastMatID := -1;
  FPalCount := 0;
end;


////////////////////////////////////////////////////////////////////////////////
// LoadAC3D
//------------------------------------------------------------------------------
// Diese Ladefunktion ist das eigentliche Kernstck. Hier werden die Daten aus
// der AC3D-Datei herrausgelesen und entsprechend weiterverarbeitet.
//------------------------------------------------------------------------------
// Parameter: Stream -> Der Stream, aus welchem die Daten geladen werden sollen.
// Rckgabe : Result -> True, wenn die Daten geladen werden konnten, anderen
//                      Falls False.
function TAC3DFile.LoadAC3D: Boolean;
var
  OldDecSep : Char;
begin
  Result := False;

  //----------------------------------------------------------------------------
  // Zuersteinmal berprfen, ob es sich berhaupt um eine AC3D-Datei handelt.
  //----------------------------------------------------------------------------
  if Copy(ReadLine, 1, 4) <> 'AC3D' then
    raise
      Exception.Create('Not a valid AC3D file!');

  //----------------------------------------------------------------------------
  // Den bisherigen Decimal-Seperator merken und ihn auf '.' setzen, damit die
  // StrToFloat-Funktion korrekt arbeitet.
  //----------------------------------------------------------------------------
  OldDecSep := DecimalSeparator;
  DecimalSeparator := '.';
  
  //----------------------------------------------------------------------------
  // Den Start-Index der Palette setzen und das Objekt laden.
  //----------------------------------------------------------------------------
  //startmatindex := num_palette;
  FObject := LoadObject(nil);

  //----------------------------------------------------------------------------
  // Wenn das Objekt geladen werden konnte, die Normalen berrechnen und True
  // zurckliefern.
  //----------------------------------------------------------------------------
  if Assigned(FObject) then
  begin
    if not FNoData then ObjectCalcVertexNormals(FObject);
    Result := True;
  end;
  DecimalSeparator := OldDecSep;
end;


////////////////////////////////////////////////////////////////////////////////
// LoadFromFile
//------------------------------------------------------------------------------
// Ldt das AC3D-Objekt aus der bergebenen Datei.
//------------------------------------------------------------------------------
// Parameter: FileName -> Der Name der Datei, welche geladen werden soll.
// Rckgabe : Result   -> True, wenn der Vorgang erfolgreich war, anderen Falls
//                        False.
function TAC3DFile.LoadFromFile(FileName: string): Boolean;
var
  FileStream : TFileStream;
  c:boolean;
begin
  //----------------------------------------------------------------------------
  // Einfach einen Stream auf die bergebene Datei erstellen und die Daten aus
  // diesem Stream laden. Dannach ein Try-Finnaly-Block, damit auf jeden Fall
  // der erstellte Stream wieder freigegeben wird.
  //----------------------------------------------------------------------------
{$IFDEF OZONE}
  c:=TOzoneMainClass(owner).O3Ini.ReadBool('ac3d','cache',false);
  if c and FileExists(FileName+'.cache') then
   FNoData:=true;
{$ENDIF}
  FileStream := TFileStream.Create(FileName, fmOpenRead);
  try
    Result := LoadFromStream(FileStream);
  finally
    FileStream.Free;
  end;
{$IFDEF OZONE}
  if c then begin
   if not FNoData then
    CacheToFile(FileName+'.cache')
   else
    RestoreCache(FileName+'.cache');
  end;
{$ENDIF}
end;


////////////////////////////////////////////////////////////////////////////////
// LoadFromResource
//------------------------------------------------------------------------------
// Ldt die AC3D-Datei aus einer Resource.
//------------------------------------------------------------------------------
// Parameter: Instance -> Die Instance des Modules, aus welchem die Daten ge-
//                        laden werden sollen.
//            ResName  -> Der Name der zu ladenden Resource.
// Rckgabe : --- Standard ---
function TAC3DFile.LoadFromResource(Instance: THandle; ResName: string):Boolean;
var
  ResStream : TResourceStream;
begin
  //----------------------------------------------------------------------------
  // Den Resource-Stream erstellen und dabei die bergebenen Daten verwenden.
  // Der Typ der Resource kann auf NIL gesetzt werden, da es sich um eine Be-
  // nutzerdefinierte Resource habdelt. Dannach alles weitere in einem Try
  // Finally Block, damit der erstellte Stream auf jeden Fall wieder freige-
  // geben wird.
  //----------------------------------------------------------------------------
  ResStream := TResourceStream.Create(Instance, ResName, nil);
  try
    Result := LoadFromStream(ResStream);
  finally
    ResStream.Free;
  end;
end;


////////////////////////////////////////////////////////////////////////////////
// LoadFromStream                                                 
//------------------------------------------------------------------------------
// Merkt sich dem bergebenen Stream und ruft die Funktion zum Laden der Daten
// aus dem bergebenen Stream auf.
//------------------------------------------------------------------------------
// Parameter: --- Standard ---
// Rckgabe : --- Standard ---
function TAC3DFile.LoadFromStream(AStream: TStream): Boolean;
begin
  //----------------------------------------------------------------------------
  // Zuersteinmal berprfen, ob AStream auch glltig ist.
  //----------------------------------------------------------------------------
  if not Assigned(AStream) then
    raise Exception.Create('Undefined stream object!');

  //----------------------------------------------------------------------------
  // Den Stream bernehmen und die Datei laden. Dannach kann der Stream wieder
  // auf NIL gesetzt werden, da er nicht mehr bentigt wird.
  //----------------------------------------------------------------------------
  FStream := AStream;
  Result := LoadAC3D;
  FStream := nil;
end;


////////////////////////////////////////////////////////////////////////////////
// NewObject
//------------------------------------------------------------------------------
// Erstellt Speicher fr ein neuen Objekt und fllt dieses gleich mit Defult-
// werten. Der reservierte Speicher wird dann zurckgeliefert.
//------------------------------------------------------------------------------
// Parameter: --- Keine ---
// Rckgabe : Result -> Zeiger auf den erstellten Speicher.
function TAC3DFile.NewObject: PACObject;
begin
  //----------------------------------------------------------------------------
  // Einfach den Speicher mit New anlegen, die einzelnen Felder mit den Def-
  // fault-Werten fllen und die Adresse zurckliefern.
  //----------------------------------------------------------------------------
  New(Result);

  with Result^ do
  begin
    Scale[0] := 1;
    Scale[1] := 1;
    Scale[2] := 1;
    Rotate[0]:= 0;
    Rotate[1]:= 0;
    Rotate[2]:= 0;
    Trans[0] := 0;
    Trans[1] := 0;
    Trans[2] := 0;
    Loc.x    := 0;
    Loc.y    := 0;
    Loc.z    := 0;
    Name     := '';
    URL      := '';
    Data     := '';
//    Vertices := nil;
    NumVert  := 0;
//    Surfaces := nil;
    NumSurf  := 0;
    Texture  := AC_NO_TEXTURE;
    TextureRepeat_X := 1;
    TextureRepeat_Y := 1;
    TextureOffset_X := 0;
    TextureOffset_Y := 0;
    Matr[0]:=1;
    Matr[1]:=0;
    Matr[2]:=0;
    Matr[3]:=0;

    Matr[4]:=0;
    Matr[5]:=1;
    Matr[6]:=0;
    Matr[7]:=0;

    Matr[8]:=0;
    Matr[9]:=0;
    Matr[10]:=1;
    Matr[11]:=0;

    Matr[12]:=0;
    Matr[13]:=0;
    Matr[14]:=0;
    Matr[15]:=1;

    Kids      := nil;
    NumKids   := 0;
    Matrix[0] := 1;
    Matrix[1] := 0;
    Matrix[2] := 0;
    Matrix[3] := 0;
    Matrix[4] := 1;
    Matrix[5] := 0;
    Matrix[6] := 0;
    Matrix[7] := 0;
    Matrix[8] := 1;
  end;
end;


////////////////////////////////////////////////////////////////////////////////
// EndOfStream
//------------------------------------------------------------------------------
// Liefert True zurck, wenn das Ende des Streams erreicht wurde.
//------------------------------------------------------------------------------
// Parameter: --- Keine ---
// Rckgabe : Result -> True, wenn das Ende des Streams erreicht wurde, sonst
//                      False.
function TAC3DFile.EndOfStream: Boolean;
begin
  Result := FStream.Position = FStream.Size;
end;


////////////////////////////////////////////////////////////////////////////////
// ReadLine
//------------------------------------------------------------------------------
// Liest aus dem bergebenen Stream die nchste Text-Zeile herraus.
//------------------------------------------------------------------------------
// Parameter: --- Keine ---
// Rckgabe : Result -> Die herrausgelesene Zeile.
function TAC3DFile.ReadLine: string;
var
  CurChar : Char;
begin
  //----------------------------------------------------------------------------
  // Result vorinitialisieren damit kein Datenmll vorhanden ist. Dann auf jeden
  // fall einmal ein Zeichen aus dem Stream herrauslesen.
  //----------------------------------------------------------------------------
  Result := '';
  if EndOfStream then Exit;
  FStream.Read(CurChar, 1);

  //----------------------------------------------------------------------------
  // Ok, das ist nun etwas komisch, aber leider notwendig. Solange alle Daten
  // aus dem Stream herrauslesen (und zwar einzeln), bis das Ende der Zeile
  // (#10) oder das Ende des Stream erreicht wurde.
  //            Korrektur: JA, es ist wirklich #10 und NICHT #13!
  //----------------------------------------------------------------------------
  while (CurChar <> #10) and (not EndOfStream) do
  begin
    Result := Result + CurChar;
    FStream.Read(CurChar, 1);
  end;

  //----------------------------------------------------------------------------
  // Das es evtl. vorkommen kann, das nun ein #10 am Anfang des Strings ist
  // (das kommt von der unterschiedlichen Speicher-Variante von Windows bzw.
  // Linux/Unix bei Test-Dateien) alle Zeichen unterhalb von #32 aus dem Er-
  // gebnis-String entfernen.
  //            Korrektur: Es kann ein #13 am Anfang sein, ein #10 (s. oben)!
  //----------------------------------------------------------------------------
  Result := Trim(Result);

  //----------------------------------------------------------------------------
  // Im Original AC3D-Loader wurde nun an dieser Stelle ein Zhler fr die
  // aktuelle Zeile erht. Dies ist aber nur fr die Fehler-Diagnose notwendig.
  // Da ich davon ausgehe, dass die dem Loader bergebenen Dateien von AC3D
  // problemlos geladen werden knnen, hab' ich mir das erspart. Wer will oder
  // notwendigkeit dafr sieht, muss das dann eben an dieser Stelle imple-
  // mentieren.
  //----------------------------------------------------------------------------
end;


////////////////////////////////////////////////////////////////////////////////
// InitSurface                                                     
//------------------------------------------------------------------------------
// Initialisiert ein bergebenes Surface mit den Default-Werteln.
//------------------------------------------------------------------------------
// Parameter: Surface -> Das zu initialisierende Surface.
// Rckgabe : Surface -> Gefllt mit den Default-Werten.
procedure TAC3DFile.InitSurface(var Surface: TACSurface);
begin
  with Surface do
  begin
//    Vertref    := nil;
//    UVs        := nil;
    NumVertRef := 0;
    Flags      := 0;
    Material   := 0;
    Normal.x   := 0.0;
    Normal.z   := 0.0;
    Normal.z   := 0.0;
  end;
end;


////////////////////////////////////////////////////////////////////////////////
// TriCalcNormal                                                   
//------------------------------------------------------------------------------
// Berechnet die Normale zwischen den Vectoren v1, v2 und V3. Der Wert der
// Normalen wird in n zurckgeliefert.
//------------------------------------------------------------------------------
// Parameter: v1, v2, v3 -> Die 3 Vectoren, fr welche die Normale berechnet
//                          werden soll.
// Rckgabe : Result     -> Die Berechnete Normale.
function TAC3DFile.TriCalcNormal(v1, v2, v3: TACPoint): TACNormal;
var
  len : double;
begin
  //----------------------------------------------------------------------------
  // Wer wissen will, was hier gemacht wird, soll sich einfach mal mit 3D-Mathe-
  // mathik beschftigen.
  //----------------------------------------------------------------------------
  Result.x := (v2.y - v1.y) * (v3.z - v1.z) - (v3.y - v1.y) * (v2.z - v1.z);
  Result.y := (v2.z - v1.z) * (v3.x - v1.x) - (v3.z - v1.z) * (v2.x - v1.x);
  Result.z := (v2.x - v1.x) * (v3.y - v1.y) - (v3.x - v1.x) * (v2.y - v1.y);
  len := sqrt(Result.x * Result.x + Result.y * Result.y + Result.z * Result.z);

  if (len > 0) then
  begin
    Result.x := Result.x / len;
    Result.y := Result.y / len;
    Result.z := Result.z / len;
  end;
end;



////////////////////////////////////////////////////////////////////////////////
// ReadSurface
//------------------------------------------------------------------------------
// Liest die Surface-Daten aus der AC3D-Datei ein.
//------------------------------------------------------------------------------
// Parameter: Surface -> Das zu fllende Surface.
//            Obj     -> Das Objekt, welchem das Surface zugehrig ist.
// Rckgabe : Result  -> True, wenn das Surface ordnungsgem geladen werden
//                       konnte, anderen Falls False.
function TAC3DFile.ReadSurface(var Surface: TACSurface;
  Obj: PACObject): Boolean;
var
  Token,
  CurLine : string;
  i : integer;
begin
  //----------------------------------------------------------------------------
  // Result und das bergebene Surface vorinitialisieren.
  //----------------------------------------------------------------------------
  Result := False;
  InitSurface(Surface);

  //----------------------------------------------------------------------------
  // Sollte das Ende der Datei erreicht werden, OHNE dass das Surface komplett
  // eingeladen wurde, ist die Datei Fehler haft. Also im Falle von EOF die
  // Schleife zum Einlesen des Surfaces abbrechen.
  //----------------------------------------------------------------------------
  while not EndOfStream do
  begin
    //--------------------------------------------------------------------------
    // Die nchste Zeile Einlesen und das Erste Token ermitteln.
    //--------------------------------------------------------------------------
    CurLine := Trim(ReadLine);
    Token   := GetNextToken(' ', CurLine);

    //--------------------------------------------------------------------------
    // berprfen, ob es sich beim aktuellen Token um ein Surface handelt.
    //--------------------------------------------------------------------------
    if (Token = 'SURF') then
    begin
      //------------------------------------------------------------------------
      // Ein Surface muss einen Parameter haben. Also den Rest von CurLine in
      // einen Integer-Wert konvertieren. Sollte dabei ein Fehler auftreten,
      // ist ein Fehler in der AC3D-Datei vorhanden. Also eine Exception aus-
      // lsen. Wenn mehr als ein Parmeter angegeben wurden, so wird der
      // Parameter, welcher zu viel ist, einfach ignoriert.
      //------------------------------------------------------------------------
      try
        Surface.Flags := StrToInt(GetNextToken(' ', CurLine));
      except
        raise
          Exception.Create('''SURF'' should be followed by one flags argument');
      end;
      Continue;
    end else

    //--------------------------------------------------------------------------
    // berprfen, ob es sich um das Token 'mat' handelt.
    //--------------------------------------------------------------------------
    if (Token = 'mat') then
    begin
      //------------------------------------------------------------------------
      // Auch hier gillt wieder das gleiche wie auch schon beim Surface.
      //------------------------------------------------------------------------
      try
        Surface.Material := StrToInt(GetNextToken(' ', CurLine));
      except
        raise
          Exception.Create('''mat'' should be followed by one flags argument');
      end;
      Continue;
    end else

    //--------------------------------------------------------------------------
    // Wenn es sich bei dem Token um 'refs' handelt, die entsprechenden Daten
    // einlesen.
    //--------------------------------------------------------------------------
		if Token = 'refs' then
    begin

      //------------------------------------------------------------------------
      // Auch hier gillt wirde: berfrlssige Parameter werden ignoriert, fal-
      // sche Parameter entsprechend mit einer Exception gemeldet.
      //------------------------------------------------------------------------
      try

        //----------------------------------------------------------------------
        // Die Anzahl ermitteln und dem entsprechenden Feld im Surface zuweisen.
        // Dannach die Dynamischen Arrays des Surfaces auf die entsprechende
        // Lnge setzen.
        //----------------------------------------------------------------------
        Surface.NumVertRef := StrToInt(GetNextToken(' ', CurLine));
//        SetLength(Surface.Vertref, Surface.NumVertRef);
//        SetLength(Surface.UVs, Surface.NumVertRef);

        //----------------------------------------------------------------------
        // Nun im folgenden die Vertex-Nummern und U/V-Textur-Koordinaten
        // einlesen.
        //----------------------------------------------------------------------
        for i := 0 to Surface.NumVertRef - 1 do
        begin
          //--------------------------------------------------------------------
          // Nun wird VOR jedem Einlesen berprft, ob das Ende der Datei er-
          // reicht wurde. Wenn ja, eine Exception auslesen. Wenn nein, die
          // nchste zeile auslesen damit Vertex-Daten geladen werden.
          //--------------------------------------------------------------------
          if EndOfStream then
            raise Exception.Create('Error while reading Surface-Data');
          CurLine := ReadLine;

          //--------------------------------------------------------------------
          // Nun die Vertex-Nummern und U/V-Daten aus dem String auslesen und
          // entsprechend konvertieren.
          //--------------------------------------------------------------------
          Surface.Vertref[i] := StrToInt(GetNextToken(' ', CurLine));
          Surface.UVs[i].u   := StrToFloat(GetNextToken(' ', CurLine));
          Surface.UVs[i].v   := StrToFloat(GetNextToken(' ', CurLine));
        end;

        //----------------------------------------------------------------------
        // Nun berprfen, ob mindestens 3 Vertex-Daten fr dieses surface vor-
        // handen sind. Wenn ja, die Normale fr dieses Surface ausrechnen.
        //----------------------------------------------------------------------
        if (Surface.NumVertRef >= 3) then
          Surface.Normal :=
            TriCalcNormal(Obj^.Vertices[Surface.Vertref[0]].Vertex,
                          Obj^.Vertices[Surface.Vertref[1]].Vertex,
                          Obj^.Vertices[Surface.Vertref[2]].Vertex);

         //----------------------------------------------------------------------
        // Ok, hier angekommen, wurde das Surface ordnungsgem. Also True zu-
        // rckliefern und die Funktion verlassen.
        //----------------------------------------------------------------------
        Result := True;
        Exit;
      except
        raise Exception.Create('Unexpected end of file!');
		  end;
    end;
  end;
end;


////////////////////////////////////////////////////////////////////////////////
// ObjectCalcVertexNormals
//------------------------------------------------------------------------------
// Berechnet die Vertex-Normalen des bergebenen Objektes.
//------------------------------------------------------------------------------
// Parameter: Obj -> Das Objekt, dessen Vertex-Normale berechnet werden sollen.
// Rckgabe : --- Standard ---
procedure TAC3DFile.ObjectCalcVertexNormals(Obj: PACObject);
var
  SurfCnt,
  VertCnt,
  VertrefCnt,
  found, i    : Integer;
  Normal      : TACNormal;
  Surface     : TACSurface;
begin
  //----------------------------------------------------------------------------
  // Nun fr jeden Vertex innerhalb des Objectes die Normale berechnen.
  //----------------------------------------------------------------------------
  for VertCnt := 0 to Obj^.NumVert - 1 do
  begin
    //--------------------------------------------------------------------------
    // Die Variablen vorinitialisieren.
    //--------------------------------------------------------------------------
    Normal.x := 0;
    Normal.y := 0;
    Normal.z := 0;
    found := 0;

    //--------------------------------------------------------------------------
    // Alle Surface durcharbeiten.
    //--------------------------------------------------------------------------
    for SurfCnt := 0 to Obj^.NumSurf - 1 do
    begin
      //--- Das Surface mit der "ID" von SurfCnt ermitteln. --------------------
	    Surface := Obj^.Surfaces[SurfCnt];

      //------------------------------------------------------------------------
      // Nun alle Vertex-Referenzen durchlaufen.
      //------------------------------------------------------------------------
	    for VertRefCnt := 0 to Surface.NumVertRef - 1 do
      begin
        //----------------------------------------------------------------------
        // berprfen, ob der Vertex im aktuellen Surface verwendet wird. Wenn
        // ja, die Normale des Surfaces zur Normale des Objektes hinzufgen.
        //----------------------------------------------------------------------
        if (Surface.Vertref[VertRefCnt] = VertCnt) then
		    begin
          Normal.x := Normal.x + Surface.normal.x;
		      Normal.y := Normal.y + Surface.normal.y;
          Normal.z := Normal.z + Surface.normal.z;
          Inc(found);
        end;
      end;
    end;

    //--------------------------------------------------------------------------
    // Wenn mindestens ein Glltiger Vertex zu einem Surface (und somit mindes-
    // tens eine Normale) gefunden wurde, die Durchschnittsnormale fr das
    // Objekt berechnen.
    //--------------------------------------------------------------------------
    if (found > 0) then
    begin
	    Normal.x := Normal.x / found;
	    Normal.y := Normal.y / found;
	    Normal.z := Normal.z / found;
    end;

    //--- Zum Schlu noch die ermittelte Objekt-Normale dem Objekt zuweisen. ---
    Obj^.Vertices[VertCnt].normal := Normal;
  end;

  //----------------------------------------------------------------------------
  // Jezt noch fr die Childs des Objektes die Normalen berechnen.
  //----------------------------------------------------------------------------
  for i := 0 to Obj^.NumKids - 1 do
    ObjectCalcVertexNormals(Obj^.Kids[i]);
end;


////////////////////////////////////////////////////////////////////////////////
// StringToObjectType
//------------------------------------------------------------------------------
// Liefert den Object-Typ, welcher als String bergeben wird, als Integer zu-
// rck.
//------------------------------------------------------------------------------
// Parameter: s      -> Der String mit dem Objekt-Typ.
// Rckgabe : Result -> Der Integer-Wert des Objekt-Typs.
function TAC3DFile.StringToObjectType(s: string): Integer;
begin
  //----------------------------------------------------------------------------
  // Einfach berprfen, was fr ein Typ im String steht und den entsprechenden
  // Integer-Wert zurckliefern.
  //----------------------------------------------------------------------------
  if s = 'world' then Result := AC_OBJECT_WORLD else
  if s = 'poly'  then Result := AC_OBJECT_NORMAL else
  if s = 'group' then Result := AC_OBJECT_GROUP else
  if s = 'light' then Result := AC_OBJECT_LIGHT else
    Result := AC_OBJECT_NORMAL;
end;


////////////////////////////////////////////////////////////////////////////////
// LoadObject                                                     
//------------------------------------------------------------------------------
// Ldt ein AC-Object und trgt als Kind dem bergebenen Parent ein.
//------------------------------------------------------------------------------
// Parameter: Parent -> Der Vater des Objektes. Das Objekt, welches nun gela-
//                      den wird, ist diesem bergebenen Vater untergeordnet.
// Rckgabe : Result -> Zeiger auf das Geladene Objekt.
function TAC3DFile.LoadObject(Parent: PACObject): PACObject;
var
  Token,
  CurLine  : string;
  i        : Integer;
  CurObj   : PACObject;
  CurMat   : TACMaterial;
begin
  //----------------------------------------------------------------------------
  // Die Variablen initialisieren.
  //----------------------------------------------------------------------------
  CurObj := nil;

  //----------------------------------------------------------------------------
  // Alle Zeilen der AC3D-Datei auslesen und verarbeiten. Dabei werden die ein-
  // zelnen Blcke in einem Try-Except-Block gefasst. Denn wenn ein Wert bzw.
  // ein Parameter eines Tolens fehlt, eine Exception auslsen. berzhlige Pa-
  // rameter werden dabei ignoriert.
  //----------------------------------------------------------------------------
  while not EndOfStream do
  begin
    //--------------------------------------------------------------------------
    // Die nchste Zeile auslesen und das erste Token zum feststellen des zu
    // bearbeitenden "Befehls" herrausfiltern.
    //--------------------------------------------------------------------------
	  CurLine := ReadLine;
    Token   := GetNextToken(' ', CurLine);

    //--------------------------------------------------------------------------
    // Von nun an die verschiedenen AC3D-Tokens berprfen. Zuerst berprfen,
    // ob es sich beim aktuellen Token um ein Material handelt.
    //--------------------------------------------------------------------------
    if Token = 'MATERIAL' then
    try
      //------------------------------------------------------------------------
      // Den Namen des Materials auslesen. Dabei das " am Anfang und Ende ent-
      // fernen (also vom 2 Zeichen bis zum vorletztem Zeichen alles herraus-
      // kopieren).
      //------------------------------------------------------------------------
      CurMat.Name  := GetNextToken(' ', CurLine);
      CurMat.Name  := Copy(CurMat.Name, 2, Length(CurMat.Name) - 2);

      //------------------------------------------------------------------------
      // Das nchste Wort ist RGB, da nun die RGB-Werte kommen. Also einmal
      // GetNextToken aufrufen, um das Wort RGB zu entfernen und dannach
      // die einzelnen R G und B Werte konvertieren.
      //------------------------------------------------------------------------
      GetNextToken(' ', CurLine);
      CurMat.RGB.r := StrToFloat(GetNextToken(' ', CurLine));
      CurMat.RGB.g := StrToFloat(GetNextToken(' ', CurLine));
      CurMat.RGB.b := StrToFloat(GetNextToken(' ', CurLine));

      //------------------------------------------------------------------------
      // Das nchste ist Ambient. Auch hier wieder das Wort Ambient entfernen
      // und die Ambient-Werte auslesen.
      //------------------------------------------------------------------------
      GetNextToken(' ', CurLine);
      CurMat.Ambient.r := StrToFloat(GetNextToken(' ', CurLine));
      CurMat.Ambient.g := StrToFloat(GetNextToken(' ', CurLine));
      CurMat.Ambient.b := StrToFloat(GetNextToken(' ', CurLine));

      //------------------------------------------------------------------------
      // Nun das Emessive-Light auslesen.
      //------------------------------------------------------------------------
      GetNextToken(' ', CurLine);
      CurMat.Emissive.r := StrToFloat(GetNextToken(' ', CurLine));
      CurMat.Emissive.g := StrToFloat(GetNextToken(' ', CurLine));
      CurMat.Emissive.b := StrToFloat(GetNextToken(' ', CurLine));

      //------------------------------------------------------------------------
      // Nun das Specular-Light.
      //------------------------------------------------------------------------
      GetNextToken(' ', CurLine);
      CurMat.Specular.r := StrToFloat(GetNextToken(' ', CurLine));
      CurMat.Specular.g := StrToFloat(GetNextToken(' ', CurLine));
      CurMat.Specular.b := StrToFloat(GetNextToken(' ', CurLine));

      //------------------------------------------------------------------------
      // Zum Schlu noch die Werte fr shininess und Tranzparenz auslesen.
      //------------------------------------------------------------------------
      GetNextToken(' ', CurLine);
      CurMat.Shininess := StrToFloat(GetNextToken(' ', CurLine));
      GetNextToken(' ', CurLine);
      CurMat.Transparency := StrToFloat(GetNextToken(' ', CurLine));

      //------------------------------------------------------------------------
      // Das ermittelte Material zur Palette hinzufgen. Dazu den Counter um
      // eins erhhen, damit die neue Lnge des Arrays gesetzt werden kann und
      // anschlieend das Material zuweisen. Dabei wird FPalCount - 1 verwen-
      // det, um auf das letzte Item des Arrays zuzugreifen. -1, da das Array
      // 0-basierend ist.
      //------------------------------------------------------------------------
      Inc(FPalCount);
      SetLength(FPalette, FPalCount);
      FPalette[FPalCount - 1] := CurMat;
      Continue;
    except
      raise
        Exception.Create('expected 21 params after ''MATERIAL''');
    end else

    //--------------------------------------------------------------------------
    // Handelt es sich beim aktuellen Token um ein Objekt?
    //--------------------------------------------------------------------------
	  if (Token = 'OBJECT') then
    begin
      //------------------------------------------------------------------------
      // Wenn ja ein neues Object erstellen. Der Rest von CurLine enthllt dann
      // noch den Objekt-Typ. Diesen auslesen, in einen Object-Typ als Integer
      // konvertieren und dem Objekt zuweisen.
      //------------------------------------------------------------------------
      CurObj := NewObject;
      CurObj^.ObjType := StringToObjectType(GetNextToken(' ', CurLine));
      Continue;
    end else

    //--------------------------------------------------------------------------
    // Handelt es sich um den Data-Bereich?
    //--------------------------------------------------------------------------
    if (Token = 'data') then
    begin
      //------------------------------------------------------------------------
      // Im Rest von CmdLine steht die Anzahl der zu lesenden Daten fr Data.
      // Diese einfach einlesen und dem Objekt zuweisen.
      //------------------------------------------------------------------------
      while (Length(CurObj^.Data) < StrToInt(CurLine)) and (not EndOfStream) do
        CurObj^.Data := CurObj^.Data + Trim(ReadLine) + #13;
      Continue;
		end else

    //--------------------------------------------------------------------------
    // Handelt es sich um den Namen des Objektes ?
    //--------------------------------------------------------------------------
    if (Token = 'name') then
    begin
      //------------------------------------------------------------------------
      // Auch hier gillt wieder: berschssige Parameter ignorieren. Also ein-
      // fach den Rest von CurLine als Objekt-Namen verwenden.
      //------------------------------------------------------------------------
      CurObj^.Name := CurLine;
      Continue;
    end else

    //--------------------------------------------------------------------------
    // Handelt es sich um die zu verwendende Textur?
    //--------------------------------------------------------------------------
    if (Token = 'texture') then
    begin
      //------------------------------------------------------------------------
      // Sollte CurLine jetzt kein Textur-Name stehen, eine Exception mit dem
      // Fehler ausgeben.
      //------------------------------------------------------------------------
      if CurLine = '' then
      begin
        raise
          Exception.Create('Expected quoted texture name!');
        Exit;
      end;

      //------------------------------------------------------------------------
      // Ok, die Anfhrungs-Zeichen am Anfang und am Ende entfernen und die
      // den Namen der Textur-Datei im Objekt ablegen (damit die Textur per
      // event beim Freigeben des Objektes ebenfalls freigegeben werden kann).
      // Dannach die Prozedur zum Laden der Lextur aufrufen falls sie zugewie-
      // sen wurde.
      //------------------------------------------------------------------------
      CurLine := Copy(CurLine, 2, Length(CurLine) - 2);
      CurObj^.TextureFile := CurLine;
{$IFDEF OZONE}
      if Assigned(owner) then
        TOzoneMainClass(owner).gfxSys.LoadTexture(CurObj^.Texture, CurObj^.TextureFile, TextureSubDir);
{$ENDIF}
      Continue;
    end else

    //--------------------------------------------------------------------------
    // Handelt es sich um die Repeat Daten der Textur?
    //--------------------------------------------------------------------------
    if (Token = 'texrep') then
    try
      //------------------------------------------------------------------------
      // Einfach die nchsten Werte in einen Float-Wert konvertieren und dem
      // entsprechenden Feld zuweisen.
      //------------------------------------------------------------------------
      CurObj^.TextureRepeat_X := StrToFloat(GetNextToken(' ', CurLine));
      CurObj^.TextureRepeat_Y := StrToFloat(GetNextToken(' ', CurLine));
      Continue;
    except
    end else

    //--------------------------------------------------------------------------
    // Handelt es sich um die Offsets der Textur (U/V-Daten) ?
    //--------------------------------------------------------------------------
    if (Token = 'texoff') then
    try
      //------------------------------------------------------------------------
      // Die Texturen-Offsets auslesen und konvertiert in die entsprechenden
      // Fehler ablegen.
      //------------------------------------------------------------------------
      CurObj^.TextureOffset_X := StrToFloat(GetNextToken(' ', CurLine));
      CurObj^.TextureOffset_Y := StrToFloat(GetNextToken(' ', CurLine));
      Continue;
    except
    end else

    //--------------------------------------------------------------------------
    // Handelt es sich um die Rotations-Matrix des Objektes?
    //--------------------------------------------------------------------------
    if (Token = 'rot') then
    try
      //------------------------------------------------------------------------
      // ersuchen, jeden Wert der Matrix zu kopieren. Im Fehlerfall sind die
      // Parameter nicht richtig.
      //------------------------------------------------------------------------
      for i := 0 to 8 do
        CurObj^.Matrix[i] := StrToFloat(GetNextToken(' ', CurLine));

      Continue;
    except
    end else

    //--------------------------------------------------------------------------
    // Handelt es sich um die Position des Objektes im 3D-Raum?
    //--------------------------------------------------------------------------
    if (Token = 'loc') then
    try
      //------------------------------------------------------------------------
      // Die nchsten drei Parameter Konvertieren und den Positions-Parameter
      // Zuweisen.
      //------------------------------------------------------------------------
      CurObj^.Loc.x := StrToFloat(GetNextToken(' ', CurLine));
      CurObj^.Loc.y := StrToFloat(GetNextToken(' ', CurLine));
      CurObj^.Loc.z := StrToFloat(GetNextToken(' ', CurLine));
      Continue;
    except
    end else

    //--------------------------------------------------------------------------
    // Handelt es sich um die URL, welche diesem Objekt zugewiesen wurde?
    //--------------------------------------------------------------------------
    if (Token = 'url') then
    begin
      //------------------------------------------------------------------------
      // Hier werden falsche Parameter komplett ignoriert. Einfach den Rest von
      // CurLine als URL dem Objekt zuweisen.
      //------------------------------------------------------------------------
      CurObj^.URL := CurLine;
      Continue;
    end else

    //--------------------------------------------------------------------------
    // Handelt es sich um die Vertex-Nummer?
    //--------------------------------------------------------------------------
    if (Token = 'numvert') then
    try
      //------------------------------------------------------------------------
      // Zuersteinmal die Anzahl der Vertices ermitteln, in NumVert ablegen und
      // das Array auf die entsprechende Gre setzen.
      //------------------------------------------------------------------------
      CurObj^.NumVert := StrToInt(GetNextToken(' ', CurLine));
//      SetLength(CurObj^.Vertices, CurObj^.NumVert);

      //----------------------------------------------------------------------
      // Nun alle Vertieces auslesen und dem Array zuweisen. Dabei aus dem
      // String aus der Datei die einzelnen X, Y und Z Werte ermitteln, konver-
      // tieren und im Array and aktueller Position dem entsprechenden
      // Feld zuweisen.
      //----------------------------------------------------------------------
      if not FNoData then begin
       for i := 0 to CurObj^.NumVert - 1 do
       begin
         CurLine := ReadLine;
         CurObj^.Vertices[i].Vertex.x := StrToFloat(GetNextToken(' ', CurLine));
         CurObj^.Vertices[i].Vertex.y := StrToFloat(GetNextToken(' ', CurLine));
         CurObj^.Vertices[i].Vertex.z := StrToFloat(GetNextToken(' ', CurLine));
       end;
      end;
      Continue;
    except
    end else

    //--------------------------------------------------------------------------
    // Handelt es sich um die Anzahl der Surfaces?
    //--------------------------------------------------------------------------
    if (Token =  'numsurf') then
    try
      //------------------------------------------------------------------------
      // Die Anzanl der Surfaces aus CurLine auslesen und im aktuellen Objekt
      // ablegen und die Gre des Arrays entsprechend setzen.
      //------------------------------------------------------------------------
      CurObj^.NumSurf := StrToInt(GetNextToken(' ', CurLine));
  //    SetLength(CurObj^.Surfaces, CurObj^.NumSurf);

      //------------------------------------------------------------------------
      // Nun alle Surface-Daten aus der Datei auslesen und im Array ablegen.
      // Sollte dabei ein Fehler auftreten, eine Exception auslsen.
      //------------------------------------------------------------------------
      if not FNoData then begin
       for i := 0 to CurObj^.NumSurf  - 1 do
       begin
         if not ReadSurface(CurObj^.Surfaces[i], CurObj) then
           raise
             Exception.Create('Error whilst reading surface!');
       end;
      end;
      Continue;
    except
    end else

    if (Token = 'kids') then // 'kids' is the last token in an object
    begin
      //------------------------------------------------------------------------
      // Die Anzahl der Kids befindet sich im Rest von CurLine. Fr den Fall,
      // dass noch mehr Parameter nach Kids stehen (welche ignoriert werden),
      // nochmal GetNextToken aufrufen-
      //------------------------------------------------------------------------
      CurObj^.NumKids := StrToInt(GetNextToken(' ', CurLine));

      //------------------------------------------------------------------------
      // Die Kids mssem logischerweise nur dann geladen werden, wenn auch
      // welche vorhanden sind.
      //------------------------------------------------------------------------
      if CurObj^.NumKids <> 0 then
      begin

        //----------------------------------------------------------------------
        // Den Speicher fr die Kids reservieren und alle Laden.
        //----------------------------------------------------------------------
        SetLength(CurObj^.Kids, CurObj^.NumKids);
        for i := 0 to CurObj^.NumKids - 1 do
        begin

          //--------------------------------------------------------------------
          // Wenn beim Laden des Kids nil zurckgelifert wurde, trat ein Fehler
          // auf. In diesem Fall eine Exceptio auslsen.
          //--------------------------------------------------------------------
          CurObj^.Kids[i] := LoadObject(CurObj);
          if CurObj^.Kids[i] = nil then
            raise Exception.Create('Error reading expected child object!');
        end;
      end;

      //------------------------------------------------------------------------
      // Da Kids immer am Ende der Liste kommt, kann die Funktion verlassen wer-
      // den.
      //------------------------------------------------------------------------
      Result := CurObj;
      Exit;
    end;
  end;

  Result := CurObj;
end;


////////////////////////////////////////////////////////////////////////////////
// Render                                                         
//------------------------------------------------------------------------------
// Rendert die aktuell geladenen Daten.
//------------------------------------------------------------------------------
// Parameter: --- Keine ---
// Rckgabe : --- Keine ---
procedure TAC3DFile.Render(x:integer=-1);
var
  Texture,
  Light    : Boolean;
  PolMode  : array[0..1] of GLenum;
begin
  //----------------------------------------------------------------------------
  // Wenn ein Objekt geladen wurde. Wenn nicht, die Funktion gleich weider ver-
  // lassen.
  //----------------------------------------------------------------------------
  if not Assigned(FObject) then Exit;

  //----------------------------------------------------------------------------
  // berprfen, ob die Render-States gesichert werden sollen. Wenn ja dies tun.
  //----------------------------------------------------------------------------
  if FSaveStates then
  begin
    glGetBooleanv(GL_LIGHTING, @Light);
    glGetBooleanv(GL_TEXTURE_2D, @Texture);
    glGetIntegerv(GL_POLYGON_MODE, @PolMode);
  end;

  //----------------------------------------------------------------------------
  // Zuerst festlegen, wie die Polygone dargestellt werden sollen. Dies soll
  // sich auf Front- und Back-Side der Ploygone auswirken, da ein- und zweisei-
  // tige Polygone in einem Objekt vorhanden sein knnen. Dannach das eigent-
  // liche Objekt rendern.
  //----------------------------------------------------------------------------
  glPolygonMode(GL_FRONT_AND_BACK, FPolygonMode);
  if x=-1 then DoRender(FObject)
          else DoRender(FObject.kids[x]);

  //----------------------------------------------------------------------------
  // Nun die gesicherten SaveStates weider herstellen, falls dies Erwnscht ist.
  // Dadurch hat der Benutzer die Mglichkeit, das AC3D-Objekt darstellen zu
  // knnen (incl. Texture-Mapping, Lightning, ect.) ohne sich nach dem Auf-
  // rufen von Render um die whrend des Rendern gesetzten RenderStates kmmern
  // zu mssen.
  //----------------------------------------------------------------------------
  if FSaveStates then
  begin
    if Light then glEnable(GL_LIGHTING) else glDisable(GL_LIGHTING);
    if Texture then glEnable(GL_TEXTURE_2D) else glDisable(GL_TEXTURE_2D);
    glPolygonMode(GL_FRONT, PolMode[0]);
    glPolygonMode(GL_BACK, PolMode[1]);
  end;
end;


////////////////////////////////////////////////////////////////////////////////
// GetPalEntry                                                     
//------------------------------------------------------------------------------
// Liefert den Paletten-Eintrag mit der bergebenen ID zurck.
//------------------------------------------------------------------------------
// Parameter: ID     -> Die ID des Eintrages aus der Palette, welcher ermittelt
//                      werden soll.
// Rckgabe : Result -> Das Material innerhalb der Palette mit der ID.
function TAC3DFile.GetPalEntry(ID: Integer): TACMaterial;
begin
  //----------------------------------------------------------------------------
  // Wenn die ID nicht auserhalb des Maximums liegt (Kleiner als Max - 1, da
  // das Array 0-basierend ist) das entsprechende Material zurckliefern.
  //----------------------------------------------------------------------------
  if ID < FPalCount then
    Result := FPalette[ID]
  else
    FillChar(Result, SizeOf(TACMaterial), #0);
end;


////////////////////////////////////////////////////////////////////////////////
// SetMaterial
//------------------------------------------------------------------------------
// Setzt das Material mit der bergebenen ID als aktuelles.
//------------------------------------------------------------------------------
// Parameter: --- Standard ---
// Rckgabe : --- Standard ---
procedure TAC3DFile.SetMaterial(MatID: Integer);
var
  Mat  : TACMaterial;
  RGBA : TACCol;
begin
  //----------------------------------------------------------------------------
  // berprfen, ob das Material vorher bereits gesetzt wurde. Wenn ja, kann
  // die Funktion wieder verlassen werden, da es ja noch aktiv ist. Falls
  // nicht, die ID des bergebenen Materials bernehmen, da es jetzt gesetzt
  // wird und das Material selber aus der Palette auslesen.
  //----------------------------------------------------------------------------
  if FLastMatID = MatID then
    Exit;
  FLastMatID := MatID;
  Mat := FPalette[MatID];

  //----------------------------------------------------------------------------
  // Den RGBA-Wert des Materiales ermitteln, mit der Transparentz des Materials
  // verrechnen und das ganze an OpenGL als diffuses Material bergeben.
  //----------------------------------------------------------------------------
  RGBA := Mat.RGB;
  RGBA.a := 1 - Mat.Transparency;
  glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, @RGBA);

  //----------------------------------------------------------------------------
  // Nun die Werte fr "Ambient" ermitteln, die Transparenzt verrechnen und das
  // ganze wieder an OpenGL bergeben.
  //----------------------------------------------------------------------------
  RGBA := Mat.Ambient;
  RGBA.a := 1 - Mat.Transparency;
  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, @RGBA);

  //----------------------------------------------------------------------------
  // Nun noch die Werte fr "Emission", "Specular" und "Shininess" setzen.
  // Damit sind alle notwendigen Farb-Werte an OpenGL bergeben.
  //----------------------------------------------------------------------------
  glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, @Mat.Emissive);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, @Mat.Specular);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, @Mat.Shininess);

  //----------------------------------------------------------------------------
  // Jetzt muss nur noch berprft werden, ob in RGBA der A-Wert < 1 ist. Wenn
  // ja, muss noch per Blending eine Transparenz gesetzt werden. Anderen Falls
  // das Blending ainfach ausschalten.
  //----------------------------------------------------------------------------
  if RGBA.a < 1 then
  begin
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);
  end;
//  glDisable(GL_BLEND);
end;


////////////////////////////////////////////////////////////////////////////////
// DoRender
//------------------------------------------------------------------------------
// Rendert das bergebene Objetk mitsamt dessen Unterobjekte.
//------------------------------------------------------------------------------
// Parameter: AObject -> Das zu rendernde Objekt.
// Rckgabe : --- keine ---
procedure TAC3DFile.DoRender(AObject: PACObject);
var
  SNr, i, st : Integer;
  tx, ty     : single;
  VertToUse  : TACVertex;
begin
  //----------------------------------------------------------------------------
  // Zuersteinmal die aktuelle Matrix sichern und das bergebene Objekt
  // Positionieren (von der aktuellen Position aus).
  //----------------------------------------------------------------------------
  glPushMatrix();
  glTranslated(AObject.Loc.x, AObject.Loc.y, AObject.Loc.z);

  //----------------------------------------------------------------------------
  // Wenn dem Objekt eine Textur und soll diese Verwendet werden?
  //----------------------------------------------------------------------------
  if (AObject.Texture <> AC_NO_TEXTURE) and FUseTexture then
  begin

    //--------------------------------------------------------------------------
    // Wenn ja, das Verwenden einer Textur in OpenGL einschalten und berprfe,
    //  ob die Textur vielleicht schon gesetzt wurde.
    //--------------------------------------------------------------------------
    glEnable(GL_TEXTURE_2D);
    if AObject.Texture <> FLastTexID then
    begin

      //------------------------------------------------------------------------
      // Endweder wurde bisher noch keine oder eine andere Textur gesetzt. Also
      // die des aktuellen Objektes setzen und als aktuelle Vermerken. Dabei
      // nicht vergessen, die Verwendung von Texturen in OpenGL auch einzu-
      // schalten.
      //------------------------------------------------------------------------
      glBindTexture(GL_TEXTURE_2D, AObject.Texture);
      FLastTexID := AObject.Texture;
    end;
  end else
    //--------------------------------------------------------------------------
    // Wenn keine Textur zugewiesen wurde, verhindern dass die zuletzt geladene
    // dargestellt wird.
    //--------------------------------------------------------------------------
    glDisable(GL_TEXTURE_2D);

  //----------------------------------------------------------------------------
  // Alle Surfaces verarbeiten und darstellen.
  //----------------------------------------------------------------------------
  for SNr := 0 to AObject.NumSurf - 1 do
  begin

    //--------------------------------------------------------------------------
    // Zursteinmal die Normale des aktuellen Surfaces bergeben.
    //--------------------------------------------------------------------------
    glNormal3fv(@AObject.Surfaces[SNr].Normal);

    //--------------------------------------------------------------------------
    // Nun berprfen, ob das Surface ein- oder zweiseitig ist und das OpenGL-
    // RenderState entsprechend setzen.
    //--------------------------------------------------------------------------
    if (AObject.Surfaces[SNr].Flags and AC_SURFACE_TWOSIDED)=AC_SURFACE_TWOSIDED
    then
	    glDisable(GL_CULL_FACE)
	  else
	    glEnable(GL_CULL_FACE);


    //--------------------------------------------------------------------------
    // Jetzt muss festgestellt werden, wass fr ein Typ OpenGL zum Rendern ver-
    // wenden soll. Zuerst berprfen, ob es sich bei dem aktuellen Surface-Typ
    // um eine geschlossene Linie handelt.
    //--------------------------------------------------------------------------
    st := AObject.Surfaces[SNr].Flags and $F;
    if st = AC_SURFACE_TYPE_CLOSEDLINE then
    begin
      //------------------------------------------------------------------------
      // Wenn ja, die Licht-Berechnung deaktivieren, damit es keine Probleme
      // gibt, den Typ Linie setzen und das Material des Objektes setzen.
      // Da es sich um eine Linie, muss vom Meterial lediglich die Farbe ge-
      // setzt werden.
      //------------------------------------------------------------------------
      glDisable(GL_LIGHTING);
      glBegin(GL_LINE_LOOP);
      glColor3fv(@FPalette[AObject.Surfaces[SNr].Material].RGB);
    end else

    //--------------------------------------------------------------------------
    // berprfen, ob es sich um eine einfache Linie handelt.
    //--------------------------------------------------------------------------
    if (st = AC_SURFACE_TYPE_LINE) then
    begin
      //------------------------------------------------------------------------
      // Auch hier wieder die Lichtrberechnung auschalten und OpenGL bescheid
      // sagen, um was es sich handelt. Ebenso wie oben muss auch hier vom
      // Meterial nur die Farbe gesetzt werden.
      //------------------------------------------------------------------------
      glDisable(GL_LIGHTING);
      glBegin(GL_LINE_STRIP);
      glColor3fv(@FPalette[AObject.Surfaces[SNr].Material].RGB);
    end else

    //--------------------------------------------------------------------------
    // Es handelt sich um keine Linienart, also die Opetionen fr die restlichen
    // Surface-Typen setzen.
    //--------------------------------------------------------------------------
    begin

      //------------------------------------------------------------------------
      // Feststellen, ob das OpenGL-Licht verwendet werden soll.
      //------------------------------------------------------------------------
      if FUseLight then
      begin

        //----------------------------------------------------------------------
        // Das das OpenGL-Licht verwendet werden soll, muss dieses aktiviert
        // und das Material des Surfaces an OpenGL bergeben werden.
        //----------------------------------------------------------------------
        glEnable(GL_LIGHTING);
        SetMaterial(AObject.Surfaces[SNr].Material);
      end else
      begin

        //----------------------------------------------------------------------
        // Wenn das Licht NICHT verwendet werden soll, es ausschalten. Das Ma-
        // terial des Surfaces zu setzen, wrde nichts bringen da die Material-
        // Eigenschaften von OpenGL nur mit aktiviertem Licht Verwendung finden.
        // Damit die Szene aber nicht so komisch einfarbig ist, wenigstens die
        // Farbe des Surfaces setzen.
        //----------------------------------------------------------------------
        glDisable(GL_LIGHTING);
        glColor3fv(@FPalette[AObject.Surfaces[SNr].Material].RGB);
      end;

      //------------------------------------------------------------------------
      // Kleine Optimierung. Wenn das aktuelle Surface aus 3 Vertices besteht,
      // das Surface als Triangle-Strip verwenden.
      //------------------------------------------------------------------------
      if AObject.Surfaces[SNr].NumVertRef = 3 then
        glBegin(GL_TRIANGLE_STRIP)
      else
        glBegin(GL_POLYGON);
	  end;

    //--------------------------------------------------------------------------
    // Ok, die Vorbereitungen sind abgeschlossen. Jetzt mssen nur noch die
    // Vertieces and OpenGL bergeben werden.
    //--------------------------------------------------------------------------
    for i := 0 to AObject.Surfaces[SNr].NumVertRef - 1 do
    begin
      //------------------------------------------------------------------------
      // Den aktuellen Vertex ermitteln.
      //------------------------------------------------------------------------
      VertToUse := AObject.Vertices[AObject.Surfaces[SNr].Vertref[i]];

      //------------------------------------------------------------------------
      // berprfen, ob dieses Objekt Texturiert ist und diese verwendet werden
      // soll.
      //------------------------------------------------------------------------
      if (AObject.Texture <> AC_NO_TEXTURE) and FUseTexture then
      begin

        //----------------------------------------------------------------------
        // Die Textur-Koordinaten berrechnen und an OpenGL bergeben. Diese
        // Berechnung wurde aus dem File-Loader von Andy direkt bernommen.
        //----------------------------------------------------------------------
        tx := AObject.TextureOffset_X + AObject.Surfaces[SNr].UVs[i].u *
              AObject.TextureRepeat_X;
        ty := AObject.TextureOffset_Y + AObject.Surfaces[SNr].UVs[i].v *
              AObject.TextureRepeat_Y;
        glTexCoord2f(tx, ty);
      end;

      //------------------------------------------------------------------------
      // Wenn das Surface "geshaded ist", die Normalen des Surfaces an OpenGL
      // bergeben.
      //------------------------------------------------------------------------
      if (AObject.Surfaces[SNr].Flags and AC_SURFACE_SHADED) = AC_SURFACE_SHADED
        then glNormal3fv(@VertToUse.Normal);

      //------------------------------------------------------------------------
      // Nun nur noch den Vertex an OpenGL bergeben.
      //------------------------------------------------------------------------
	    glVertex3fv(@VertToUse.Vertex);
    end;

    //--------------------------------------------------------------------------
    // Die Daten dieses Surfaces sind an OpenGL bergeben worden. Also teilen
    // wir dies OpenGL auch mit.
    //--------------------------------------------------------------------------
    glEnd;
  end;

  //----------------------------------------------------------------------------
  // Wenn es "Kinder" des Objektes gibt, diese ebenfalls darstellen.
  //----------------------------------------------------------------------------
  if AObject.NumKids <> 0 then
    for i := 0 to AObject.NumKids - 1 do
      DoRender(AObject.Kids[i]);

  //----------------------------------------------------------------------------
  // Da nun alles gezeichnet wurde, die am Anfang der Prozedur gesicherte Ma-
  // trix weider herstellen.
  //----------------------------------------------------------------------------
  glPopMatrix;
//  glTranslated(-AObject.Loc.x, -AObject.Loc.y, -AObject.Loc.z);
end;


////////////////////////////////////////////////////////////////////////////////
// CreateDisplayList
//------------------------------------------------------------------------------
// Renderd die Daten des Objektes in eine Display-Liste.
//------------------------------------------------------------------------------
// Parameter: --- Keine ---
// Rckgabe : --- Keine ---
procedure TAC3DFile.CreateDisplayList;
var x:cardinal;
begin
  //----------------------------------------------------------------------------
  // Falls bereits eine Display-Liste erstellt wurde, die Funktion verlassen.
  //----------------------------------------------------------------------------
  if FDispListID <> 0 then Exit;
  if FObject.NumKids <= 0 then Exit;
  //----------------------------------------------------------------------------
  // Eine neue Display-Liste erstellen und aktivieren damit sie gefllt werden
  // kann. Im fehlerfall eine Exception auslsen.
  //----------------------------------------------------------------------------

  FDispListID := glGenLists(FObject.NumKids);
  if FDispListID = 0 then
    raise Exception.Create('Error generating a DisplayList!');
  for x:=0 to FObject.NumKids-1 do begin
   glNewList(FDispListID+x, GL_COMPILE);

  //----------------------------------------------------------------------------
  // Nun das Objekt rendern damit die DisplayListe gefllt wird. Dannach das
  // rendern in die Display-Liste beenden, damit weitere Render-Aufrufe nicht
  // in diese Display-Liste geleitet werden.
  //----------------------------------------------------------------------------
   Render(x);
   glEndList;
  end;
end;


////////////////////////////////////////////////////////////////////////////////
// ReleaseDisplayList
//------------------------------------------------------------------------------
// Falls eine DisplayListe erstellt wurde, diese freigeben.
//------------------------------------------------------------------------------
// Parameter: --- Keine ---
// Rckgabe : --- Keine ---
procedure TAC3DFile.ReleaseDisplayList;
begin
  if FDispListID <> 0 then
    glDeleteLists(FDispListID, FObject.NumKids);
  FDispListID := 0;
end;


////////////////////////////////////////////////////////////////////////////////
// RenderDisplayList                                               
//------------------------------------------------------------------------------
// Rendert die erstellte Display-Liste.
//------------------------------------------------------------------------------
// Parameter: --- Keine ---
// Rckgabe : --- Keine ---
procedure TAC3DFile.RenderDisplayList;
var x:cardinal;
begin
 if (TOzoneMainClass(owner).gfxSys.renderPass<>0) then exit;
 glPushMatrix;
 glMultMatrixf(@FObject.matr);
 for x:=0 to FObject.NumKids-1 do begin
  with FObject.kids[x]^ do begin
   glPushMatrix;
   glMultMatrixf(@matr);
   glTranslated(Loc.x,Loc.y,Loc.z);
   glRotated(Rotate[0],1,0,0);
   glRotated(Rotate[1],0,1,0);
   glRotated(Rotate[2],0,0,1);
   glTranslated(-Loc.x,-Loc.y,-Loc.z);
   glTranslated(Trans[0],Trans[1],Trans[2]);
   glScaled(Scale[0],Scale[1],Scale[2]);
   glCallList(FDispListID+x);
   glPopMatrix;
  end;
 end;
 glPopMatrix;
end;


////////////////////////////////////////////////////////////////////////////////
// acEnable                                                           26.05.2003
//------------------------------------------------------------------------------
// Aktiviert den in FlagID bergebenen RenderState.
//------------------------------------------------------------------------------
// Parameter: FlagID -> Der zu aktivierende Render-State.
// Rckgabe : --- Keine ---
procedure TAC3DFile.acEnable(FlagID: TACRenderStateEnum);
begin
  //----------------------------------------------------------------------------
  // Einfach ermitteln, welcher Render-State gesetzt werden soll und eben dies
  // tun.
  //----------------------------------------------------------------------------
  case FlagID of
    AC_USE_TEXTURE : FUseTexture := True;
    AC_USE_LIGHT   : FUseLight := True;
    AC_SOLIDDRAW   : FPolygonMode := GL_FILL;
    AC_WIREFRAME   : FPolygonMode := GL_LINE;
    AC_POINTS      : FPolygonMode := GL_POINT;
    AC_SAVESTATES  : FSaveStates := True;
  end;
end;


////////////////////////////////////////////////////////////////////////////////
// acDisable                                                          26.05.2003
//------------------------------------------------------------------------------
// Deaktiviert den in FlagID bergebenen Render-State.
//------------------------------------------------------------------------------
// Parameter: FlagID -> Die ID des zu deaktivierenden Render-States.
// Rckgabe : --- Keine ---
procedure TAC3DFile.acDisable(FlagID: TACRenderStateEnum);
begin
  //----------------------------------------------------------------------------
  // Ermitteln, welcher Render-State deaktiviert werden soll und eben dies tun.
  // Wenn der WireFrame bzw. Point-Modus abgeschaltet werden soll, muss die
  // Polygon-Darstellung auf GL_FILL gesetzt werden.
  //----------------------------------------------------------------------------
  case FlagID of
    AC_USE_TEXTURE : FUseTexture := False;
    AC_USE_LIGHT   : FUseLight := False;
    AC_SAVESTATES  : FSaveStates := False;
    AC_WIREFRAME,
    AC_POINTS      : FPolygonMode := GL_FILL;
  end;
end;

end.


  
