library control_mouse;

uses
 {$IFDEF WIN32}
  ShareMem,
 {$ENDIF}
  SysUtils,
  Classes,
  o3control,
  dglOpenGL,
  o3actors,
  o3vectors,
  o3game,
  o3interfaces,
  Math;

{$LIBPREFIX ''}

type TMouseControl = class;
     TControlListener = class(TO3InputListener)
      owner:TMouseControl;
      procedure keyDown(key:TO3Key); override;
      procedure keyUp(key:TO3Key); override;
      procedure mouseDown(button:integer); override;
      procedure mouseUp(button:integer); override;
      procedure mouseMove(dx,dy:integer); override;
     end;
     TMouseControl = class(To3Control)
      MouseDeltaX,MouseDeltaY:integer;
      MouseSensivity:single;
      KeyboardOnly:boolean;
      keysPressed:array[TO3Key] of integer;   // 0 = released, 1 = pressed,
      buttonsPressed:array[0..10] of integer; //    2 = fresh pressed
      Listener:TControlListener;
      procedure MouseMove(dx,dy:integer);
      procedure MouseDown(button:integer);
      procedure MouseUp(button:integer);
      procedure KeyDown(key:TO3Key);
      procedure KeyUp(key:TO3Key);
      function Steer(ac:pointer;t:integer):TVector3d; override;
      constructor Create;
      destructor Destroy; override;
     end;
var Ozone:TOzoneMainClass;

procedure TControlListener.keyDown;
begin
 owner.keyDown(key);
end;

procedure TControlListener.keyUp;
begin
 owner.keyUp(key);
end;

procedure TControlListener.mouseDown;
begin
 owner.MouseDown(button);
end;

procedure TControlListener.mouseUp;
begin
 owner.MouseUp(button);
end;

procedure TControlListener.mouseMove;
begin
 owner.MouseMove(dx,dy);
end;

function MakeMouseControl:TO3Control;
begin
 Result:=TMouseControl.Create;
end;

procedure RegisterLib(entry:pointer;o3_state:pointer);
begin
 ozone:=o3_state;
 PControlEntry(entry).MakeControl:=MakeMouseControl;
end;

 procedure TMouseControl.MouseMove;
 begin
  inc(MouseDeltaX,dx);
  inc(MouseDeltaY,dy);
 end;

 procedure TMouseControl.MouseDown;
 begin
  buttonsPressed[button]:=2;
 end;

 procedure TMouseControl.MouseUp;
 begin
  buttonsPressed[button]:=0;
 end;

 procedure TMouseControl.KeyDown;
 begin
  keysPressed[key]:=2;
 end;

 procedure TMouseControl.KeyUp;
 begin
  keysPressed[key]:=0;
 end;

 constructor TMouseControl.Create;
 begin
  inherited Create;
  listener:=TControlListener.Create;
  listener.owner:=Self;
  ozone.InputSys.addListener(listener);
//  ozone.InputSys.addMouseMoveListener(MouseMove);
//  ozone.InputSys.addKeyDownListener(keyDown);
//  ozone.InputSys.addKeyUpListener(keyUp);
//  ozone.InputSys.addMouseDownListener(mouseDown);
//  ozone.InputSys.addMouseUpListener(mouseUp);
  MouseDeltaX:=0;
  MouseDeltaY:=0;
  MouseSensivity:=ozone.O3Ini.ReadFloat('control_mouse','sensivity',1);
  keyboardOnly:=ozone.O3Ini.ReadBool('control_mouse','keyboard_only',false);
 end;


 function TMouseControl.Steer;
 var s:integer;
     actor:TActor;
     c:boolean;
 const MinSens = 1;
       MaxSens = 30;
 begin
  actor:=TActor(ac);
  if actor=nil then Exit;
  if (keysPressed[O3K_LSHIFT]>0) or (keysPressed[O3K_RSHIFT]>0) then begin
   if (keysPressed[O3K_LEFT]>1) then begin
    actor.Look:=(actor.Look+1)mod LCount;
    keysPressed[O3K_LEFT]:=1;
   end;
   if (keysPressed[O3K_RIGHT]>1) then begin
    actor.Look:=(actor.Look+LCount-1)mod LCount;
    keysPressed[O3K_RIGHT]:=1;
   end;
  end;
  if keyboardOnly then begin
   mouseDeltaX:=0;
   mouseDeltaY:=0;
   if keysPressed[O3K_UP]>0 then
    mouseDeltaY:=-t;
   if keysPressed[O3K_DOWN]>0 then
    mouseDeltaY:=t;
   if keysPressed[O3K_LEFT]>0 then
    mouseDeltaX:=-t;
   if keysPressed[O3K_RIGHT]>0 then
    mouseDeltaX:=t;
  end;
  actor.canTakeUpItems:=true;
  if keysPressed[O3K_SPACE]>1 then begin
   keysPressed[O3K_SPACE]:=1;
   actor.LookFar:=not actor.LookFar; 
  end;
  if buttonsPressed[1]>0 then begin
   if buttonsPressed[1]>1 then
    actor.UseFirstItem;
   buttonsPressed[1]:=1;
   actor.canTakeUpItems:=false;
  end;

  if buttonsPressed[3]>0 then begin
   if buttonsPressed[3]>1 then
    actor.RotateItems;
   buttonsPressed[3]:=1;
   actor.canTakeUpItems:=false;
  end;
  c:=false;
  if (keysPressed[O3K_PAGEUP] > 1) then begin MouseSensivity:=mouseSensivity+1; c:=true; keysPressed[O3K_PAGEUP]:=1; end;
  if (keysPressed[O3K_PAGEDOWN] > 1) then begin MouseSensivity:=mouseSensivity-1; c:=true; keysPressed[O3K_PAGEDOWN]:=1; end;
  if MouseSensivity<MinSens then MouseSensivity:=MinSens;
  if MouseSensivity>MaxSens then MouseSensivity:=MaxSens;
  if c then
   actor.Inventary.StartStaticText('Sensivity: '+Copy(FloatToStr(mouseSensivity),1,3),1000);
  if (keysPressed[O3K_TAB]>1) then begin
   keysPressed[O3K_TAB]:=1;
   keyboardOnly:=not keyboardOnly;
   if keyboardOnly then
    actor.Inventary.StartStaticText('Keyboard only mode',1000)
   else
    actor.Inventary.StartStaticText('Mouse and Keyboard mode',1000);
  end;
  if L[actor.Look].SwapXY then begin
   s:=MouseDeltaX;
   MouseDeltaX:=MouseDeltaY;
   MouseDeltaY:=s;
  end;
   Result := getVector(
                       sign(MouseDeltaX)*power(abs(MouseDeltaX)*0.4,1.2)*
                       L[actor.Look].XS*MouseSensivity,
                       0,
                       sign(MouseDeltaY)*power(abs(MouseDeltaY)*0.4,1.2)*
                       L[actor.Look].YS*MouseSensivity);
  MouseDeltaX:=0;
  MouseDeltaY:=0;
 end;


 destructor TMouseControl.Destroy;
 begin
//  ozone.InputSys.deleteMouseMoveListener(MouseMove);
//  ozone.InputSys.deleteKeyDownListener(keyDown);
//  ozone.InputSys.deleteKeyUpListener(keyUp);
//  ozone.InputSys.deleteMouseDownListener(mouseDown);
//  ozone.InputSys.deleteMouseUpListener(mouseUp);
  ozone.InputSys.deleteListener(listener);
  listener.Free;
  ozone.O3Ini.WriteFloat('control_mouse','sensivity',MouseSensivity);
  ozone.O3Ini.WriteBool('control_mouse','keyboard_only',keyboardOnly);
  inherited Destroy;
 end;

exports RegisterLib;
begin
end.
