library item_wormhole;

uses
 {$IFDEF WIN32}
  ShareMem,
 {$ENDIF}
  SysUtils,
  Classes,
  o3item,
  ac3d,
  o3utils,
  o3inventary,
  o3opengl,
  StrUtils,
  o3actors,
  dglOpenGL,
  o3vectors,
  o3physics,
  o3world,
  odeimport,
  o3particles,
  o3game,
  o3m;

{$LIBPREFIX 'lib'}

type
 Twormhole = class(TItem)
  targetx,
  targety,
  targetz,
  force,
  range:single;
  active:boolean;
  TheForceField:TForceField;
  rotation:single;
  constructor Create(s:string;own:Pointer); override;
  procedure Render(deltaTime:integer); override;
  destructor Destroy; override;
  procedure OnActorOver(actor:pointer;level:Pointer;x,y,z:integer); override;
  procedure UpdateGeomPos; override;
  procedure processParameterList(p:TParameterList); override;
  function getParameterList():TParameterList; override;
 end;

 var graph:TO3mFile;
     p:string;
     ozone:TOzoneMainClass;

procedure TWormhole.processParameterList(p:TParameterList);
begin
 info['cmd']:=GetToken(info['cmd'],1,':')+':'+floatToStr(p[0].posValue[0])+'$'+floatToStr(p[1].heightValue)+'$'+floatToStr(p[0].posValue[1])+'$'+floatToStr(p[2].floatValue)+'$'+floatToStr(p[3].floatValue);
 if p[4].boolValue then
  info['cmd']:=info['cmd']+'$on';
end;

function TWormhole.getParameterList:TParameterList;
var t:TObjParameter;
    p:TPosParam;
begin
 Result:=TParameterList.Create;
 
 p[0]:=targetx;
 p[1]:=targetz;
 t:=TObjParameter.Create;
 t.posValue:=p;
 t.desc:='Destination Position';
 Result.Add(t);
 
 t:=TObjParameter.Create;
 t.HeightValue:=targety;
 t.desc:='Destination Height';
 Result.Add(t);

 t:=TObjParameter.Create;
 t.FloatValue:=force;
 t.desc:='Force';
 Result.Add(t);

 t:=TObjParameter.Create;
 t.FloatValue:=range;
 t.desc:='Range';
 Result.Add(t);

 t:=TObjParameter.Create;
 t.BoolValue:=active;
 t.desc:='Initially Active';
 Result.Add(t);
end;

function GetFieldForce(dist:TVector3d;Data:Pointer;Actor:TActor):single;
var l:single;
    t:Tdvector3;
begin
 result:=0;
 if not TWormHole(Data).active then exit;
 l:=len(dist,1,1,1);
 if l=0 then l:=1E-50;
 if l>TWormHole(Data).Range then exit;
 if l>0.5 then
  Result:=TWormHole(Data).force/(l*l)
 else begin
  odeimport.dWorldGetGravity(TLevel(actor.owner).Physics.theWorld,t);
  odeimport.dBodySetForce(actor.body,-t[0],-t[1],-t[2]);
  Result:=30;
 end;
end;

procedure TWormhole.UpdateGeomPos;
begin
 inherited;
 TheForceField.cx:=cx;
 TheForceField.cy:=cy;
 TheForceField.cz:=cz;
end;

procedure WormCallBack(P:PParticle;t:integer);
var acp,dist:TVector3d;
begin
 if (p.data=nil)or(TGameObject(p.data)=nil) then EXIT;
 acp:=getVector(TGameObject(p.data).cx,TGameObject(p.data).cy,TGameObject(p.data).cz);
 dist:=sum(acp,mul(p.pos,-1));
 p.speed:=sum(p.speed,mul(norm(dist),t/500));
end;

procedure WormCallBack2(P:PParticle;t:integer);
var acp,dist:TVector3d;
begin
 if (p.data=nil)or(TGameObject(p.data)=nil) then EXIT;
 acp:=getVector(TGameObject(p.data).cx,TGameObject(p.data).cy,TGameObject(p.data).cz);
 dist:=sum(acp,mul(p.pos,-1));
 p.speed:=sum(p.speed,mul(norm(dist),t/100));
end;

procedure TWormhole.OnActorOver;
var Dist:TVector3d;
const WarpDist  = 0.5;
      WarpDistY = 0.2;
begin
 if active then begin
  Dist:=getVector(cx-TActor(actor).cx,cy-TActor(actor).cy,cz-TActor(actor).cz);
  if (abs(cy-TActor(actor).cy)<=WarpDistY)and(Len(dist,1,0,1)<=WarpDist)and(TActor(actor).physics.godFor<=0) then begin
   TActor(actor).x:=targetx;
   TActor(actor).y:=targety;
   TActor(actor).z:=targetz;
   ozone.soundLib.playSound('wormhole_warp',cx,cy,cz);
//   theParticleSystem.spawn(100,2,getVector(1,0,0),getVector(1,1,1),getVector(contact[i].geom.pos[0],contact[i].geom.pos[1],contact[i].geom.pos[2]),1000,100);
//   TActor(actor).spawnParticlesOwnColor(600,2,getVector(0,0,1),getVector(1,1,1),getVector(cx,cy,cz),2000,50,-1,WormCallback2,actor);
   TActor(actor).spawn;
    //   TLevel(owner).particleSystem.spawn(100,2,getVector(1,0,0),getVector(1,1,1),getVector(cx,cy,cz),1000,100);
//   TLevel(owner).particleSystem.spawn(100,2,getVector(1,0,0),getVector(1,1,1),getVector(targetx,targety,targetz),1000,100);
   if TActor(actor).body<>nil then
    dBodySetLinearVel(TActor(actor).body,0,0,0);
  end;
 end;
end;

procedure Twormhole.Render;
begin
 graph.FObject.kids[0].Rotate[1]:=rotation;
 rotation:=rotation-(deltaTime*force)/20;
 if rotation<-1E10 then rotation:=0;
 if (ozone<>nil)and(active) then
  inherited Render(deltaTime);
  //ozone.gfxSys.DrawModel(graph,0,0,0,o3utils.ozone);
 graph.FObject.kids[0].Rotate[1]:=0;
end;

constructor Twormhole.Create;
var t,n,tx,ty,tz,f,r,o:string;
begin
 rotation:=0;
 targetx:=0;
 targety:=0;
 targetz:=0;
 active:=false;
 force:=0;
 range:=0;
 n:=GetToken(s,1,':');
 t:=GetToken(s,2,':');

 tx:=GetToken(t,1,'$');
 ty:=GetToken(t,2,'$');
 tz:=GetToken(t,3,'$');
 f:=GetToken(t,4,'$');
 r:=GetToken(t,5,'$');
 o:=GetToken(t,6,'$');
 inherited Create(s,own);
 gfx:=graph;
 gfxtype:=o3utils.ozone;
 DecimalSeparator:='.';
 targetx:=StrToFloat(tx);
 targety:=StrToFloat(ty);
 targetz:=StrToFloat(tz);
 force:=StrToFloat(f);
 range:=StrToFloat(r);
 if o='on' then
  active:=true;
 TheForceField.cx:=cx;
 TheForceField.cy:=cy;
 TheForceField.cz:=cz;
 TheForceField.getForce:=getFieldForce;
 TheForceField.Data:=self;
 TLevel(Owner).Physics.ForceFields.Add(@TheForceField)
end;

destructor Twormhole.Destroy;
begin
 TLevel(owner).Physics.ForceFields.Remove(@TheForceField);
 inherited Destroy;
end;

function MakeItem(s:string;owner:Pointer):TItem;
begin
 Result:=Twormhole.Create(s,owner);
end;

procedure InitLib(path:string;owner:Pointer);
begin
 ozone:=owner;
 p:=path;
 graph:=TO3mFile.Create(owner);
 graph.LoadFromFile(p+'it-wormhole.o3m');
end;

procedure QuitLib;
begin
 if graph<>nil then
  graph.Free;
end;

exports
 MakeItem,
 InitLib,
 QuitLib;
begin
end.

