library stone_lasers;

uses
  {$IFDEF WIN32}ShareMem,{$ENDIF}
  SysUtils,
  Classes,
  Math,
  ac3d,
  o3m,
  o3utils,
  o3stones,
  o3world,
  o3game,
  o3actors,
  o3vectors;

{$LIBPREFIX 'lib'}

type
    TDir = (dn,de,ds,dw);
    PDir = ^TDir;
    TLBeam = (bn,bs,bns,bew,bw,be);
    TMirrorType = (mtPlane,mtTriangle);
    TLaser = class(TStone)
     state:boolean;
     dir:TDir;
     procedure processParameterList(p:TParameterList); override;
     function getParameterList:TParameterList; override;
     procedure Render(timeDelta:integer); override;
     constructor create(s:string;own:pointer); reintroduce;
     procedure processMessage(m:string;data:array of pointer); override;
    end;
    TMirror = class(TStone)
     MType: TMirrorType;
     MOrient: integer;
     OOrient: integer;
     SFrame: single;
     doReflect: array[TDir] of boolean;
     procedure Render(timeDelta:integer); override;
     constructor create(s:string;own:pointer); reintroduce;
     procedure processMessage(m:string;data:array of pointer); override;
     procedure playSound(reason:string;data:array of pointer); override;
    end;
var laserGfx: array[TDir] of TO3mFile;
    laserBeams: array[TLBeam] of TO3mFile;
    mirrorGfx: array[TMirrorType] of TO3mFile;

 procedure TMirror.playSound;
 begin
  if reason='collision' then begin
  end
  else
   inherited playSound(reason,data);
 end;

 function nx(x:integer; dir:TDir):integer;
 begin
  Result:=x;
  case dir of
   dw:Result:=x-1;
   de:Result:=x+1;
  end;
 end;

 function nz(z:integer; dir:TDir):integer;
 begin
  Result:=z;
  case dir of
   ds:Result:=z+1;
   dn:Result:=z-1;
  end;
 end;


 procedure drawBeam(x,y,z:integer;dir:TDir;o3:TOzoneMainClass;half:boolean;inv:boolean=false);
 var t:TLBeam;
 begin
 t:=bns;
 if half then
  case dir of
   dn: if inv then t:=bn else t:=bs;
   ds: if inv then t:=bs else t:=bn;
   dw: if inv then t:=bw else t:=be;
   de: if inv then t:=be else t:=bw;
  end
 else
  case dir of
   dn,ds: t:=bns;
   dw,de: t:=bew;
  end;
  o3.gfxSys.DrawModel(laserBeams[t],x,y,z,ozone,nil,true);
 end;

 procedure castBeam(sx,sy,sz:integer;dir:TDir;level:TLevel);
 var tx,tz,i:integer;
     cell:TCell;
     answ:boolean;
     ph:boolean;
     invDir:Tdir;
     delta: single;
 begin
  if(TOzoneMainClass(level.owner).gfxSys.renderPass<>0) then EXIT;
  tx:=nx(sx,dir);
  tz:=nz(sz,dir);
  case dir of
   dn: invDir:=ds;
   ds: invDir:=dn;
   dw: invDir:=de;
   de: invDir:=dw;
  end;
  if (tx<0)or(tz<0)or(tx>=level.FX)or(tz>=level.FZ) then EXIT;
  cell:=level.getCell(tx,sy,tz);
  drawBeam(sx,sy,sz,dir,TOzoneMainClass(level.owner),true,true);
  if(cell=nil) then begin
   drawBeam(tx,sy,tz,dir,TOzoneMainClass(level.owner),false);
   castBeam(tx,sy,tz,dir,level);
   EXIT;
  end;
  if(cell.FStone<>nil) then begin
   answ:=false;
   ph:=false;
   cell.FStone.processMessage('laserhit',[@invdir,@answ,@ph]);
   if (cell.FStone.Info['laserpass']='yes')or answ then begin
    drawBeam(tx,sy,tz,dir,TOzoneMainClass(level.owner),false);
    castBeam(tx,sy,tz,dir,level);
   end
   else
    if ph then drawBeam(tx,sy,tz,dir,TOzoneMainClass(level.owner),true);
  end
  else if(cell.FItem<>nil) then begin
   answ:=false;
   ph:=true;
   cell.FItem.processMessage('laserhit',[@invdir,@answ,@ph]);
   if (cell.FItem.Info['laserpass']='yes')or answ then begin
    drawBeam(tx,sy,tz,dir,TOzoneMainClass(level.owner),false);
    castBeam(tx,sy,tz,dir,level);
   end
   else
    if ph then drawBeam(tx,sy,tz,dir,TOzoneMainClass(level.owner),true);
  end
  else begin
   answ:=true;
   delta:=0.5;
   if level.FActors.List.Count>0 then
    for i:=0 to level.FActors.List.Count-1 do
     if level.FActors.List[i]<>nil then
      if (floor(TActor(level.FActors.List[i]).x)=tx) and
         (floor(TActor(level.FActors.List[i]).y)=sy) and
         (floor(TActor(level.FActors.List[i]).z)=tz) then
         begin
          if (dir=dn)or(dir=ds) then delta:=TActor(level.FActors.List[i]).z-tz;
          if (dir=dw)or(dir=de) then delta:=TActor(level.FActors.List[i]).x-tx;
          answ:=false;
          TActor(level.FActors.List[i]).processMessage('laserhit',[@dir,@answ]);
          break;
         end;

   if answ then begin
    drawBeam(tx,sy,tz,dir,TOzoneMainClass(level.owner),false);
    castBeam(tx,sy,tz,dir,level);
   end
   else begin
    if delta<0.25 then begin end
    else if delta<0.8 then drawBeam(tx,sy,tz,dir,TOzoneMainClass(level.owner),true)
    else drawBeam(tx,sy,tz,dir,TOzoneMainClass(level.owner),false)
   end;
  end;
 end;

 function CollFuncMirror(dir:integer;call:TStone;x,y,z:integer;actor:Pointer;level:Pointer;geom:pointer):boolean;
 begin
  Result:=true;
  if dir=0 then exit;
  if (TMirror(call).sFrame>0)and(TMirror(call).sFrame<=0.5) then exit;
  TMirror(call).OOrient:=TMirror(call).MOrient;
  TMirror(call).SFrame:=0.0001;
  TMirror(call).MOrient:=(TMirror(call).MOrient+3)mod 4;
  TOzoneMainClass(TLevel(call.owner).owner).soundLib.playSound('mirrorturn',call.cx,call.cy,call.cz);
 end;

 procedure TMirror.processMessage;
 var d:PDir;
     h:PBoolean;
 begin
  if (m='signal') then begin
   OOrient:=MOrient;
   SFrame:=0;
   MOrient:=(MOrient+3)mod 4;
  end;
  if (m='laserhit') then begin
   if MOrient<>OOrient then EXIT;
   d:=data[0];
   h:=data[2];
   if doReflect[d^] then doReflect[d^]:=false else EXIT; // Avoid endless loops, so only reflect once per time step
   h^:=true;
   if mType=mtPlane then begin
   case MOrient of
     0:
      case d^ of
       ds: castBeam(x,y,z,dn,owner);
       dn: castBeam(x,y,z,dn,owner);
       else h^:=false;
      end;
     1:
      case d^ of
       de: castBeam(x,y,z,ds,owner);
       dn: castBeam(x,y,z,dw,owner);
       ds: castBeam(x,y,z,de,owner);
       dw: castBeam(x,y,z,dn,owner);
      end;
     2:
      case d^ of
       de: castBeam(x,y,z,de,owner);
       dw: castBeam(x,y,z,dw,owner);
       else h^:=false;
      end;
     3:
      case d^ of
       de: castBeam(x,y,z,dn,owner);
       dw: castBeam(x,y,z,ds,owner);
       dn: castBeam(x,y,z,de,owner);
       ds: castBeam(x,y,z,dw,owner);
      end;
    end;
   end;
   if mType=mtTriangle then begin
    case MOrient of
     0: // v
      case d^ of
       ds: begin castBeam(x,y,z,dw,owner); castBeam(x,y,z,de,owner); end;
       de: castBeam(x,y,z,ds,owner);
       dn: castBeam(x,y,z,dn,owner);
       dw: castBeam(x,y,z,ds,owner);
      end;
     1: // >
      case d^ of
       de: begin castBeam(x,y,z,ds,owner); castBeam(x,y,z,dn,owner); end;
       dn: castBeam(x,y,z,de,owner);
       ds: castBeam(x,y,z,de,owner);
       dw: castBeam(x,y,z,dw,owner);
      end;
     2: // ^
      case d^ of
       dn: begin castBeam(x,y,z,dw,owner); castBeam(x,y,z,de,owner); end;
       de: castBeam(x,y,z,dn,owner);
       ds: castBeam(x,y,z,ds,owner);
       dw: castBeam(x,y,z,dn,owner);
      end;
     3: // <
      case d^ of
       dw: begin castBeam(x,y,z,dn,owner); castBeam(x,y,z,ds,owner); end;
       de: castBeam(x,y,z,de,owner);
       ds: castBeam(x,y,z,dw,owner);
       dn: castBeam(x,y,z,dw,owner);
      end;
    end;
   end;
  end;
 end;

 constructor TMirror.Create;
 var t:string;
 begin
  inherited Create(s,'block',own);
  t:=GetToken(s,1,':');
  MType:=mtPlane;
  if t[Length(t)]='3' then
   MType:=mtTriangle;
  MOrient:=0;
  t:=GetToken(s,2,':');
  if t<>'' then
   MOrient:=StrToInt(t) mod 4;
  OOrient:=MOrient;
  collisionCallback:=CollFuncMirror;
  gfx:=mirrorGfx[MType];
  gfxtype:=o3utils.ozone;
  shadow:=TOzoneMainClass(TLevel(owner).owner).shadowLib['sh-stone'];
 end;

 procedure TMirror.Render;
 var mo:integer;
 begin
  doReflect[dn]:=true;
  doReflect[de]:=true;
  doReflect[dw]:=true;
  doReflect[ds]:=true;
  if OOrient = MOrient then begin
   if MType=mtPlane then
    To3mFile(gfx).FObject.Kids[0].Rotate:=getVector(0,MOrient*45,0)
   else
    To3mFile(gfx).FObject.Kids[0].Rotate:=getVector(0,MOrient*90,0);
  end
  else begin
   mo:=OOrient;
   if mo=0 then inc(mo,4);
   if MType=mtPlane then
    To3mFile(gfx).FObject.Kids[0].Rotate:=getVector(0,((1-SFrame)*mo+SFrame*MOrient)*45,0)
   else
    To3mFile(gfx).FObject.Kids[0].Rotate:=getVector(0,((1-SFrame)*mo+SFrame*MOrient)*90,0);
   SFrame:=SFrame+timeDelta/200;
   if SFrame>=1 then begin
    SFrame:=0;
    OOrient:=MOrient;
   end;
  end;
  inherited Render(timeDelta);
  //TOzoneMainClass(TLevel(owner).owner).gfxSys.DrawModel(gfx,0,0,0,ozone);
  To3mFile(gfx).FObject.Kids[0].Rotate:=getVector(0,0,0);
 end;


 procedure TLaser.processMessage;
 begin
  if (m='signal') then begin
   state:=not state;
   if state then
    TOzoneMainClass(TLevel(owner).owner).soundLib.playSound('laser',cx,cy,cz);
  end;
 end;

 procedure TLaser.processParameterList(p:TParameterList);
 begin
  if p[0].boolValue then
   info['cmd']:=GetToken(info['cmd'],1,':')+':on'
  else
   info['cmd']:=GetToken(info['cmd'],1,':');
 end;

 function TLaser.getParameterList:TParameterList;
 var t:TObjParameter;
 begin
  Result:=TParameterList.Create;
  t:=TObjParameter.Create;
  t.desc:='Initially On';
  t.boolValue:=state;
  Result.add(t);
 end;
 
 constructor TLaser.create;
 var t:string;
 begin
  inherited Create(s,'block',own);
  dir:=dn;
  state:=false;
  t:=GetToken(s,1,':');
  case t[Length(t)] of
   'n': dir:=dn;
   's': dir:=ds;
   'e': dir:=de;
   'w': dir:=dw;
  end;
  t:=GetToken(s,2,':');
  if t='on' then state:=true;
  gfx:=laserGfx[dir];
  gfxtype:=o3utils.ozone;
  shadow:=TOzoneMainClass(TLevel(owner).owner).shadowLib['sh-laser'];
  if state then
   TOzoneMainClass(TLevel(owner).owner).soundLib.playSound('laser',cx,cy,cz);
  end;

 procedure TLaser.Render;
 begin
//  TOzoneMainClass(TLevel(owner).owner).gfxSys.DrawModel(gfx,0,0,0,ozone);
  inherited Render(timeDelta);
  if state then
   if visible then
   castBeam(x,y,z,dir,owner);
 end;


  function MakeStone(s:string;o:Pointer):TStone;
  begin
   Result:=nil;
   if Copy(s,1,8)='st-laser' then
    Result:=TLaser.create(s,o);
   if Copy(s,1,9)='st-mirror' then
    Result:=TMirror.create(s,o);
  end;

  procedure InitLib(path:string;owner:Pointer);
  begin

   laserGfx[dn]:=TO3mFile.Create(owner);
   laserGfx[dn].LoadFromFile(path+'st-laser-n.o3m');
   laserGfx[ds]:=TO3mFile.Create(owner);
   laserGfx[ds].LoadFromFile(path+'st-laser-s.o3m');
   laserGfx[de]:=TO3mFile.Create(owner);
   laserGfx[de].LoadFromFile(path+'st-laser-e.o3m');
   laserGfx[dw]:=TO3mFile.Create(owner);
   laserGfx[dw].LoadFromFile(path+'st-laser-w.o3m');

   laserBeams[bn]:=TO3mFile.Create(owner);
   laserBeams[bn].LoadFromFile(path+'laserbeam-n.o3m');
   laserBeams[bs]:=TO3mFile.Create(owner);
   laserBeams[bs].LoadFromFile(path+'laserbeam-s.o3m');
   laserBeams[bw]:=TO3mFile.Create(owner);
   laserBeams[bw].LoadFromFile(path+'laserbeam-w.o3m');
   laserBeams[be]:=TO3mFile.Create(owner);
   laserBeams[be].LoadFromFile(path+'laserbeam-e.o3m');
   laserBeams[bew]:=TO3mFile.Create(owner);
   laserBeams[bew].LoadFromFile(path+'laserbeam-ew.o3m');
   laserBeams[bns]:=TO3mFile.Create(owner);
   laserBeams[bns].LoadFromFile(path+'laserbeam-ns.o3m');
   
   mirrorGfx[mtPlane]:=TO3mFile.Create(owner);
   mirrorGfx[mtPlane].LoadFromFile(path+'st-mirror1.o3m');
   mirrorGfx[mtTriangle]:=TO3mFile.Create(owner);
   mirrorGfx[mtTriangle].LoadFromFile(path+'st-mirror2.o3m');


  end;

  procedure QuitLib;
  var b1:TLBeam;
      b2:TDir;
  begin
   for b1:=bn to be do
    if laserBeams[b1]<>nil then
     laserBeams[b1].Free;
   for b2:=dn to dw do
    if laserGfx[b2]<>nil then
     laserGfx[b2].Free;
   if mirrorGfx[mtPlane]<>nil then mirrorGfx[mtPlane].Free;
   if mirrorGfx[mtTriangle]<>nil then mirrorGfx[mtTriangle].Free;
  end;


  exports InitLib,
          QuitLib,
          MakeStone;
begin
end.
