unit oooal;

interface

uses openal, sysutils;

type
        TAlSource = class
         private
          _pos: array [0..2] of TALFloat;
          _vel: array [0..2] of TALFloat;
          _id: TALuint;
          _pitch: TALFloat;
          _gain: TALFloat;
          _loop: TALInt;
          function is_playing:boolean;
         public
          constructor Create(buffer:TALuint);
          procedure Play(x,y,z:single;xv,yv,zv:single);
          procedure Stop;
          procedure Pause;
          procedure Update;
          destructor Destroy; override;
          property xpos: TALFloat read _pos[0] write _pos[0];
          property ypos: TALFloat read _pos[1] write _pos[1];
          property zpos: TALFloat read _pos[2] write _pos[2];
          property xvel: TALFloat read _vel[0] write _vel[0];
          property yvel: TALFloat read _vel[1] write _vel[1];
          property zvel: TALFloat read _vel[2] write _vel[2];
          property pitch: TALFloat read _pitch write _pitch;
          property gain: TALFloat read _gain write _gain;
          property loop: TALInt read _loop write _loop;
          property playing: boolean read is_playing;
        end;
        TAlObject = class
        private
          _buffer: TALuint;
          _sources: array of TAlSource;
        protected
        public
          constructor Create;
          destructor Destroy; override;
          procedure LoadFromFile(filename: string);
          function Play(x,y,z:single;xv,yv,zv:single):TALSource;
        end;

implementation

constructor TALSource.Create(buffer:TALuint);
begin
  AlGenSources(1, @_id);
//    if alGetError <> AL_NO_ERROR then raise Exception.Create('Cannot create Source');
  AlSourcei ( _id, AL_BUFFER, buffer);
  gain := 1.0;
  pitch := 1.0;
  xpos := 0.0;
  ypos := 0.0;
  zpos := 0.0;
  xvel := 0.0;
  yvel := 0.0;
  zvel := 0.0;
end;

procedure TALSource.Play(x,y,z:single;xv,yv,zv:single);
begin
 xpos:=x;
 ypos:=y;
 zpos:=z;
 xvel:=xv;
 yvel:=yv;
 zvel:=zv;
 AlSourcePlay(_id);
 Update;
end;

procedure TalSource.Update;
begin
 //pass the 'changed' values on to openal
 AlSourcef ( _id, AL_PITCH, _pitch );
 AlSourcef ( _id, AL_GAIN, _gain );
 AlSourcefv ( _id, AL_POSITION, @_pos);
 AlSourcefv ( _id, AL_VELOCITY, @_vel);
 AlSourcei ( _id, AL_LOOPING, _loop);
end;

procedure TALSource.Stop;
begin
 AlSourceStop(_id);
end;

procedure TALSource.Pause;
begin
 AlSourcePause(_id);
end;

function TALSource.is_playing:boolean;
var t:TALint;
begin
 alGetSourcei(_id,AL_SOURCE_STATE,@t);
 Result:=(t=AL_PLAYING);
end;

destructor TALSource.Destroy;
begin
  if playing then stop;
  AlDeleteSources(1, @_id);
//  if alGetError <> AL_NO_ERROR then raise Exception.Create('Cannot delete Source');

end;

constructor TalObject.Create;
begin
  setLength(_sources,0);
  AlGenBuffers(1, @_buffer);
//  if alGetError <> AL_NO_ERROR then raise Exception.Create('Cannot create Buffer');
end;

destructor TalObject.Destroy;
var x:integer;
begin
  if Length(_sources)>0 then
   for x:=0 to Length(_sources)-1 do
    if _sources[x]<>nil then _sources[x].Free;
  setLength(_sources,0);
  AlDeleteBuffers(1, @_buffer);
  //if alGetError <> AL_NO_ERROR then raise Exception.Create('Cannot delete Buffer');

  inherited destroy;
end;

procedure TalObject.LoadFromFile(filename: string);
var
  format: TALEnum;
  size: TALSizei;
  freq: TALSizei;
  floop: TALInt;
  data: TALVoid;
begin
  //load the wavedata from the file
  AlutLoadWavFile(filename, format, data, size, freq, floop);
  //assign the wavedata to the buffer
  AlBufferData(_buffer, format, data, size, freq);
  if alGetError <> AL_NO_ERROR then raise Exception.Create('Cannot assign wave data to buffer');
  //remove the wavedata from memory
  AlutUnloadWav(format, data, size, freq);
end;

function TalObject.Play;
var i,s:integer;
begin
 s:=-1;
 if Length(_sources)=0 then begin
  SetLength(_sources,1);
  _sources[0]:=TALSource.Create(_buffer);
  s:=0;
 end
 else begin
  for i:=0 to Length(_sources)-1 do begin
   if not _sources[i].playing then begin
    s:=i;
    break;
   end;
  end;
  if (s=-1)and(Length(_sources)<10) then begin
   s:=Length(_sources);
   setLength(_sources,s+1);
   _sources[s]:=TALSource.Create(_buffer);
  end;
 end;
 if s<>-1 then begin
  _sources[s].Play(x,y,z,xv,yv,zv);
  Result:=_sources[s];
 end
 else
  Result:=nil;
end;

end.
