unit o3game;

interface
uses o3io_sdlogl,
     o3world,
     o3objects,
     o3control,
     o3opengl,
     o3specialstones,
     o3specialitems,
     o3actorinterface,
     o3particles,
     Logger,
     SysUtils,
     o3interfaces,
     IniFiles,
     o3sndlib,
     o3menu,
     o3shadowlib;
type TStartResult = (won, lost, aborted, restart);
     TOzoneMainClass = class;
     TOzoneInGameListener = class(TO3InputListener)
      owner:TOzoneMainClass;
      procedure keyDown(key:TO3Key); override;
     end;
     TOzoneMainClass = class
     public
      path:string;
      O3Ini:TIniFile;
      O3Log:TLogger;
      done : boolean;
      MDone : boolean;
      gfxSys:TO3GfxSystem;
      soundSys:TO3SoundSystem;
      soundLib:TSoundLib;
      invSys:TO3InvSystem;
      particleSysGfx:TO3ParticleSystemGfx;
      particleSystem:TParticleSystem;
      Objects:TObjectList;
      IOSys:TO3IOSystem;
      InputSys:TO3InputSystem;
      controls:TControlList;
      SpecialStones:TSpecialStoneList;
      SpecialItems:TSpecialItemList;
      SpecialActors:TSpecialActorList;
      theLevel:TLevel;
      VidW:integer;
      VidH:integer;
      vsplit:integer;
      FVolMusic:integer;
      FVolSound:integer;
      ActMenu:TO3Menu;
      MainMenu:TO3Menu;
      ShadowLib:TShadowLib;
      procedure volMusic(const vol:integer);
      procedure volSound(const vol:integer);
      property musicVolume:integer read FVolMusic write volMusic;
      property effectVolume:integer read FVolSound write volSound;
      function StartLevelPack(lp:string;slevel:integer):integer;
      function StartLevelOxyd(loader:string;level:string;args:string):TStartResult;
      procedure LoadTiles(frompath:string);
      procedure FreeTiles;
      destructor Destroy; override;
      constructor Create(s:string);
      procedure ShowMainMenu;
      procedure MKeyDown(sender:TO3menu;key:TO3Key);
      procedure OnePlayerStart(sender:TO3Control);
      procedure StartEditor(sender:TO3Control);
      procedure TwoPlayerStart(sender:TO3Control);
      procedure ExitMenu(sender:TO3Control);
      procedure CreateMainMenu;
     end;

implementation
uses o3utils,dglOpenGL,o3sound,o3opengl_font,o3inventary,ac3d,o3levelmenu,o3editor;

procedure TOzoneMainClass.OnePlayerStart;
begin
 ActMenu.Active:=false;
 ActMenu:=TO3LevelMenu.CreateFromLevelPack(GetLevelPack(O3Ini.ReadString('game','levelpack','test')),self);
 ActMenu.OnKeyDown:=MKeyDown;
// StartLevelPack(GetLevelPack(O3Ini.ReadString('game','levelpack','test')));
 ActMenu.Active:=true;
// ActMenu:=OA;
end;

procedure TOzoneMainClass.TwoPlayerStart;
begin
 ActMenu.Active:=false;
 ActMenu:=TO3LevelMenu.CreateFromLevelPack(GetLevelPack(O3Ini.ReadString('game','battlegroundpack','battlegrounds')),self);
 ActMenu.OnKeyDown:=MKeyDown;
// StartLevelPack(GetLevelPack(O3Ini.ReadString('game','battlegroundpack','battlegrounds')),0);
 ActMenu.Active:=true;
// ActMenu:=OA;
end;

procedure TOzoneMainClass.StartEditor;
var editor:TO3editorMainMenu;
begin
 ActMenu.Active:=false;
 editor:=TO3EditorMainMenu.Create(self);
 editor.run;
 editor.Free;
 ActMenu.Active:=true;
end;

procedure TOzoneMainClass.ExitMenu;
begin
 MKeyDown(ActMenu,O3K_ESCAPE);
end;


procedure TOzoneInGameListener.keyDown;
begin
 if key=O3K_ESCAPE then owner.done:=true;
 if key=O3K_F12 then
 if owner.theLevel<>nil then
  owner.theLevel.state:=LevWon;
 if key=O3K_F3 then
 if owner.theLevel<>nil then begin
  owner.theLevel.state:=LevRestart;
  owner.theLevel.theActorToFollow.physics.godFor:=0;
  owner.theLevel.theActorToFollow.alive:=false;
  if owner.theLevel.theActorToFollow2<>nil then begin
   owner.theLevel.theActorToFollow2.physics.godFor:=0;
   owner.theLevel.theActorToFollow2.alive:=false;
  end;
  if owner.theLevel.state <> LevLost then
   owner.theLevel.state:=LevRestart;
 end;
end;

procedure TOzoneMainClass.CreateMainMenu;
var Tbutton:TO3CustomButton;
begin
  MainMenu:=TO3Menu.Create(Self);

  Tbutton:=TO3CustomButton.Create(gfxSys,'1player');
  TButton.bgModel:=gfxSys.CreatePlaneXYModel(0.5,0.5,'button_1pstart.png',BitmapSubDir);//TAC3dFile.Create(self);
  TButton.bgMtype:=plane;
  TButton.x:=1;
  TButton.y:=0.7;
  TButton.ToolTip:='Start a one player game';
  TButton.OnClick:=OnePlayerStart;
  MainMenu.AddControl(TButton);

  Tbutton:=TO3CustomButton.Create(gfxSys,'editor');
  TButton.bgModel:=gfxSys.CreatePlaneXYModel(0.5,0.5,'button_editor.png',BitmapSubDir);//TAC3dFile.Create(self);
  TButton.bgMtype:=plane;
  TButton.x:=0;
  TButton.y:=0.7;
  TButton.ToolTip:='Enter the Level Editor';
  TButton.OnClick:=StartEditor;
  MainMenu.AddControl(TButton);

  Tbutton:=TO3CustomButton.Create(gfxSys,'2player');
  TButton.bgModel:=gfxSys.CreatePlaneXYModel(0.5,0.5,'button_2pstart.png',BitmapSubDir);//TAC3dFile.Create(self);
  TButton.bgMtype:=plane;
  TButton.x:=-1;
  TButton.y:=0.7;
  TButton.ToolTip:='Start a two player battle';
  TButton.OnClick:=TwoPlayerStart;
  MainMenu.AddControl(TButton);

  Tbutton:=TO3CustomButton.Create(gfxSys,'exit');
  TButton.bgModel:=gfxSys.CreatePlaneXYModel(0.5,0.5,'button_exit.png',BitmapSubDir);//TAC3dFile.Create(self);
  TButton.bgMtype:=plane;
  TButton.x:=1;
  TButton.y:=-0.7;
  TButton.ToolTip:='Quit OZONE';
  TButton.OnClick:=ExitMenu;
//  TAC3dFile(TButton.bgModel).LoadFromFile(path+MenuModelSubDir+'button_exit.ac');
//  TAC3dFile(TButton.bgModel).acEnable(AC_SOLIDDRAW);
//  TAC3dFile(TButton.bgModel).acDisable(AC_USE_LIGHT);
  MainMenu.AddControl(TButton);

  Tbutton:=TO3CustomButton.Create(gfxSys,'switches');
  TButton.bgModel:=gfxSys.CreatePlaneXYModel(0.5,0.5,'button_switches.png',BitmapSubDir);//TAC3dFile.Create(self);
  TButton.bgMtype:=plane;
  TButton.x:=-1;
  TButton.y:=-0.7;
  TButton.ToolTip:='Go to the switches! (n/a yet!)';
//  TButton.OnClick:=ExitMenu;
//  TAC3dFile(TButton.bgModel).LoadFromFile(path+MenuModelSubDir+'button_switches.ac');
//  TAC3dFile(TButton.bgModel).acEnable(AC_WIREFRAME);
//  TAC3dFile(TButton.bgModel).acDisable(AC_USE_LIGHT);


  MainMenu.AddControl(TButton);
  MainMenu.OnKeyDown:=MKeyDown;
end;

constructor TOzoneMainClass.Create;
const Log = false;
begin
  path:=s;
  if Log then
   O3Log:=TDummyLogger.Create
  else
   O3Log:=TLogger.Create;
  vsplit:=1;
  O3Ini:=TIniFile.Create(path+'ozone.ini');
  VidW:=O3Ini.ReadInteger('video','width',640);
  VidH:=O3Ini.ReadInteger('video','height',480);
  IOSys:=TO3IOSystem_SDLOGL.Create(self,VidH,VidW,O3Ini.ReadInteger('video','bpp',16),O3Ini.ReadBool('video','fullscreen',false));
  shadowLib:=TShadowLib.Create(self,path+ModelSubDir);
  soundLib:=TSoundLib.Create(path+SoundSubDir+'index',SoundSys,self);
  controls:=TControlList.Create(path+ControlSubDir+'index',self);
  particleSystem:=particleSysGfx.NewParticleSystem;
  LoadTiles(path+ModelSubDir);
  CreateMainMenu;
end;

procedure TOzoneMainClass.volMusic;
begin
 FVolMusic:=vol;
 soundSys.SetMusicVolume(FVolMusic);
end;

procedure TOzoneMainClass.ShowMainMenu;
begin
 ActMenu:=MainMenu;
 MDone:=False;
 ActMenu.Active:=true;
 while not MDone do begin
  gfxSys.DoMenu(ActMenu);
  inputSys.Tick;
  if ActMenu.exitMenu then
   ExitMenu(nil);
 end;
end;

procedure TOzoneMainClass.MKeyDown;
begin
 if key=O3K_ESCAPE then begin
  if actMenu=MainMenu then
   MDone:=true
  else begin
   actMenu.Free;
   actMenu:=MainMenu;
   actMenu.active:=true;
  end;
 end;
end;

procedure TOzoneMainClass.volSound;
begin
 FVolSound:=vol;
 soundLib.fxVolume:=FVolSound;
end;

destructor TOzoneMainClass.Destroy;
begin
 if controls<>nil then controls.Free;
// QuitAudio;
// QuitFont;
// quitParticleSystem;
// DeleteInvGfx;
 MainMenu.Free;
 ShadowLib.Free;
 FreeTiles;
 particleSystem.Free;
 SoundLib.Free;
 if IOSys<>nil then IOSys.Free;
 O3Ini.Free;
end;

procedure TOzoneMainClass.LoadTiles(frompath:string);
begin
  gfxSys.writeProgress('LOADING OBJECTS');
  Objects := TObjectList.create(frompath+'index',self);
  gfxSys.WriteProgress('LOADING STONES');
  SpecialStones := TSpecialStoneList.create(frompath+'special-stones',self);
  gfxSys.WriteProgress('LOADING ITEMS');
  SpecialItems := TSpecialItemList.create(frompath+'special-items',self);
  gfxSys.WriteProgress('LOADING ACTORS');
  SpecialActors := TSpecialActorList.create(frompath+'special-actors',self);
  gfxSys.WriteProgress('LOADING MENUS');
end;

procedure TOzoneMainClass.FreeTiles;
begin
  Objects.Free;
  SpecialStones.Free;
  SpecialItems.Free;
  SpecialActors.Free;
end;

function TOzoneMainClass.StartLevelPack;
type TLevPackLevel = record
      t:string;
      filename:string;
      loader:string;
      args:string;
     end;
var x:integer;
    Inp:TextFile;
    levels:array of TLevPackLevel;
    actLevel:integer;
    s:string;
    b:pointer;
begin
// controls:=TControlList.Create(path+ControlSubDir+'index');
 b:=SoundSys.LoadMusic(path+SoundSubDir+'bgmusic.ogg');
 SoundSys.PlayMusic(b,99999); // "Infinite" Repeat!
 actLevel:=slevel;
 SetLength(levels,0);
 Assign(Inp,path+lp+'index');
 Reset(Inp);
 x:=1;
 while not Eof(Inp) do begin
  ReadLn(Inp,s);
  if (s<>'')and(s[1]<>';') then begin
   SetLength(levels,x);
   levels[x-1].t:=GetToken(s,1,':');
   levels[x-1].loader:=GetToken(s,2,':');
   levels[x-1].filename:=GetToken(s,3,':');
   levels[x-1].args:=GetToken(s,4,':');
   inc(x);
  end;
 end;
 Close(Inp);
 while actLevel<Length(levels) do begin
  if levels[actLevel].t='oxyd' then begin
   case StartLevelOxyd(path+LevelSubDir+levels[actLevel].loader+LibExt,lp+levels[actLevel].filename,levels[actLevel].args) of
    won: inc(actLevel);
    aborted,lost: begin
              SoundSys.HaltMusic;
              SoundSys.FreeMusic(b);
              Result:=actLevel;
              Exit;
             end;
   end;
  end;
 end;
 SoundSys.HaltMusic;
 SoundSys.FreeMusic(b);
 Result:=Length(levels)-1;
end;


function TOzoneMainClass.StartLevelOxyd;
var listener:TOzoneInGameListener;
begin
  Done := False;
  O3Log.LogStatus('---- LOADING LEVEL ----','StartLevelOxyd');
  if loader<>'nil' then theLevel := TLevel.Create(self,loader,level,args,CreateStone,CreateItem,CreateActor);
  O3Log.LogStatus('loaded level','StartLevelOxyd');
  Result := aborted;
  listener:=TOzoneInGameListener.Create;
  listener.owner:=self;
  inputSys.addListener(listener);
  O3Log.LogStatus('entering game loop','StartLevelOxyd');
  while ( not Done ) do
  begin
//   ProcessGameEvents;
//   ControlsTick;
   inputSys.Tick;
   gfxSys.DoLevel(theLevel);
   if theLevel.state=LevWon then begin
    soundLib.playSound('levend',theLevel.theActorToFollow.x,theLevel.theActorToFollow.y,theLevel.theActorToFollow.z);
    gfxSys.MakeLevelEnd(theLevel);
    Done:=true;
    Result:=won;
   end;
   if theLevel.state=LevLost then begin
    Done:=true;
    Result:=lost;
   end;
   if theLevel.state=LevRestart then begin
    Done:=true;
    Result:=restart;
   end;
  end;
  if loader<>'nil' then theLevel.Free;
  inputSys.deleteListener(listener);
  listener.Free;
  particleSystem.clear;
end;

end.

