unit o3m;

// FINISH POINTER TO ARRAYS


interface
{$DEFINE OZONE} // Include OZONE-dependant parts
uses
  Classes, dglOpenGL;

type
  TString255 = string[255];
  PVector3s = ^TVector3s;
  TVector3s = record
    x, y, z: Single;
  end;

  PVector3sa = ^TVector3sa;
  TVector3sa = array[0..2] of Single;

  PO3Vertex = ^TO3Vertex;
  TO3Vertex = record
    Vertex,Normal:TVector3s;
  end;
  TO3VertexArray = array[0..50000] of TO3Vertex;
  PO3VertexArray = ^TO3VertexArray;

  PO3UV = ^TO3UV;
  TO3UV = record
    u, v : Single;
  end;

  PO3Surface = ^TO3Surface;
  TO3Surface = record
    Vertref     : array[0..2] of Integer;
    UVs         : array[0..2] of TO3UV;
//    Normal      : TVector3s;
    Material    : Word;
    TwoSide     : Boolean;
  end;
  TO3SurfaceArray = array[0..50000] of TO3Surface;
  PO3SurfaceArray = ^TO3SurfaceArray;


  PO3Object = ^TO3Object;
  TPO3ObjectArray = array[0..50000] of PO3Object;
  PPO3ObjectArray = ^TPO3ObjectArray;
  TO3Object = record
   // User data
    Scale            : TVector3d;
    Rotate           : TVector3d;
    Trans            : TVector3d;
    Matr             : array [0..15] of Single;
    Texture          : Cardinal;
   // Load-/Saveable
    Loc              : TVector3s;
    
    NumVert          : Word;
    Vertices         : PO3VertexArray;

    NumSurf          : Word;
    Surfaces         : PO3SurfaceArray;

    TextureRepeat_X,
    TextureRepeat_Y  : Single;
    TextureOffset_X,
    TextureOffset_Y  : Single;

    TextureFile      : TString255;

    NumKids          : Word;
    Kids             : PPO3ObjectArray;
  end;

  PO3Col = ^TO3Col;
  TO3Col = record
    r, g, b, a : Single;
  end;

  PO3Material = ^TO3Material;
  TO3Material = record
    RGB          : TO3Col;
    Ambient      : TO3Col;
    Specular     : TO3Col;
    Emissive     : TO3Col;
    Shininess    : Single;
    Transparency : Single;
  end;
  TO3MaterialArray = array[0..50000] of TO3Material;
  PO3MaterialArray = ^TO3MaterialArray;


  TO3MFile = class
  public
    highlightColor:TVector3d;
    FObject : PO3Object;
    FPalCount : Word;
    FPalette :PO3MaterialArray;
    IsTransparent:boolean;
    FLastTexID  : Cardinal;
    FLastMatID  : Integer;
    FDispListID : Cardinal;

    owner : pointer;
    procedure FreeObject(AObject : PO3Object = nil);
    procedure InitSurface(var Surface: TO3Surface);

    function NewObject: PO3Object;
    function GetPalEntry(ID: Integer): TO3Material;
    constructor Create(own:pointer);
    destructor Destroy; override;

    // Renderer for openGL
    procedure GLRender(x:integer = -1);
    procedure GLDoRender(AObject: PO3Object);
    procedure GLCreateDisplayList;
    procedure GLRenderDisplayList;
    procedure GLReleaseDisplayList;
    procedure GLSetMaterial(MatID: Integer);

    procedure Release;

    function LoadFromStream(AStream: TStream): Boolean;
    function LoadFromFile(FileName: string): Boolean;
    procedure LoadObject(var obj:PO3Object;str:TStream);

    function SaveToStream(AStream: TStream): Boolean;
    function SaveToFile(FileName: string): Boolean;
    procedure SaveObject(obj:PO3Object;str:TStream);

  end;
const
  NO_TEXTURE = Cardinal(-1);
  NO_DISPLAYLIST = Cardinal(-1);

implementation

uses
  SysUtils{$IFDEF OZONE},o3game,o3utils{$ENDIF};

constructor TO3MFile.Create;
begin
  owner:=own;
  Release;
  highlightColor[0]:=-1;
end;


destructor TO3MFile.Destroy;
begin
  FreeObject;
  inherited;
end;

procedure TO3MFile.FreeObject(AObject : PO3Object = nil);
var
  i : Integer;
  Obj : PO3Object;
begin
  if Assigned(AObject) then
    Obj := AObject
  else
  begin
    if not Assigned(FObject) then Exit;
    Obj := FObject;
  end;
  {$IFDEF OZONE}
  if Obj^.Texture <> NO_TEXTURE then
    if Assigned(owner) then
      TOzoneMainClass(owner).gfxSys.FreeTexture(Obj^.Texture);
  {$ENDIF}

  for i := 0 to Obj^.NumKids - 1 do
    FreeObject(Obj^.Kids^[i]);

//  finalize(Obj^.Surfaces,Obj^.NumSurf*sizeOf(TO3Surface));
//  finalize(Obj^.Vertices,Obj^.NumVert*sizeOf(TO3Vertex));
//  finalize(Obj^.Kids,Obj^.NumKids*sizeOf(Pointer));
  freeMem(Obj^.Surfaces,Obj^.NumSurf*sizeOf(TO3Surface));
  freeMem(Obj^.Vertices,Obj^.NumVert*sizeOf(TO3Vertex));
  freeMem(Obj^.Kids,Obj^.NumKids*sizeOf(Pointer));

  Dispose(Obj);
  if AObject = nil then FObject := nil;
end;


procedure TO3MFile.Release;
begin
  //GLReleaseDisplayList;
  FreeObject;
  FreeMem(FPalette,sizeOf(TO3Material)*FPalCount);
  FObject := nil;
  FLastTexID := NO_TEXTURE;
  FLastMatID := -1;
  FPalCount := 0;
  isTransparent:=false;
end;

function TO3MFile.LoadFromFile(FileName: string): Boolean;
var
  FileStream : TFileStream;
begin
  FileStream := TFileStream.Create(FileName, fmOpenRead);
  try
    Result := LoadFromStream(FileStream);
  finally
    FileStream.Free;
  end;
end;

function TO3MFile.SaveToFile(FileName: string): Boolean;
var
  FileStream : TFileStream;
  t:file of byte;
begin
  Assign(t,FileName); ReWrite(t); Close(t);
  FileStream := TFileStream.Create(FileName, fmOpenWrite);
  try
    Result := SaveToStream(FileStream);
  finally
    FileStream.Free;
  end;
end;

procedure TO3MFile.LoadObject(var obj:PO3Object;str:TStream);
var x:integer;
begin
 obj:=NewObject;
 with obj^ do begin
  str.Read(Loc,sizeOf(TVector3s));
  str.Read(NumVert,sizeOf(Word));
  getMem(Vertices,NumVert*sizeOf(TO3Vertex));
  if NumVert>0 then
   for x:=0 to NumVert-1 do begin
    str.Read(Vertices^[x],sizeOf(TO3Vertex));
   end;
  str.Read(NumSurf,sizeOf(Word));
  getMem(Surfaces,NumSurf*sizeOf(TO3Surface));
  if NumSurf>0 then
   for x:=0 to NumSurf-1 do begin
    str.Read(Surfaces^[x],sizeOf(TO3Surface));
   end;
  str.Read(TextureRepeat_X,sizeOf(single));
  str.Read(TextureRepeat_Y,sizeOf(single));
  str.Read(TextureOffset_X,sizeOf(single));
  str.Read(TextureOffset_Y,sizeOf(single));
  str.Read(TextureFile,sizeOf(TString255));
  {$IFDEF OZONE}
   if TextureFile<>'' then
    TOzoneMainClass(owner).gfxSys.LoadTexture(Texture,TextureFile,TextureSubDir);
  {$ENDIF}
  str.Read(NumKids,sizeOf(Word));
  getMem(kids,numKids*sizeOf(Pointer));
  if numKids>0 then
   for x:=0 to numKids-1 do
    LoadObject(kids^[x],str);
 end;
end;

procedure TO3MFile.SaveObject(obj:PO3Object;str:TStream);
var x:integer;
begin

 with obj^ do begin
  str.Write(Loc,sizeOf(TVector3s));
  str.Write(NumVert,sizeOf(Word));
  if NumVert>0 then
   for x:=0 to NumVert-1 do
    str.Write(Vertices^[x],sizeOf(TO3Vertex));
  str.Write(NumSurf,sizeOf(Word));
  if NumSurf>0 then
   for x:=0 to NumSurf-1 do
    str.Write(Surfaces^[x],sizeOf(TO3Surface));
  str.Write(TextureRepeat_X,sizeOf(single));
  str.Write(TextureRepeat_Y,sizeOf(single));
  str.Write(TextureOffset_X,sizeOf(single));
  str.Write(TextureOffset_Y,sizeOf(single));
  str.Write(TextureFile,sizeOf(TString255));
  str.Write(NumKids,sizeOf(Word));
  if numKids>0 then
   for x:=0 to numKids-1 do
    SaveObject(kids^[x],str);
 end;
end;

function TO3MFile.LoadFromStream(AStream: TStream): Boolean;
var x:integer;
begin
  if not Assigned(AStream) then
    raise Exception.Create('Undefined stream object!');
  Astream.Read(FPalCount,sizeOf(Word));
  getMem(FPalette,FPalCount*sizeOf(TO3Material));
  if FPalCount>0 then
   for x:=0 to FPalCount-1 do begin
    Astream.Read(FPalette^[x],sizeOf(TO3Material));
    if FPalette^[x].Transparency<>0 then
     IsTransparent:=true;
   end;
  LoadObject(FObject,AStream);
  Result:=true;
end;

function TO3MFile.SaveToStream(AStream: TStream): Boolean;
var x:integer;
begin
  if not Assigned(AStream) then
    raise Exception.Create('Undefined stream object!');
  Astream.Write(FPalCount,sizeOf(Word));
  if FPalCount>0 then
   for x:=0 to FPalCount-1 do
    Astream.Write(FPalette^[x],sizeOf(TO3Material));
  SaveObject(FObject,AStream);
  Result:=true;
end;

function TO3MFile.NewObject: PO3Object;
begin
  New(Result);
  with Result^ do
  begin
    Scale[0] := 1;
    Scale[1] := 1;
    Scale[2] := 1;
    Rotate[0]:= 0;
    Rotate[1]:= 0;
    Rotate[2]:= 0;
    Trans[0] := 0;
    Trans[1] := 0;
    Trans[2] := 0;
    Loc.x    := 0;
    Loc.y    := 0;
    Loc.z    := 0;
    Vertices := nil;
    NumVert  := 0;
    Surfaces := nil;
    NumSurf  := 0;
    Texture  := NO_TEXTURE;
    TextureRepeat_X := 1;
    TextureRepeat_Y := 1;
    TextureOffset_X := 0;
    TextureOffset_Y := 0;
    Matr[0]:=1;
    Matr[1]:=0;
    Matr[2]:=0;
    Matr[3]:=0;

    Matr[4]:=0;
    Matr[5]:=1;
    Matr[6]:=0;
    Matr[7]:=0;

    Matr[8]:=0;
    Matr[9]:=0;
    Matr[10]:=1;
    Matr[11]:=0;

    Matr[12]:=0;
    Matr[13]:=0;
    Matr[14]:=0;
    Matr[15]:=1;

    Kids      := nil;
    NumKids   := 0;
  end;
end;


procedure TO3MFile.InitSurface(var Surface: TO3Surface);
begin
  with Surface do
  begin
    Material   := 0;
//    Normal.x   := 0.0;
//    Normal.y   := 0.0;
//    Normal.z   := 0.0;
  end;
end;

procedure TO3MFile.GLRender(x:integer=-1);
begin
  if not Assigned(FObject) then Exit;
  if FDispListID<>NO_DISPLAYLIST then begin
   if x=-1 then GLDoRender(FObject)
           else GLDoRender(FObject.kids^[x]);
  end
  else begin
   {$IFDEF OZONE}
   if isTransparent and (TOzoneMainClass(owner).gfxSys.renderPass<>1) then exit;
   if (not isTransparent) and (TOzoneMainClass(owner).gfxSys.renderPass<>0) then exit;
   {$ENDIF}
   glPushMatrix;
   glMultMatrixf(@FObject.matr);
   for x:=0 to FObject.NumKids-1 do begin
    with FObject.kids[x]^ do begin
     glPushMatrix;
     glMultMatrixf(@matr);
     glTranslated(Loc.x,Loc.y,Loc.z);
     glRotated(Rotate[0],1,0,0);
     glRotated(Rotate[1],0,1,0);
     glRotated(Rotate[2],0,0,1);
     glTranslated(-Loc.x,-Loc.y,-Loc.z);
     glTranslated(Trans[0],Trans[1],Trans[2]);
     glScaled(Scale[0],Scale[1],Scale[2]);
     GLDoRender(FObject.kids^[x]);
     glPopMatrix;
    end;
   end;
   glPopMatrix;
  end;
end;

function TO3MFile.GetPalEntry(ID: Integer): TO3Material;
begin

  if ID < FPalCount then
    Result := FPalette^[ID]
  else
    FillChar(Result, SizeOf(TO3Material), #0);
end;


procedure TO3MFile.GLSetMaterial(MatID: Integer);
var
  Mat  : TO3Material;
  RGBA : TO3Col;
begin
//  if FLastMatID = MatID then
//    Exit;
  FLastMatID := MatID;
  Mat := FPalette^[MatID];

  RGBA := Mat.RGB;
  RGBA.a := 1 - Mat.Transparency;
  glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, @RGBA);

  RGBA := Mat.Ambient;
  RGBA.a := 1 - Mat.Transparency;
  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, @RGBA);

  glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, @Mat.Emissive);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, @Mat.Specular);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, @Mat.Shininess);

  if RGBA.a > 0 then
  begin
//    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);
  end;
end;


procedure TO3MFile.GLDoRender(AObject: PO3Object);
var
  SNr, i : Integer;
  tx, ty     : single;
  VertToUse  : TO3Vertex;
begin

  glPushMatrix();
  glTranslated(AObject.Loc.x, AObject.Loc.y, AObject.Loc.z);

  if (AObject.Texture <> NO_TEXTURE) then
  begin

    glEnable(GL_TEXTURE_2D);
    if AObject.Texture <> FLastTexID then
    begin
      glBindTexture(GL_TEXTURE_2D, AObject.Texture);
      FLastTexID := AObject.Texture;
    end;
  end
  else
    glDisable(GL_TEXTURE_2D);

  if AObject.NumSurf>0 then begin
   for SNr := 0 to AObject.NumSurf - 1 do
   begin
    if (AObject.Surfaces^[SNr].TwoSide) then
     glDisable(GL_CULL_FACE)
    else
     glEnable(GL_CULL_FACE);
//    glEnable(GL_LIGHTING);
    GLSetMaterial(AObject.Surfaces^[SNr].Material);
    glBegin(GL_TRIANGLES);
    for i := 0 to 2 do
    begin
     VertToUse := AObject.Vertices^[AObject.Surfaces^[SNr].Vertref[i]];
     if (AObject.Texture <> NO_TEXTURE) then
     begin
       tx := AObject.TextureOffset_X + AObject.Surfaces^[SNr].UVs[i].u *
             AObject.TextureRepeat_X;
       ty := AObject.TextureOffset_Y + AObject.Surfaces^[SNr].UVs[i].v *
             AObject.TextureRepeat_Y;
       glTexCoord2f(tx, ty);
     end;
     glNormal3fv(@VertToUse.Normal);
     glVertex3fv(@VertToUse.Vertex);
    end;
    glEnd;
   end;
  end;
  if AObject.NumKids > 0 then
    for i := 0 to AObject.NumKids - 1 do
      GLDoRender(AObject.Kids^[i]);

  glPopMatrix;
end;


procedure TO3MFile.GLCreateDisplayList;
var x:cardinal;
begin
  if FDispListID <> 0 then Exit;
  if FObject.NumKids <= 0 then Exit;

  FDispListID := glGenLists(FObject.NumKids);
  if FDispListID = 0 then
    raise Exception.Create('Error generating a DisplayList!');
  for x:=0 to FObject.NumKids-1 do begin
   glNewList(FDispListID+x, GL_COMPILE);
   GLRender(x);
   glEndList;
  end;
end;


procedure TO3MFile.GLReleaseDisplayList;
begin
  if FDispListID <> 0 then
    glDeleteLists(FDispListID, FObject.NumKids);
  FDispListID := 0;
end;


procedure TO3MFile.GLRenderDisplayList;
var x:cardinal;
begin
 {$IFDEF OZONE}
 if isTransparent and (TOzoneMainClass(owner).gfxSys.renderPass<>1) then exit;
 if (not isTransparent) and (TOzoneMainClass(owner).gfxSys.renderPass<>0) then exit;
 {$ENDIF}
 glPushMatrix;
 glMultMatrixf(@FObject.matr);
 for x:=0 to FObject.NumKids-1 do begin
  with FObject.kids[x]^ do begin
   glPushMatrix;
   glMultMatrixf(@matr);
   glTranslated(Loc.x,Loc.y,Loc.z);
   glRotated(Rotate[0],1,0,0);
   glRotated(Rotate[1],0,1,0);
   glRotated(Rotate[2],0,0,1);
   glTranslated(-Loc.x,-Loc.y,-Loc.z);
   glTranslated(Trans[0],Trans[1],Trans[2]);
   glScaled(Scale[0],Scale[1],Scale[2]);
//   glLightfv(GL_LIGHT0,GL_COLOR,@color);
   glCallList(FDispListID+x);
   glPopMatrix;
  end;
 end;
 glPopMatrix;
end;

(*{procedure TO3MFile.ObjectCalcVertexNormals(Obj: PO3Object);
var
  SurfCnt,
  VertCnt,
  VertrefCnt,
  found, i    : Integer;
  Normal      : TACNormal;
  Surface     : TO3Surface;
begin
  for VertCnt := 0 to Obj^.NumVert - 1 do
  begin
    Normal.x := 0;
    Normal.y := 0;
    Normal.z := 0;
    found := 0;

    for SurfCnt := 0 to Obj^.NumSurf - 1 do
    begin
	    Surface := Obj^.Surfaces[SurfCnt];
	    for VertRefCnt := 0 to Surface.NumVertRef - 1 do
      begin
        if (Surface.Vertref[VertRefCnt] = VertCnt) then
		    begin
          Normal.x := Normal.x + Surface.normal.x;
		      Normal.y := Normal.y + Surface.normal.y;
          Normal.z := Normal.z + Surface.normal.z;
          Inc(found);
        end;
      end;
    end;

    if (found > 0) then
    begin
	    Normal.x := Normal.x / found;
	    Normal.y := Normal.y / found;
	    Normal.z := Normal.z / found;
    end;

    Obj^.Vertices[VertCnt].normal := Normal;
  end;

  for i := 0 to Obj^.NumKids - 1 do
    ObjectCalcVertexNormals(Obj^.Kids[i]);
end;
} *)


end.
