unit o3opengl;

interface
uses SDL,
     dglOpenGL,
     o3world,
     o3objects,
     o3vectors,
     o3utils,
     o3actors,
     SysUtils,
     o3opengl_font,
     o3interfaces,
     StrUtils,
     SDL_Image,
     ac3d,
     o3m;
type TGLPlaneModel = record
      Dlist:Cardinal;
      tex:Cardinal;
     end;
     TFogColor = record
      r,g,b,a:Single;
     end;
     PGLPlaneModel = ^TGLPlaneModel;
     TO3GfxSystem_OGLSDL = class(TO3GfxSystem)
      private
       t1,t2:string;
       procedure renderEnvBox(lev:pointer);
      public
      s:integer;
      xpos,ypos,zpos:single;
      timestamp:cardinal;
      ftime:cardinal;
      CFPS:single;
      delta:integer;
      FontSystem:TO3GfxSystem_SDLOGL_FontSubSys;
      DrawFloorShadow:boolean;
      DrawStoneShadow:boolean;
      DrawActorShadow:boolean;
      DrawItemShadow:boolean;
      gridM:TAC3dFile;
      LightPosition:array[0..3] of GLfloat;
      procedure DrawText( tx, ty : single; Text : string; fontset : cardinal; s:single;r,g,b:single;reset:boolean=true ); override;
      function LoadGLTexture(var tex:cardinal;filename:string) : Boolean;
      function LoadGLTexturePNG(var tex:cardinal;filename:string) : Boolean;
      procedure LoadTexture(var tex:Cardinal;text:string;subdir:string); override;
      procedure FreeTexture(var tex:Cardinal); override;
      constructor Create(o:pointer); override;
      procedure DrawModel(m:pointer;px,py,pz:single;mtype:TModelType;gfxinfo:PGfxInfo=nil;abscoord:boolean=false); override;
//      procedure HighlightModel(m:pointer;mtype:TModelType;r,g,b:single); override;
      procedure DoLevel(lev:pointer); override;

      procedure lookAt(xc,yc,zc:single;xp,yp,zp:single;xu,yu,zu:single); override;
      procedure viewport(x,y,w,h:single); override;
      procedure popmatrix; override;
      procedure pushmatrix; override;
      procedure swap; override;
      procedure clear; override;
      procedure setLightPos(x,y,z,w:single); override;
      procedure PaintLevel(lev:pointer;delta:integer;ex,ey,ez,cx,cy,cz:single;plane:integer=-1;grid:boolean=false); override;
      procedure PaintGrid; override;

      procedure enableLighting; override;
      procedure disableLighting; override;
      procedure setColor(r,g,b,a:single); override;
      procedure drawLine(x1,y1,z1,x2,y2,z2:single;w:integer); override;

      procedure disableDepthTest; override;
      procedure enableDepthTest; override;


      procedure loadIdentityMatrix; override;
      procedure setProjectionMatrix; override;
      procedure setViewMatrix; override;
      procedure setPerspective(fov,aspect,zNear,zFar:Single); override;

      procedure DoMenu(menu:pointer); override;
      procedure MakeLevelEnd(lev:pointer); override;
      procedure writeProgress(s:string); override;
      procedure translate(x,y,z:single); override;
      function CreatePlaneXYModel(xs,ys:single;text:string;sDir:string):pointer; override;
      procedure FreePlaneXYModel(var model:pointer); override;
      procedure DrawShadows(y:integer;lev:pointer;cx:single=0;cy:single=0;cz:single=0;dc:boolean=false);
      procedure rotate(ang,x,y,z:single); override;
      destructor destroy; override;
     end;
implementation
 uses odeimport,
 o3inventary,
 o3particles,
 o3game,
 md2,
 o3menu;
const fcol:TFogColor = (r:0.7;g:0.7;b:1;a:0);

procedure TO3GfxSystem_OGLSDL.disableDepthTest;
begin
 glDisable(GL_DEPTH_TEST);
end;

procedure TO3GfxSystem_OGLSDL.enableDepthTest;
begin
 glEnable(GL_DEPTH_TEST);
end;


procedure TO3GfxSystem_OGLSDL.enableLighting;
begin
 glEnable(GL_LIGHTING);
end;

procedure TO3GfxSystem_OGLSDL.disableLighting;
begin
 glDisable(GL_LIGHTING);
end;

procedure TO3GfxSystem_OGLSDL.setColor(r,g,b,a:single);
begin
 glColor4f(a,g,b,a);
end;

procedure TO3GfxSystem_OGLSDL.drawLine(x1,y1,z1,x2,y2,z2:single;w:integer);
begin
 glPushAttrib(GL_ALL_ATTRIB_BITS);
 glLineWidth(w);
 glDisable(GL_TEXTURE_2D);
 glBegin(GL_LINES);
  glVertex3f(x1,y1,z1);
  glVertex3f(x2,y2,z2);
 glEnd;
 glPopAttrib;
 glLineWidth(1);
end;


destructor TO3GfxSystem_OGLSDL.destroy;
begin
 gridm.Free;
 inherited Destroy;
end;

procedure TO3GfxSystem_OGLSDL.PaintGrid;
begin
// dglopengl.glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
 self.DrawModel(gridm,0,0,0,ac);
 glColor3f(1,1,1);
// dglopengl.glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
end;


{procedure TO3GfxSystem_OGLSDL.HighlightModel;
var oz:TO3mFile;
begin
 if mtype=ozone then begin
  oz:=m;
  oz.highlightColor[0]:=r;
  oz.highlightColor[1]:=g;
  oz.highlightColor[2]:=b;
 end;
end;
 }

procedure TO3GfxSystem_OGLSDL.rotate;
begin
 glRotatef(ang,x,y,z);
end;

procedure TO3GfxSystem_OGLSDL.setLightPos;
var LP:array[0..3] of GLFloat;
begin
 LP[0]:=LightPosition[0];
 LP[1]:=LightPosition[1];
 LP[2]:=LightPosition[2];
 LP[3]:=LightPosition[3];
 if(w>=-1)and(w<=1) then begin
  LP[0]:=x;
  LP[1]:=y;
  LP[2]:=z;
  LP[3]:=w;
 end;
 glLightfv(GL_LIGHT0, GL_POSITION, @LP[0]);
 glEnable(GL_LIGHTING);
end;

procedure TO3GfxSystem_OGLSDL.loadIdentityMatrix;
begin
 glLoadIdentity;
end;

procedure TO3GfxSystem_OGLSDL.swap;
begin
 SDL_GL_SwapBuffers;
end;

procedure TO3GfxSystem_OGLSDL.clear;
begin
 glClear(GL_DEPTH_BUFFER_BIT or GL_COLOR_BUFFER_BIT or GL_STENCIL_BUFFER_BIT);
end;

procedure TO3GfxSystem_OGLSDL.setProjectionMatrix;
begin
 glMatrixMode(GL_PROJECTION);
end;

procedure TO3GfxSystem_OGLSDL.setViewMatrix;
begin
 glMatrixMode(GL_MODELVIEW);
end;

procedure TO3GfxSystem_OGLSDL.setPerspective;
begin
 gluPerspective(fov,aspect,zNear,zFar);
end;

procedure TO3GfxSystem_OGLSDL.lookAt;
begin
 gluLookAt(xc,yc,zc,xp,yp,zp,xu,yu,zu);
end;

procedure TO3GfxSystem_OGLSDL.viewport;
var xt,yt:single;
begin
 xt:=round(1/2.66*TOzoneMainClass(owner).VidW);
 yt:=round(1/2*TOzoneMainClass(owner).VidH);
 glViewport(round(x*xt+TOzoneMainClass(owner).VidW div 2),round(-y*yt+TOzoneMainClass(owner).VidH div 2),round(w*xt),round(h*yt))
end;
procedure TO3GfxSystem_OGLSDL.popmatrix;
begin
 glPopMatrix;
end;
procedure TO3GfxSystem_OGLSDL.pushmatrix;
begin
 glPushMatrix;
end;
procedure TO3GfxSystem_OGLSDL.PaintLevel;
var level:Tlevel;
    i:integer;
begin
   level:=TLevel(lev);
   renderEnvBox(lev);
//   glEnable(GL_FOG);
   glLightfv(GL_LIGHT0, GL_POSITION, @LightPosition[0]);
 //  glFogfv(GL_FOG_COLOR, @fcol);
 //  glFogi(GL_FOG_MODE,GL_LINEAR);
  // glFogf(GL_FOG_START, 15);
  // glFogf(GL_FOG_END, 30);
   renderPass:=0;
   level.Render(delta,-1E555,0,0,0,plane,grid);
   level.FActors.RenderAll(delta);
   renderPass:=1;
   level.FActors.RenderAll(0);

   if DrawShadow then begin
    glClear(GL_STENCIL_BUFFER_BIT);
    if DrawFloorShadow then begin
     for i:=0 to level.FY-1 do begin
      if ex>-1E555 then DrawShadows(i,level,ex,ey,ez,true)
      else DrawShadows(i,level);
     end;
    end
    else begin
     for i:=1 to level.FY-1 do begin
      if ex>-1E555 then DrawShadows(i,level,ex,ey,ez,true)
      else DrawShadows(i,level);
     end;
    end;
    DrawShadows(-2,level);

    level.Render(0,ex,ez,cx,cz,plane,grid);
    glPushAttrib(GL_ALL_ATTRIB_BITS);
    glPushMatrix;
    TOzoneMainClass(owner).particleSystem.render(delta);
    glPopMatrix;
    glPopAttrib;
//    glDisable(GL_FOG);

    glPushAttrib(GL_ALL_ATTRIB_BITS);
    glEnable(GL_STENCIL_TEST);
    glDisable(GL_LIGHTING);
    glColor4f(0,0,0,0.4 );
    glStencilFunc(GL_NOTEQUAL,0,$FFFFFFFF);
    glStencilOp(GL_KEEP,GL_KEEP,GL_KEEP);
    glPushMatrix;
    glLoadIdentity;
    glBegin(GL_TRIANGLE_STRIP);
     glVertex3f(-1, 1,-1);
     glVertex3f(-1,-1,-1);
     glVertex3f( 1, 1,-1);
     glVertex3f( 1,-1,-1);
     glEnd;
    glPopMatrix;
    glPopAttrib;
   end
   else begin
    level.Render(0,ex,ez,cx,cz,plane,grid);
    TOzoneMainClass(owner).particleSystem.render(delta);
//    glDisable(GL_FOG);
   end;

   renderPass:=0;
end;


function TO3GfxSystem_OGLSDL.CreatePlaneXYModel;
var t:pGLPlaneModel;
begin
 New(t);
 LoadTexture(t.tex,text,sDir);
 glEnable(GL_TEXTURE_2D);
 t.Dlist:=glGenLists(1);
 glBindTexture( GL_TEXTURE_2D, t.Tex );
 glNewList(t.dlist,GL_COMPILE);
  glBegin( GL_QUADS );
   glTexCoord2f(0, 0);
   glVertex2f(-xs/2, ys/2);
   glTexCoord2f(1, 0);
   glVertex2f(xs/2, ys/2);
   glTexCoord2f(1, 1);
   glVertex2f(xs/2, -ys/2);
   glTexCoord2f(0, 1);
   glVertex2f(-xs/2, -ys/2);
  glEnd;
 glEndList;
 glDisable(GL_TEXTURE_2D);
 Result:=t;
end;

procedure TO3GfxSystem_OGLSDL.FreePlaneXYModel;
begin
 FreeTexture(PGLPlaneModel(model).tex);
 glDeleteLists(PGLPlaneModel(model).dList,1);
 Dispose(model);
 model:=nil;
end;
procedure TO3GfxSystem_OGLSDL.translate;
begin
 glTranslatef(x,y,z);
 xpos:=xpos+x;
 ypos:=ypos+y;
 zpos:=zpos+z;
end;

procedure TO3GfxSystem_OGLSDL.writeProgress;
begin
  if not noclear then glClear( GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT or GL_STENCIL_BUFFER_BIT);
  glLoadIdentity;
  if (s<>'')and(s[1]=UpCase(s[1])) then begin
   t1:=s;
   t2:='';
   TOzoneMainClass(owner).O3Log.LogStatus('-- '+t1+' --','ozone init');
  end
  else begin
   t2:=s;
   TOzoneMainClass(owner).O3Log.LogStatus(t2,'ozone init');
  end;
  DrawText(-Length('[\]^_')*8*0.005*4+0.05,0.2,'[\]^_',1,4,1,1,1);
  DrawText(-Length(t1)*8*0.005*2+0.05,0,t1,1,2,1,0,0);
  DrawText(-Length(t2)*8*0.005+0.05,-0.2,t2,0,1,0,0,1);
  if not NoSwap then SDL_GL_SwapBuffers;
end;

procedure TO3GfxSystem_OGLSDL.DrawText;
begin
 //glPushMatrix;
 glPushAttrib(GL_ALL_ATTRIB_BITS);
 FontSystem.DrawText(tx,ty,text,fontset,s,r,g,b,reset);
 glPopAttrib;
// dglopengl.
// glColor3f( 1,1,1 );
 //glPopMatrix;
end;

function TO3GfxSystem_OGLSDL.LoadGLTexture;
var
  TextureImage : PSDL_Surface;
begin
  if RightStr(filename,3)='png' then begin
   Result:=LoadGLTexturePNG(tex,filename);
   Exit;
  end;
  TextureImage := SDL_LoadBMP(PChar(filename));
  if ( TextureImage <> nil ) then
  begin
//    if flip then
//     SDL_FlipRectV(textureImage,nil);
    glGenTextures(1, @tex );
    glBindTexture( GL_TEXTURE_2D, tex );
    gluBuild2dMipMaps(GL_TEXTURE_2D,3,TextureImage.w,TextureImage.h,
                      GL_BGR,GL_UNSIGNED_BYTE,TextureImage.pixels);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
  end
  else
  begin
    result:=false;
    exit;
  end;
  if ( TextureImage <> nil ) then
   SDL_FreeSurface(TextureImage);
  result := true;
end;


function TO3GfxSystem_OGLSDL.LoadGLTexturePNG;
var
  TextureImage : PSDL_Surface;
begin
  TextureImage := SDL_Image.IMG_Load(PChar(filename));

  if ( TextureImage <> nil ) then
  begin
//    if flip then
//     SDL_FlipRectV(textureImage,nil);
    glGenTextures(1, @tex );
    glBindTexture( GL_TEXTURE_2D, tex );
    gluBuild2dMipMaps(GL_TEXTURE_2D,4,TextureImage.w,TextureImage.h,
                      GL_RGBA,GL_UNSIGNED_BYTE,TextureImage.pixels);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
  end
  else
  begin
    result:=false;
    exit;
  end;
  if ( TextureImage <> nil ) then
   SDL_FreeSurface(TextureImage);
  result := true;
end;

procedure TO3GfxSystem_OGLSDL.FreeTexture(var tex:Cardinal);
begin
 glDeleteTextures(1,@tex);
end;

procedure TO3GfxSystem_OGLSDL.LoadTexture;
begin
 LoadGLTexture(tex,TOzoneMainClass(owner).path+subdir+text);
end;

procedure TO3GfxSystem_OGLSDL.DrawModel;
var model:TAC3dFile;
    p:pGLPlaneModel;
    md:TQuakeModel;
    om:TO3MFile;
begin
 glPushMatrix;
 if (gfxinfo<>nil) then begin
  glPushAttrib(GL_ALL_ATTRIB_BITS);
  if (gfxinfo.highlightColor[0]<>-1)and((gfxinfo.highlightColor[0]<>1)or(gfxinfo.highlightColor[1]<>1)or(gfxinfo.highlightColor[2]<>1)or(gfxinfo.highlightColor[3]<>1)) then begin
   glDisable(GL_LIGHTING);
   glColor4fv(@gfxinfo.highlightColor[0]);
   if (gfxinfo.highlightColor[3]<1) then begin
    glDepthMask(false);
   end;
  end;
  if gfxinfo.wireframe then
   glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
  if gfxinfo.hide then begin
   glPopMatrix;
   glPopAttrib;
   EXIT;
  end;
 end;
 if abscoord then
  glTranslatef(-xpos,-ypos,-zpos);
 if mtype=ac then begin
  model:=m;
  if model.DisplayListID=0 then
   model.CreateDisplayList;
  glTranslatef(px,py,pz);
  model.RenderDisplayList;
 end;
 if mtype=quake then begin
  md:=m;
  glTranslatef(px,py,pz);
  md.Show;
 end;
 if mtype=plane then begin
  p:=m;
  glEnable(GL_TEXTURE_2D);
  glBindTexture( GL_TEXTURE_2D, p.Tex );
  glTranslatef(px,py,pz);
  glCallList(p.Dlist);
  glDisable(GL_TEXTURE_2D);
 end;
 if mtype=ozone then begin
  om:=m;
  if om.FDispListID=0 then
   om.GLCreateDisplayList;
  glTranslatef(px,py,pz);
  if om.FDispListID<>NO_DISPLAYLIST then om.GLRenderDisplayList
  else om.GLRender();
 end;
 if (gfxinfo<>nil) then
  glPopAttrib;
 glPopMatrix;
end;

constructor TO3GfxSystem_OGLSDL.Create;
begin
  inherited;
  FontSystem:=TO3GfxSystem_SDLOGL_FontSubSys.create(self);
  DrawShadow:=TOzoneMainClass(owner).O3Ini.ReadBool('gfx','shadow',true);
  DrawFloorShadow:=TOzoneMainClass(owner).O3Ini.ReadBool('gfx','floorshadow',true);
  DrawItemShadow:=TOzoneMainClass(owner).O3Ini.ReadBool('gfx','itemshadow',true);
  DrawStoneShadow:=TOzoneMainClass(owner).O3Ini.ReadBool('gfx','stoneshadow',true);
  DrawActorShadow:=TOzoneMainClass(owner).O3Ini.ReadBool('gfx','actorshadow',true);
  timestamp:=0;
  glShadeModel( GL_SMOOTH );
  glClearColor( 0.0, 0.0, 0.0, 0 );
  glClearDepth( 1.0 );
  glClearStencil( 0 );
  LightPosition[0] := 500.0;
  LightPosition[1] := 1000.0;
  LightPosition[2] := 500.0;
  LightPosition[3] := 0.0;
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glLightfv(GL_LIGHT0, GL_POSITION, @LightPosition[0]);
  glEnable( GL_DEPTH_TEST );
  glDepthFunc( GL_LEQUAL );
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
  glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );
  glHint( GL_FOG_HINT, GL_NICEST );
  s:=0;
  ftime:=0;
  delta:=0;
  cFPS:=100;
  gridM:=TAC3dFile.Create(owner);
  gridM.LoadFromFile(TOzoneMainClass(owner).path+ModelSubDir+'grid.ac');
end;

procedure TO3GfxSystem_OGLSDL.renderEnvBox;
var FrontTex,BackTex,TopTex,BottomTex,RightTex,LeftTex:cardinal;
    level:TLevel;
    matrix:array[0..15] of Single;
begin
  glEnable(GL_TEXTURE_2D);
  glDisable(GL_CULL_FACE);
  glDisable(GL_LIGHTING);
  glGetFloatv(GL_MODELVIEW_MATRIX,@matrix);
  glPushMatrix;
  matrix[12]:=0;
  matrix[13]:=0;
  matrix[14]:=0;
  glLoadMatrixf(@matrix);
  glDisable(GL_DEPTH_TEST);
  glDepthMask(false);
  level:=lev;
  FrontTex:=level.envBoxTex[ebFront];
  BackTex:=level.envBoxTex[ebBack];
  TopTex:=level.envBoxTex[ebTop];
  BottomTex:=level.envBoxTex[ebBottom];
  LeftTex:=level.envBoxTex[ebLeft];
  RightTex:=level.envBoxTex[ebRight];
  glBindTexture(GL_TEXTURE_2D, FrontTex);  // Bind the Texture to the object
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
  glBegin(GL_QUADS);
    glNormal3f( 0.0, 0.0,-1.0);
    glTexCoord2f(0.0, 1.0); glVertex3f(-1.01, -1.01,  1.0);
    glTexCoord2f(1.0, 1.0); glVertex3f( 1.005, -1.005,  1.0);
    glTexCoord2f(1.0, 0.0); glVertex3f( 1.005,  1.005,  1.0);
    glTexCoord2f(0.0, 0.0); glVertex3f(-1.005,  1.005,  1.0);
  glend;

  glBindTexture(GL_TEXTURE_2D, BackTex);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
  glBegin(GL_QUADS);
    glNormal3f( 0.0, 0.0,1.0);
    glTexCoord2f(1.0, 1.0); glVertex3f(-1.005, -1.005, -1.0);
    glTexCoord2f(1.0, 0.0); glVertex3f(-1.005,  1.005, -1.0);
    glTexCoord2f(0.0, 0.0); glVertex3f( 1.005,  1.005, -1.0);
    glTexCoord2f(0.0, 1.0); glVertex3f( 1.005, -1.005, -1.0);
  glEnd;

  glBindTexture(GL_TEXTURE_2D, TopTex);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
  glBegin(GL_QUADS);
    glNormal3f( 0.0, -1.0, 0.0);
    glTexCoord2f(0.0, 1.0); glVertex3f(-1.005,  1.0, -1.005);
    glTexCoord2f(1.0, 1.0); glVertex3f(-1.005,  1.0,  1.005);
    glTexCoord2f(1.0, 0.0); glVertex3f( 1.005,  1.0,  1.005);
    glTexCoord2f(0.0, 0.0); glVertex3f( 1.005,  1.0, -1.005);
  glEnd;

  glBindTexture(GL_TEXTURE_2D, BottomTex);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
  glBegin(GL_QUADS);
    glNormal3f( 0.0,1.0, 0.0);
    glTexCoord2f(0.0, 0.0); glVertex3f(-1.005, -1.0, -1.005);
    glTexCoord2f(0.0, 1.0); glVertex3f( 1.005, -1.0, -1.005);
    glTexCoord2f(1.0, 1.0); glVertex3f( 1.005, -1.0,  1.005);
    glTexCoord2f(1.0, 0.0); glVertex3f(-1.005, -1.0,  1.005);
  glEnd;

  glBindTexture(GL_TEXTURE_2D, LeftTex);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
  glBegin(GL_QUADS);
    glNormal3f( -1.0, 0.0, 0.0);
    glTexCoord2f(1.0, 1.0); glVertex3f( 1.0, -1.005, -1.005);
    glTexCoord2f(1.0, 0.0); glVertex3f( 1.0,  1.005, -1.005);
    glTexCoord2f(0.0, 0.0); glVertex3f( 1.0,  1.005,  1.005);
    glTexCoord2f(0.0, 1.0); glVertex3f( 1.0, -1.005,  1.005);
  glEnd;

  glBindTexture(GL_TEXTURE_2D, RightTex);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
  glBegin(GL_QUADS);
    glNormal3f(1.0, 0.0, 0.0);
    glTexCoord2f(0.0, 1.0); glVertex3f(-1.0, -1.005, -1.005);
    glTexCoord2f(1.0, 1.0); glVertex3f(-1.0, -1.005,  1.005);
    glTexCoord2f(1.0, 0.0); glVertex3f(-1.0,  1.005,  1.005);
    glTexCoord2f(0.0, 0.0); glVertex3f(-1.0,  1.005, -1.005);

  glEnd();
  glDepthMask(true);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_LIGHTING);
  glPopMatrix;
//  glEnable(GL_TEXTURE_2D);
//  glDisable(GL_CULL_FACE);
end;

procedure TO3GfxSystem_OGLSDL.MakeLevelEnd;
const len = 3000;
var time:integer;
    ox,oy,oz:single;
    level:TLevel;
    p:TActor;
    fcol:TFogColor;
begin
 fcol.r:=0.7;
 fcol.g:=0.7;
 fcol.b:=1;
 fcol.a:=0;
 level:=TLevel(lev);
 p:=level.theActorToFollow;
 time:=0;
 delta:=0;
// p.physics.gravity:=mul(p.physics.gravity,-4);
// p.physics.speed:=GetVector(0,0,0);
 dWorldSetGravity(level.Physics.theWorld,0,9.81,0);
 p.physics.noclip:=true;
 ox:=p.x;
 oy:=p.y;
 oz:=p.z;
 p.Inventary.StartStaticText('Level finished',len*2);
 repeat
  timeStamp:=SDL_GetTicks;
  glViewport( 0, 0, TOzoneMainClass(owner).vidW, TOzoneMainClass(owner).vidH);
  glMatrixMode( GL_PROJECTION );
  glLoadIdentity;
  gluPerspective( TOzoneMainClass(owner).O3Ini.ReadInteger('gfxdebug','viewport1',55), TOzoneMainClass(owner).vidW / TOzoneMainClass(owner).vidH, 0.1, 31.0 );
  glMatrixMode( GL_MODELVIEW );    
  if not noclear then glClear( GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT or GL_STENCIL_BUFFER_BIT);
  glLoadIdentity;
  xpos:=0;
  ypos:=0;
  zpos:=0;
  glPushMatrix;
  if time<(len*0.6) then begin
   gluLookAt(ox+L[p.Look].LX,oy+L[p.Look].LY+time*20/len,oz+L[p.Look].LZ,ox,oy+0.5,oz,0+time/(3*len),1-time/(3*len),0);
   TOzoneMainClass(owner).soundSys.UpdateListener(ox+L[p.Look].LX,oy+L[p.Look].LY+time*20/len,oz+L[p.Look].LZ,-L[p.Look].LX,0.5-time*20/len-L[p.Look].LY,-L[p.Look].LZ);
  end else begin
   gluLookAt(ox+L[p.Look].LX,oy+L[p.Look].LY+(len*0.6)*20/len,oz+L[p.Look].LZ,ox,oy+0.5,oz,0+(len*0.6)/(3*len),1-(len*0.6)/(3*len),0);
   TOzoneMainClass(owner).soundSys.UpdateListener(ox+L[p.Look].LX,oy+L[p.Look].LY+time*20/len,oz+L[p.Look].LZ,-L[p.Look].LX,0.5-(len*0.6)*20/len-L[p.Look].LY,-L[p.Look].LZ);
  end;
{  renderEnvBox(lev);
  glEnable(GL_FOG);
  glFogfv(GL_FOG_COLOR, @fcol);
   glFogi(GL_FOG_MODE,GL_LINEAR);
   glFogf(GL_FOG_START, 15);
   glFogf(GL_FOG_END, 30);
  renderPass:=0;
  level.Render(delta,-1E555,0,0,0);
  level.FActors.RenderAll(delta);
  renderPass:=1;
  level.FActors.RenderAll(0);
//  glDepthMask(false);
  level.Render(0,-1E555,0,0,0);
//  glDepthMask(true);
  renderPass:=0;
  if delta>0 then
   level.Update(delta);
  glDisable(GL_FOG);
  TOzoneMainClass(owner).particleSystem.render(delta);
  glPopMatrix;}
  PaintLevel(lev,delta,-1E555,0,0,p.x,p.y,p.z);
  if delta>0 then
   level.Update(delta);
  glPopMatrix;
  FontSystem.DrawText( -1.2, 0.8, 'FPS: '+FloatToStrF(cFPS,ffGeneral,4,5), 0 , 0.75, 1, 1, 1);
  p.Inventary.Render(delta);
  if not NoSwap then SDL_GL_SwapBuffers;
  inc(s);
  delta:=SDL_GetTicks-timeStamp;
  inc(time,delta);
  inc(FTime,delta);
  if FTime>30 then begin
   cFPS:=s/(FTime/1000);
   s:=0;
   FTime:=0;
  end;
 until(time>len);
 timeStamp:=0;
 ftime:=0;
 delta:=0;
 cFPS:=100;
 TOzoneMainClass(owner).particleSystem.clear;
end;

procedure TO3GfxSystem_OGLSDL.DoMenu;
var   m:TO3Menu;
begin
  glMatrixMode(GL_PROJECTION);
  glPushMatrix;
  glLoadIdentity;
  glOrtho(-1.33,1.33,-1,1,-10,10);
  glDisable(GL_LIGHTING);
//  glColor3f( 1,1,1 );
  glDisable(GL_CULL_FACE);
//  glDepthFunc(GL_GEQUAL);
  glMatrixMode(GL_MODELVIEW);
  m:=TO3Menu(menu);
  timeStamp:=SDL_GetTicks;
  glClearColor(0,0,0.05,0);
  if not noclear then glClear( GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT or GL_STENCIL_BUFFER_BIT);
  glLoadIdentity;
//  gluLookAt(0,0,-2.4,0,0,0,0,1,0);
  glTranslated( 0,0, 5 );
//  glEnable(GL_TEXTURE_2D);
//  glBlendFunc(GL_ONE,GL_ONE);
  if m.showToolTips then begin
   DrawText(-Length('[\]^_')*8*0.005*4+0.05,-0.1,'[\]^_',1,4,1,1,1,false);
   DrawText(-Length(m.aToolTip)*2*0.005*4+0.05,-0.3,m.aToolTip,0,1,1,1,1,false);
  end
  else
   DrawText(0,0,'',1,4,1,1,1,false);
//  TOzoneMainClass(owner).particleSystem.render(delta);
  renderPass:=0;
  m.Render(delta);
  //menus are renderes 1-pass
{  glColor3f(1,1,1);
  glBegin(GL_QUADS);
   glVertex3d(-1,-1,0);
   glVertex3d(-1,1,0);
   glVertex3d(1,1,0);
   glVertex3d(1,-1,0);
  glEnd;}
  if not NoSwap then SDL_GL_SwapBuffers;
  delta:=SDL_GetTicks-timeStamp;
  inc(FTime,delta);
  glClearColor(0,0,0,0);
  glMatrixMode(GL_PROJECTION);
  glPopMatrix;
  glMatrixMode(GL_MODELVIEW);
end;

procedure TO3GfxSystem_OGLSDL.DrawShadows;
var x,z:integer;
    level:TLevel;
    cell:TCell;
    actor:TActor;
    gfx:TO3mFile;
begin
// glClear(GL_STENCIL_BUFFER_BIT);
 glPushAttrib(GL_ALL_ATTRIB_BITS);
 glColorMask(false,false,false,false);
// glDisable(GL_LIGHTING);
 glDepthMask(false);
 glEnable(GL_STENCIL_TEST);
 glColorMask(false,false,false,false);
 glStencilFunc(GL_ALWAYS,1,$FFFFFFFF);
 Level:=TLevel(lev);
 glEnable(GL_CULL_FACE);
 if DrawActorShadow then begin
  if level.FActors.List.Count>0 then
   for x:=0 to level.FActors.List.Count-1 do begin
    actor:=TActor(level.FActors.List[x]);
    if ((trunc(actor.y)=y)or((trunc(actor.y)>=level.FY)and(y=-2)))and(actor.shadow>0) then begin
//     if sqr(actor.x-x)
     gfx:=TOzoneMainClass(owner).ShadowLib.ShadowGFX[actor.shadow].gfx;
     translate(actor.x,actor.y,actor.z);
     if gfx<>nil then begin
      if gfx.FDispListID=0 then
       gfx.GLCreateDisplayList;
      glCullFace(GL_FRONT);
      glStencilOp(GL_KEEP, GL_INCR, GL_KEEP);
      gfx.GLRenderDisplayList;
      glCullFace(GL_BACK);
      glStencilOp(GL_KEEP, GL_DECR, GL_KEEP);
      gfx.GLRenderDisplayList;
      translate(-actor.x,-actor.y,-actor.z);
     end;
    end;
   end;
  end;
 if y>-2 then begin
  translate(0,y,0);
  for x:=0 to level.FX-1 do begin
   for z:=0 to level.FZ-1 do begin
//    if dc and (sqr(x-cx)+sqr(y-cy)+sqr(z-cz)>500) then CONTINUE;
    translate(x,0,z);
    cell:=level.getCell(x,y,z);
    if cell=nil then continue;
    if DrawItemShadow then begin
     if (cell.FItem<>nil)and(cell.FItem.shadow>0) then begin
      gfx:=TOzoneMainClass(owner).ShadowLib.ShadowGFX[cell.FItem.shadow].gfx;
      if gfx<>nil then begin
       if gfx.FDispListID=0 then
        gfx.GLCreateDisplayList;
       glCullFace(GL_FRONT);
       glStencilOp(GL_KEEP, GL_INCR, GL_KEEP);
       gfx.GLRenderDisplayList;
       glCullFace(GL_BACK);
       glStencilOp(GL_KEEP, GL_DECR, GL_KEEP);
       gfx.GLRenderDisplayList;
      end;
     end;
    end;
    if DrawStoneShadow then begin
     if (cell.FStone<>nil)and(cell.FStone.shadow>0) then begin
      gfx:=TOzoneMainClass(owner).ShadowLib.ShadowGFX[cell.FStone.shadow].gfx;
      if gfx<>nil then begin
       if gfx.FDispListID=0 then
        gfx.GLCreateDisplayList;
       glCullFace(GL_FRONT);
       glStencilOp(GL_KEEP, GL_INCR, GL_KEEP);
       gfx.GLRenderDisplayList;
       glCullFace(GL_BACK);
       glStencilOp(GL_KEEP, GL_DECR, GL_KEEP);
       gfx.GLRenderDisplayList;
      end;
     end;
    end;
    translate(-x,0,-z);
   end;
  end;
  translate(0,-y,0);
 end;
 glDisable(GL_CULL_FACE);

 glColorMask(true,true,true,true);
 glPopAttrib;
 glDisable(GL_STENCIL_TEST);
end;

procedure TO3GfxSystem_OGLSDL.DoLevel;
const Nearest = 0;
      Farest = 9;
var   level:TLevel;
      p:TActor;
      p2:TActor;
      w,h:integer;
      ex,ez:single;
begin

  w:=TOzoneMainClass(owner).vidW;
  h:=TOzoneMainClass(owner).vidH;
  level:=TLevel(lev);
  p:=level.theActorToFollow;
  p2:=level.theActorToFollow2;
  timeStamp:=SDL_GetTicks;
  if not noclear then glClear( GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
  if p2<>nil then begin
   TOzoneMainClass(owner).vsplit:=2;
   glViewport( 0, h div 2, w, h div 2 );
   glMatrixMode( GL_PROJECTION );
   glLoadIdentity;
   gluPerspective( TOzoneMainClass(owner).O3Ini.ReadInteger('gfxdebug','viewport2',32), w / (h div 2), 0.1, 31.0 );
   glMatrixMode( GL_MODELVIEW );
  end
  else begin
   TOzoneMainClass(owner).vsplit:=1;
   glViewport( 0, 0, w, h);
   glMatrixMode( GL_PROJECTION );
   glLoadIdentity;
   gluPerspective( TOzoneMainClass(owner).O3Ini.ReadInteger('gfxdebug','viewport1',45), w / h, 0.1, 31.0 );
   glMatrixMode( GL_MODELVIEW );
  end;
   glLoadIdentity;
   xpos:=0;
   ypos:=0;
   zpos:=0;
   glPushMatrix;
   ex:=0;
   ez:=0;
   if p<>nil then begin
     if p.LookFar then
      p.LookFarness:=p.LookFarness+delta/50
     else
      p.LookFarness:=p.LookFarness-delta/50;
     if p.LookFarness>Farest then p.LookFarness:=Farest;
     if p.LookFarness<Nearest then p.LookFarness:=Nearest;
     ex:=p.x+L[p.Look].LX-L[p.Look].LX*p.LookFarness/10;
     ez:=p.z+L[p.Look].LZ-L[p.Look].LZ*p.LookFarness/10;
     gluLookAt(ex,p.y+L[p.Look].LY+p.LookFarness,ez,p.x,p.y+0.5,p.z,L[p.Look].UX,L[p.Look].UY,L[p.Look].UZ);
     TOzoneMainClass(owner).soundSys.UpdateListener(ex,p.y+L[p.Look].LY+p.LookFarness,ez,-L[p.Look].LX+L[p.Look].LX*p.LookFarness/10,0.5-L[p.Look].LY-p.LookFarness,-L[p.Look].LZ+L[p.Look].LZ*p.LookFarness/10);
   end;
   paintLevel(lev,delta,ex,p.y+L[p.Look].LY+p.LookFarness,ez,p.x,p.y,p.z);

   level.Update(delta);
   glPopMatrix;

   FontSystem.DrawText( -1.2, 0.8, 'FPS: '+FloatToStrF(cFPS,ffGeneral,4,5), 0 , 0.75, 1, 1, 1);
   if level.paused then
    FontSystem.DrawText( -0.8, 0, 'PAUSED' , 1 , 3, 1, 1, 1);
   p.Inventary.Render(delta);
  if p2<>nil then begin
   glViewport( 0, 0, w, h div 2 );
   glMatrixMode( GL_PROJECTION );
   glLoadIdentity;
   gluPerspective( TOzoneMainClass(owner).O3Ini.ReadInteger('gfxdebug','viewport2',32), w / (h div 2), 0.1, 100.0 );
   glMatrixMode( GL_MODELVIEW );
   glLoadIdentity;
   xpos:=0;
   ypos:=0;
   zpos:=0;
   glPushMatrix;
   ex:=0;
   ez:=0;
   if p2<>nil then begin
     if p2.LookFar then
      p2.LookFarness:=p2.LookFarness+delta/50
     else
      p2.LookFarness:=p2.LookFarness-delta/50;
     if p2.LookFarness>Farest then p2.LookFarness:=Farest;
     if p2.LookFarness<Nearest then p2.LookFarness:=Nearest;
     ex:=p2.x+L[p2.Look].LX;
     ez:=p2.z+L[p2.Look].LZ;
     gluLookAt(ex,p2.y+L[p2.Look].LY+p2.LookFarness,ez,p2.x,p2.y+0.5,p2.z,L[p2.Look].UX,L[p2.Look].UY,L[p2.Look].UZ);
     TOzoneMainClass(owner).soundSys.UpdateListener(ex,p2.y+L[p2.Look].LY+p2.LookFarness,ez,-L[p2.Look].LX+L[p2.Look].LX*p2.LookFarness/10,0.5-L[p2.Look].LY-p2.LookFarness,-L[p2.Look].LZ+L[p2.Look].LZ*p2.LookFarness/10);
   end;
   paintLevel(lev,0,ex,p2.y+L[p2.Look].LY+p2.LookFarness,ez,p2.x,p2.y,p2.z);
   level.Update(0);
  { renderEnvBox(lev);
   glEnable(GL_FOG);
   glLightfv(GL_LIGHT0, GL_POSITION, @LightPosition[0]);
   glFogfv(GL_FOG_COLOR, @fcol);
   glFogi(GL_FOG_MODE,GL_LINEAR);
   glFogf(GL_FOG_START, 15);
   glFogf(GL_FOG_END, 30);
   renderPass:=0;
   level.Render(0,-1E555,0,0,0);
   level.FActors.RenderAll(0);
   renderPass:=1;
   level.FActors.RenderAll(0);
   level.Render(0,ex,ez,p2.x,p2.z);
   if DrawShadow then begin
    glClear(GL_STENCIL_BUFFER_BIT);
    if DrawFloorShadow then begin
     for i:=0 to level.FY-1 do begin
      DrawShadows(i,level);
     end;
    end
    else begin
     for i:=1 to level.FY-1 do begin
      DrawShadows(i,level);
     end;
    end;
    DrawShadows(-2,level);
    glPushAttrib(GL_ALL_ATTRIB_BITS);
    glEnable(GL_STENCIL_TEST);
    glDisable(GL_LIGHTING);
    glColor4f(0,0,0,0.4 );
    glStencilFunc(GL_NOTEQUAL,0,$FFFFFFFF);
    glStencilOp(GL_KEEP,GL_KEEP,GL_KEEP);
    glPushMatrix;
    glLoadIdentity;
    glBegin(GL_TRIANGLE_STRIP);
     glVertex3f(-1, 1,-1);
     glVertex3f(-1,-1,-1);
     glVertex3f( 1, 1,-1);
     glVertex3f( 1,-1,-1);
     glEnd;
    glPopMatrix;
    glPopAttrib;
   end;
   renderPass:=0;
   level.Update(0);
   glDisable(GL_FOG);
  TOzoneMainClass(owner).particleSystem.render(delta);}
   glPopMatrix;
   FontSystem.DrawText( -1.2, 0.8, 'FPS: '+FloatToStrF(cFPS,ffGeneral,4,5), 0 , 0.75, 1, 1, 1);
   if level.paused then
    FontSystem.DrawText( -0.8, 0, 'PAUSED' , 1 , 3, 1, 1, 1);
   p2.Inventary.Render(delta);
  end;
//  TOzoneMainClass(owner).O3Log.LogStatus('4','---');
  if not NoSwap then SDL_GL_SwapBuffers;
  inc(s);
  if FTime>30 then begin
   cFPS:=s/(FTime/1000);
   s:=0;
   FTime:=0;
  end;

  delta:=SDL_GetTicks-timeStamp;
  inc(FTime,delta);
  glViewport( 0, 0, w, h);
  TOzoneMainClass(owner).vsplit:=1;
end;


end.
