unit o3physics;

interface

uses o3actors,dglOpenGL,Math,o3vectors,types,odeimport,o3utils,Classes,o3particles;

type
 TODEBodyData = record
  stabilize:boolean;
  fix:TVector3d;
  maxSpeedSquare:single;
  misc:array of pointer;
 end;
 PODEBodyData = ^TODEBodyData;
 TTerrainForceType = (collision,friction);
 TTerrainForces = array[collision..friction] of TVector3d;
 TForceFieldProc = function(dist:TVector3d;Data:Pointer;actor:TActor):single;
 TForceField = record
  getForce:TForceFieldProc;
  cx,cy,cz:single;
  Data:Pointer;
 end;
 PForceField = ^TForceField;
 TO3Physics = class
  public
   owner:pointer;
   theWorld:PdxWorld;
   theSpace:PdxSpace;
   ElectricCharged:TCallBackList;
   contactgroup: TdJointGroupID;
   ptime,wtime:integer;
   ForceFields:TList;
   noclip:boolean;
   constructor Create(own:pointer);
   destructor Destroy; override;
//   function getTerrainForces(level:Pointer;actor:TActor):TTerrainForces;
//   procedure applyPhysics(deltaTime:integer;actor:TActor;level:Pointer);
   procedure Step(deltaTime:integer);
 end;
 PO3Physics = TO3Physics;

const odeb_inactive = -$DEAD;
// procedure InitAcPhysics(var p:TActorPhysic);
var NumParticles:integer = 0;
    MatrIdent:TdMatrix3;
 function NewODEBodyData:PODEBodyData;

implementation
uses o3objects,o3stones,o3world,o3basicobj,o3item,o3game;

 function NewODEBodyData:PODEBodyData;
 begin
  New(result);
  with result^ do begin
   stabilize:=false;
   fix[0]:=odeb_inactive;
   fix[1]:=odeb_inactive;
   fix[2]:=odeb_inactive;
   maxSpeedSquare:=1E500;
  end;
 end;

 procedure LimitParticlesCallback(p:PParticle;t:integer);
 const ParticleLimit = 20;
 begin
  if p.data=nil then begin
   inc(NumParticles);
   p.data:=Pointer($FF);
  end;
  if NumParticles>ParticleLimit then
   p.life:=0;
  if p.life<=0 then
   dec(NumParticles);
 end;

// this is called by dSpaceCollide when two objects in space are
// potentially colliding.
 function BodyVel(t:PdxBody):TVector3d;
 begin
  if t=nil then
   Result:=getVector(0,0,0)
  else
   Result:=toVector3d(dBodyGetLinearVel(t));
 end;

 procedure nearCallback (data : pointer; o1, o2 : PdxGeom); cdecl;
 const
  cCOL_MAX = 100;
  cCOLLISION_SPEED = 0.2;
  sCOLLISION_SPEED = 0.2;
 var
  i : integer;
  b1, b2, actb : PdxBody;
  numc : integer;
  contact : array[0..cCOL_MAX-1] of TdContact;
  c : TdJointID;
  actor,obj:PdxGeom;
  isStoneContact,noContact:boolean;
  theStone:Tstone;
  theFirstActor:TActor;
  theSecondActor:TActor;
  sent:boolean;
  cForce:single;
  procedure prepare;
  begin
    if (TGameObject(dGeomGetData(o1)).objectType=otStone)or(TGameObject(dGeomGetData(o1)).objectType=otItem) then begin
     obj:=o1;
     actor:=o2;
     actb:=b2;
    end else begin
     obj:=o2;
     actor:=o1;
     actb:=b1;
    end;
    sent:=false;
    theFirstActor:=nil;
    theSecondActor:=nil;
    theStone:=nil;
    theFirstActor:= TActor(dGeomGetData(actor));
    isStoneContact:=false;
    noContact:=false;
    if (TGameObject(dGeomGetData(obj)).objectType=otStone)or(TGameObject(dGeomGetData(obj)).objectType=otItem)then begin
     if TGameObject(dGeomGetData(obj)).objectType=otStone then begin
      isStoneContact:=true;
      theStone:=TStone(dGeomGetData(obj));
     end
     else begin
      noContact:=true
     end;
    end
    else
      theSecondActor:=TActor(dGeomGetData(obj));
  end;

 begin
  if (TGameObject(o1.data).objectType<>otActor)and(TGameObject(o2.data).objectType<>otActor) then
   Exit;
  // exit without doing anything if the two bodies are connected by a joint
  b1 := dGeomGetBody(o1);
  b2 := dGeomGetBody(o2);
  if (assigned(b1) and assigned(b2) and (dAreConnected (b1,b2)<>0)) then
    exit;

  for i :=0 to cCOL_MAX-1 do
  begin
    contact[i].surface.mode := dContactBounce;
    // This determines friction, play around with it!
    contact[i].surface.mu:=100;
    contact[i].surface.mu2 := 0;
    contact[i].surface.bounce := 0.9;
    contact[i].surface.bounce_vel := 0.05;
  end;

  numc := dCollide (o1,o2,cCOL_MAX,contact[0].geom,sizeof(TdContact));
  if (numc>0) then
  begin
   for i := 0 to numc-1 do
   begin
    prepare;
//    contact[i].surface.bounce:=0;
    if noContact then
     Break;
//    if isStone and ((@theStone.isCollisionPossible=nil)or(not theStone.isCollisionPossible(0,theStone,theStone.x,theStone.y,theStone.z,theActor,theStone.Owner))) then
//     break;
//    if not isStone then
//     theItem.OnActorOver(@theActor,theActor.Owner);
//    with PLevel(theStone.Owner)^.Physics do
    if isStoneContact then begin
     contact[i].surface.mu:=TheStone.mu;
    // Stone-Actor Contact
     // X+ Collision
     contact[i].surface.bounce:=TheFirstActor.bounceity*TheStone.bounceity;
     if contact[i].geom.normal[0]>0 then begin
      if not((@theStone.collisionCallback<>nil)and(theStone.collisionCallback(1,theStone,theStone.x,theStone.y,theStone.z,theFirstActor,theStone.Owner,@contact[i].geom))) then
       break;
     end;
     // X- Collision
     if contact[i].geom.normal[0]<0 then begin
      if not((@theStone.collisionCallback<>nil)and(theStone.collisionCallback(2,theStone,theStone.x,theStone.y,theStone.z,theFirstActor,theStone.Owner,@contact[i].geom))) then
       break;
     end;
     // Y+ Collision
     if contact[i].geom.normal[1]>0 then begin
      if contact[i].surface.mu=0 then
       contact[i].surface.mu:=10;
      if not((@theStone.collisionCallback<>nil)and(theStone.collisionCallback(3,theStone,theStone.x,theStone.y,theStone.z,theFirstActor,theStone.Owner,@contact[i].geom))) then
       break;
      contact[i].surface.bounce:=0.2*(TheFirstActor.bounceity*TheStone.bounceity);
      theFirstActor.physics.floorSteerPower:=theStone.steer;
      theFirstActor.physics.jumping:=false;
      if theFirstActor.physics.godFor>0 then
       theFirstActor.physics.floorSteerPower:=1;
      dBodyAddForce(actb,
                    -(theStone.fric+theFirstActor.physics.selfFric) * theFirstActor.physics.mass * dBodyGetLinearVel(actb)[0]/numc,
                    -(theStone.fric+theFirstActor.physics.selfFric) * theFirstActor.physics.mass * dBodyGetLinearVel(actb)[1]/numc,
                    -(theStone.fric+theFirstActor.physics.selfFric) * theFirstActor.physics.mass * dBodyGetLinearVel(actb)[2]/numc);
     end;
     // Y- Collision
     if contact[i].geom.normal[1]<0 then begin
      if not((@theStone.collisionCallback<>nil)and(theStone.collisionCallback(4,theStone,theStone.x,theStone.y,theStone.z,theFirstActor,theStone.Owner,@contact[i].geom))) then
       break;
      contact[i].surface.bounce:=0.2*(TheFirstActor.bounceity*TheStone.bounceity);
     end;
     // Z+ Collision
     if contact[i].geom.normal[2]>0 then begin
      if not((@theStone.collisionCallback<>nil)and(theStone.collisionCallback(5,theStone,theStone.x,theStone.y,theStone.z,theFirstActor,theStone.Owner,@contact[i].geom))) then
       break;
     end;
     if contact[i].geom.normal[2]<0 then begin
     // Z- Collision
      if not((@theStone.collisionCallback<>nil)and(theStone.collisionCallback(6,theStone,theStone.x,theStone.y,theStone.z,theFirstActor,theStone.Owner,@contact[i].geom))) then
       break;
     end;
     cForce:=len(sum(BodyVel(b1),
             BodyVel(b2)),
             contact[i].geom.normal[0],
             contact[i].geom.normal[1],
             contact[i].geom.normal[2]);
     if
      (not sent) and
         (cForce>sCOLLISION_SPEED)
      then
      begin
       sent:=true; // play the sound only once!
       theFirstActor.playSound('collision',[@cForce,theStone]);
       theStone.playSound('collision',[@cForce,theFirstActor]);
      end;
    end
    else begin
     if contact[i].geom.normal[1]>0 then
      theFirstActor.physics.jumping:=false;
     if contact[i].geom.normal[1]<0 then
      theSecondActor.physics.jumping:=false;
     contact[i].surface.bounce:=(TheFirstActor.bounceity*TheSecondActor.bounceity);
     TheFirstActor.processMessage('collision',[theSecondActor,theFirstActor]);// Actor-Actor Collision
     TheSecondActor.processMessage('collision',[theFirstActor,theSecondActor]);// Actor-Actor Collision
     cForce:=len(sum(BodyVel(b1),
             BodyVel(b2)),
             contact[i].geom.normal[0],
             contact[i].geom.normal[1],
             contact[i].geom.normal[2]);
     if
      (not sent) and
          (cForce>sCOLLISION_SPEED) then

      begin
       sent:=true;
       TheFirstActor.playSound('accollision1',[@cForce,theSecondActor]);
       TheSecondActor.playSound('accollision2',[@cForce,theFirstActor]);
      end;
    end;
// spawn particles if speed greater than cCOLLISION_SPEED
    if
     len(sum(BodyVel(b1),
            BodyVel(b2)),
            contact[i].geom.normal[0],
            contact[i].geom.normal[1],
            contact[i].geom.normal[2])>cCOLLISION_SPEED then
       TOzoneMainClass(TLevel(TO3Physics(data).owner).owner).particleSystem.spawn(10 div numc,2,getVector(1,0.5,0),getVector(2,2,2),getVector(contact[i].geom.pos[0],contact[i].geom.pos[1],contact[i].geom.pos[2]),200,10,-1,LimitParticlesCallback,nil);
//    odeimport.
//    if contact[i].geom.depth>0 then
//     theParticleSystem.spawn(100,2,getVector(1,0,0),getVector(1,1,1),getVector(contact[i].geom.pos[0],contact[i].geom.pos[1],contact[i].geom.pos[2]),1000,100);
//if contact[i].geom.pos
    c := dJointCreateContact(TO3Physics(data).theWorld,TO3Physics(data).contactgroup,@contact[i]);
    dJointAttach (c,b1,b2);
//    odeimport.
   end;
  end;
 end;


  constructor TO3Physics.Create;
  begin
   owner:=own;
   noclip:=false;
   MatrIdent[0]:=1;
   MatrIdent[4]:=0;
   MatrIdent[8]:=0;
   MatrIdent[1]:=0;
   MatrIdent[5]:=1;
   MatrIdent[9]:=0;
   MatrIdent[2]:=0;
   MatrIdent[6]:=0;
   MatrIdent[10]:=1;
   theWorld:=dWorldCreate;
   ElectricCharged:=TCallbackList.Create;
   ForceFields:=TList.Create;
//   dWorldSetCFM(theworld,1e-5);
   theSpace:=odeimport.dHashSpaceCreate(nil);
   dWorldSetGravity (theWorld,0,-10.81,0);
   contactGroup:=dJointGroupCreate(0);
   ptime:=0;
   wtime:=0;
  end;

  destructor TO3Physics.Destroy;
  begin
   dWorldDestroy(theWorld);
   dSpaceDestroy(theSpace);
   dJointGroupDestroy (contactGroup);
   TOzoneMainClass(TLevel(owner).owner).O3Log.LogStatus('freeing lists','TO3Physics.Free');
   ElectricCharged.FList.Clear;
   ElectricCharged.Free;
   ForceFields.Free;
  end;

  procedure TO3Physics.Step(deltaTime:integer);
  const Pstep=10;
  var b:PdxBody;
      x:Pointer;
      t,s:PdVector3;
      w:single;
  begin
   inc(wtime,deltaTime);
   while ptime<wtime do begin
    if not noclip then begin
     dSpaceCollide(theSpace,self,@nearCallback);
     b:=theWorld.firstbody;
     while b<>nil do begin
      x:=dBodyGetData(b);
      if x<>nil then begin
       with PODEBodyData(x)^ do begin
        if stabilize then begin
         dBodySetAngularVel(b,0,0,0);
         dBodySetRotation(b,MatrIdent);
       end;
        t:=dBodyGetLinearVel(b);
 //       t[0]:=Min(maxSpeed[0],t[0]);
 //       t[1]:=Min(maxSpeed[1],t[1]);
 //       t[2]:=Min(maxSpeed[2],t[2]);
        w:=t[0]*t[0]+t[1]*t[1]+t[2]*t[2];
        if w>maxSpeedSquare then begin
         t[0]:=t[0]*sqrt(maxSpeedSquare)/sqrt(w);
         t[1]:=t[1]*sqrt(maxSpeedSquare)/sqrt(w);
         t[2]:=t[2]*sqrt(maxSpeedSquare)/sqrt(w);
        end;
        s:=dBodyGetPosition(b);
        if fix[0]<>odeb_inactive then begin
         s[0]:=fix[0];
         t[0]:=0;
        end;
        if fix[1]<>odeb_inactive then begin
         s[1]:=fix[1];
         t[1]:=0;
        end;
        if fix[2]<>odeb_inactive then begin
         s[2]:=fix[2];
         t[2]:=0;
        end;
        dBodySetPosition(b,s[0],s[1],s[2]);
        dBodySetLinearVel(b,t[0],t[1],t[2]);
       end;
      end;
      b:=PdxBody(b.BaseObject.next);
     end;
    end;
    {$IFDEF WIN32}
     odeimport.dWorldQuickStep(theWorld,Pstep/1000);
    {$ELSE}
     odeimport.dWorldStep(theWorld,Pstep/1000);
    {$ENDIF}
//    odeimport.dWorldQuickStep(theWorld,);
    dJointGroupEmpty(contactgroup);
    inc(ptime,Pstep);
   end;
  end;

(* procedure InitAcPhysics(var p:TActorPhysic);
 begin
  with p do begin
   noclip:=false;
   envsteerPower:=0;
   steerPower:=0.02;
   steer[0]:=0;
   steer[1]:=0;
   steer[2]:=0;

   mass:=5;
   gravity[0]:=0;
   gravity[1]:=-0.00981*mass;
   gravity[2]:=0;
   speed[0]:=0;
   speed[1]:=0;
   speed[2]:=0;
   CantDieFor:=0;
  end;
 end;

 function GreaterOrZero(z:integer):integer;
 begin
  if z<=0 then Result:=0
  else Result:=z;
 end;

(* procedure TDiPhysics.applyPhysics;
 var Fres:TVector3d;
     FTerrain:TTerrainForces;
 const maxSpeed = 0.3;
 begin
  if @actor.ControlProc<>nil then
   actor.ControlProc(actor);
  with actor.physics do begin
   steer:=mul(steer,envSteerPower*steerPower);
   if CantDieFor<>0 then
    FRes:=steer
    else Fres:=sum(steer,gravity);
   speed:=sum(speed,mul(mul(Fres,1/mass),deltaTime/1000));
   roundIt(speed);
   FTerrain:=getTerrainForces(level,actor);
   speed:=sum(speed,mul(mul(FTerrain[friction],1/mass),deltaTime/1000));
   if not noclip then
    speed:=sum(speed,FTerrain[collision]);

   if len(speed,1,1,1)>maxSpeed then
    speed:=mul(norm(speed),maxSpeed);

   actor.x:=actor.x+speed[0];
   actor.y:=actor.y+speed[1];
   actor.z:=actor.z+speed[2];
   if noclip then
    actor.alive:=true;
   if CantDieFor<>0 then
    actor.alive:=true;
   if CantDieFor>0 then
    cantDieFor:=greaterOrZero(cantDieFor-deltaTime);
  end;
 end;


// function s(d:double):integer;
//begin
// if d>0 then Result:=1
// else if d<-0 then Result:=-1
// else Result:=0;
//end;


function inv(x:single):single;
begin
 if x<>0 then result:=1/x
 else Result:=0;
end;


(*function TDiPhysics.getTerrainForces;
var EmptyCell:TCell;
//    lx,ly,lz:integer;
 function gFCells(x,y,z:single):TCell;
 begin
  Result:=PLevel(level)^.getCell(trunc(x),trunc(y),trunc(z));
  if Result=nil then Result:=EmptyCell;
//  if (x>=0)and(x<PLevel(level)^.FX)and(y>=0)and(y<PLevel(level)^.FY)and(z>=0)and(z<PLevel(level)^.FZ) then
//   Result:=PLevel(level)^.FCells[trunc(x),trunc(y),trunc(z)];
 end;
begin
 EmptyCell:=TCell.Create('nil','nil',@);
 Result[collision]:=getVector(0,0,0);
 Result[friction]:=getVector(0,0,0);
 actor.physics.envSteerPower:=0;
{ for lx:=-1 to 1 do
  for ly:=-1 to 1 do
   for lz:=-1 to 1 do
    with
     gFCells(actor.x+lx*actor.sphereRad,
            actor.y+ly*actor.sphereRad,
            actor.z+lz*actor.sphereRad)
            do begin
             if not((ly<>0)and(not((lx=0)and(lz=0)))) then
             if FStone<>nil then begin
              if len(actor.physics.speed,lx,ly,lz)>0 then
               if (@FStone.isCollisionPossible<>nil) and FStone.isCollisionPossible(abs(lx)+abs(ly)*10+abs(lz)*100,FStone) then
                Result[collision]:=sum(Result[collision],getVector(abs(lx),abs(ly),abs(lz)));
                if(lx=0)and(lz=0)and(ly=-1) then begin
                 actor.physics.envsteerPower:=FStone.steer;
                 Result[friction,0]:=-s(actor.physics.speed[0])*FStone.fric;
                 Result[friction,2]:=-s(actor.physics.speed[2])*FStone.fric;
                end;
             end;
            end;}
 with
    gFCells(actor.x,
           actor.y,
           actor.z)
            do begin
             if FItem<>nil then begin
              FItem.OnActorOver(@actor,@level);
             end;
             if FStone<>nil then begin
              if @Actor.OnStoneOver<>nil then
               if (@FStone.isCollisionPossible<>nil) and FStone.isCollisionPossible(-1,FStone,FStone.x,FStone.y,FStone.z,@actor,level) then
                Actor.OnStoneOver(FStone,@level,Actor);
             end;
            end;

 with
    gFCells(actor.x+actor.sphereRad,
           actor.y,
           actor.z)
            do begin
             if FStone<>nil then begin
              if actor.physics.speed[0]>0 then
               if (@FStone.isCollisionPossible<>nil) and FStone.isCollisionPossible(1,FStone,FStone.x,FStone.y,FStone.z,@actor,level) then begin
                Result[collision,0]:=actor.physics.speed[0]*-1.9;
                PLevel(level)^.UpdateAllOxydsOpen;
               end;
             end;
            end;
 with
    gFCells(actor.x-actor.sphereRad,
           actor.y,
           actor.z)
            do begin
             if FStone<>nil then begin
              if actor.physics.speed[0]<0 then
               if (@FStone.isCollisionPossible<>nil) and FStone.isCollisionPossible(2,FStone,FStone.x,FStone.y,FStone.z,@actor,level) then begin
                Result[collision,0]:=actor.physics.speed[0]*-1.9;
                PLevel(level)^.UpdateAllOxydsOpen;
               end;
             end;
            end;
 with
    gFCells(actor.x,
           actor.y+actor.sphereRad,
           actor.z)
            do begin
             if FStone<>nil then begin
              if actor.physics.speed[1]>0 then
               if (@FStone.isCollisionPossible<>nil) and FStone.isCollisionPossible(3,FStone,FStone.x,FStone.y,FStone.z,@actor,level) then begin
                Result[collision,1]:=actor.physics.speed[1]*-1.2;
                actor.physics.envsteerPower:=FStone.steer;
                Result[friction]:=mul(mul(mul(actor.physics.speed,FStone.fric),inv(len(actor.physics.speed,1,0,1))),-power(len(actor.physics.speed,1,0,1),0.8));
                PLevel(level)^.UpdateAllOxydsOpen;
//                Result[friction,0]:=-s(actor.physics.speed[0])*FStone.fric;
//                Result[friction,2]:=-s(actor.physics.speed[2])*FStone.fric;
               end;
             end;
            end;
 with
    gFCells(actor.x,
           actor.y-actor.sphereRad,
           actor.z)
            do begin
             if FStone<>nil then begin
              if actor.physics.speed[1]<0 then
               if (@FStone.isCollisionPossible<>nil) and FStone.isCollisionPossible(4,FStone,FStone.x,FStone.y,FStone.z,@actor,level) then begin
                Result[collision,1]:=actor.physics.speed[1]*-1.2;
                actor.physics.envsteerPower:=FStone.steer;
                Result[friction]:=mul(mul(mul(actor.physics.speed,FStone.fric),inv(len(actor.physics.speed,1,0,1))),-power(len(actor.physics.speed,1,0,1),0.8));
                PLevel(level)^.UpdateAllOxydsOpen;
//                Result[friction,0]:=-s(actor.physics.speed[0])*FStone.fric;
//                Result[friction,2]:=-s(actor.physics.speed[2])*FStone.fric;
               end;
             end;
            end;
 with
    gFCells(actor.x,
           actor.y,
           actor.z+actor.sphereRad)
            do begin
             if FStone<>nil then begin
              if actor.physics.speed[2]>0 then
               if (@FStone.isCollisionPossible<>nil) and FStone.isCollisionPossible(5,FStone,FStone.x,FStone.y,FStone.z,@actor,level) then begin
                Result[collision,2]:=actor.physics.speed[2]*-1.9;
                PLevel(level)^.UpdateAllOxydsOpen;
               end;
             end;
            end;
 with
    gFCells(actor.x,
           actor.y,
           actor.z-actor.sphereRad)
            do begin
             if FStone<>nil then begin
              if actor.physics.speed[2]<0 then
               if (@FStone.isCollisionPossible<>nil) and FStone.isCollisionPossible(6,FStone,FStone.x,FStone.y,FStone.z,@actor,level) then begin
                Result[collision,2]:=actor.physics.speed[2]*-1.9;
                PLevel(level)^.UpdateAllOxydsOpen;
               end;
             end;
            end;

// oneVec(Result[collision]);
// Result[collision]:=mul(mulVec(Result[collision],actor.physics.speed),-1.3);
 EmptyCell.Free;
end;*)
end.
