unit o3world;
interface
uses dglOpenGL,ac3d,o3m,o3utils,o3particles,SysUtils,Classes,Math,o3control,o3actors,o3item,o3stones,o3vectors,o3physics,o3levelinterface;
type
        TCell = class
         public
          FStone:TStone;
          FItem:TItem;
          owner:Pointer;
          procedure Render(delta:integer;grid:boolean=false);
          constructor Create(stone,item:string;own:Pointer);
          destructor Destroy; override;
        end;
        TStoneFunc = function(s:string;owner:Pointer):TStone;
        TItemFunc = function(s:string;owner:Pointer):TItem;
        TActorFunc = function(name:string;sx,sy,sz:single;own:Pointer;info:string):TActor;
        TGetSteer = function(c:string):TO3Control;
        TLevel = class;
        TActorList = class
         private
          FList:TList;
          level:TLevel;
          function getActor(s:string):TActor;
         public
          constructor Create(lev:TLevel);
          destructor Destroy; override;
          function AddActor(actor:TActor;name:string):Integer;
          function RemoveActor(actor:TActor):Integer;
          procedure RenderAll(deltaTime:integer;plane:integer=-1);
          procedure UpdateAll(deltaTime:integer);
          property Actor[s:string]:TActor read getActor;default;
          property List:TList read FList;
        end;
        TEnvBoxSide = (ebLeft,ebRight,ebFront,ebBack,ebTop,ebBottom);
        TLevel = class
         public
          envBoxTex: array[TEnvBoxSide] of Cardinal;
          envBoxTheme: string;
          owner:pointer;
          FCells:array of array of array of TCell;
          signals:TSignals;
          Physics:TO3Physics;
          CallBackBuffer:TCallBackList;
          FActors:TActorList;
          FX,FY,FZ:integer;
          paused:boolean;
          theActorToFollow:TActor;
	  theActorToFollow2:TActor;
          loadingLib:TLevelLib;
          MakeStone:TStoneFunc;
          MakeItem:TItemFunc;
          MakeActor:TActorFunc;
          state:(LevLost,LevWon,LevRestart,nochange);
          LostOnDeath:boolean;
          disposeList:TList;
          editorMode:boolean;
          function getParticleSystem:TParticleSystem;
          property particleSystem:TParticleSystem read getParticleSystem;
          procedure SendSignal(source:TGameObject;val:integer);
          function getCell(x,y,z:integer):TCell;
          constructor Create(own:pointer;loader:string;filename:string;args:string;st:TStoneFunc;it:TItemFunc;ac:TActorFunc);
          procedure saveToFile(filename,args:string);
          destructor Destroy; override;
          procedure Render(delta:integer;ex,ez,cx,cz:single;plane:integer=-1;grid:boolean=false);
          procedure Update(deltaTime:integer);
          function UpdateAllOxydsOpen:boolean;
          procedure Notify1STdestroyed(got:boolean);
        end;
        PLevel = ^TLevel;
const
        cellSize = 1;
implementation

uses o3objects,odeimport,IniFiles,o3game;

destructor TCell.Destroy;
begin
 if (FStone<>nil)and(not FStone.freed) then FStone.Free;
 if (FItem<>nil)and(not FItem.freed) then FItem.Free;
end;

function TLevel.getParticleSystem;
begin
 Result:=TOzoneMainClass(owner).particleSystem;
end;

procedure TLevel.SendSignal;
begin
 if signals<>nil then
  signals.SendSignal(source,val);
end;

procedure TLevel.Notify1STdestroyed;
begin
 TOzoneMainClass(owner).O3Log.LogStatus('notify','TLevel.Notify1STDestroyed');
 if not got then
  state:=LevLost;
 if LostOnDeath then
  state:=LevRestart;
 if state=levLost then
  TOzoneMainClass(owner).soundLib.playSound('looser',theActorToFollow.x,theActorToFollow.y,theActorToFollow.z);
end;

function TLevel.getCell;
begin
 Result:=nil;
 if FCells=nil then TOzoneMainClass(owner).O3Log.LogError('FCells is NIL, level error','TLevel.getCell');
 if (x<FX)and(x>=0)and(y<FY)and(y>=0)and(z<FZ)and(z>=0) then
  Result:=FCells[x,y,z];
end;

function TActorList.getActor;
var x:integer;
begin
 Result:=nil;
 if FList.Count>0 then begin
  for x:=0 to FList.Count-1 do begin
   if FList.Items[x]<>nil then begin
    if TActor(FList.Items[x]).name=s then
      Result:=TActor(FList.Items[x]);
   end;
  end;
 end;
end;

function TActorList.RemoveActor(actor:TActor):Integer;
begin
 Result:=FList.Remove(actor);
end;

function TActorList.AddActor;
begin
 actor.name:=name;
 Result:=FList.Add(actor);
end;

procedure TActorList.RenderAll;
var x:integer;
    m:PdMatrix3;
begin
 if FList.Count>0 then begin
  for x:=0 to FList.Count-1 do begin
   if FList.Items[x]<>nil then begin
    if FList.Items[x]<>nil then begin
      m:=dBodyGetRotation(TActor(FList.Items[x]).body);
//      gfx.FObject.Rotate:=sum(gfx.FObject.Rotate,mul(toVector3d(t),deltaTime/40));
       if (TActor(FList.Items[x]).gfxType=ac) then
        with TAC3dFile(TActor(FList.Items[x]).gfx).FObject^ do begin
         matr[0] := m[0];
         matr[1] := m[4];
         matr[2] := m[8];
         matr[3] := 0;
         matr[4] := m[1];
         matr[5] := m[5];
         matr[6] := m[9];
         matr[7] := 0;
         matr[8] := m[2];
         matr[9] := m[6];
         matr[10] := m[10];
         matr[11] := 0;
         matr[12] := 0;
         matr[13] := 0;
         matr[14] := 0;
         matr[15] := 1;
        end
       else if (TActor(FList.Items[x]).gfxType=ozone) then
        with TO3mFile(TActor(FList.Items[x]).gfx).FObject^ do begin
         matr[0] := m[0];
         matr[1] := m[4];
         matr[2] := m[8];
         matr[3] := 0;
         matr[4] := m[1];
         matr[5] := m[5];
         matr[6] := m[9];
         matr[7] := 0;
         matr[8] := m[2];
         matr[9] := m[6];
         matr[10] := m[10];
         matr[11] := 0;
         matr[12] := 0;
         matr[13] := 0;
         matr[14] := 0;
         matr[15] := 1;
        end;
     if plane<>-1 then begin
      if ((TActor(FList.Items[x]).y<0)and(plane=0))or((TActor(FList.Items[x]).y>=TLevel(TActor(FList.Items[x]).owner).FY)and(plane=-2))or(trunc(TActor(FList.Items[x]).y)=plane) then
       TActor(FList.Items[x]).Render(deltaTime);
     end
     else TActor(FList.Items[x]).Render(deltaTime);
    end;
   end;
  end;
 end;
end;

procedure TActorList.UpdateAll;
var x,i:integer;
    steer:TVector3d;
    actor:TActor;
    forceField:TForceField;
    chargedObject:TGameObject;
    l,l2:single;
    t:PdVector3;
begin
 if FList.Count>0 then begin
  for x:=0 to FList.Count-1 do begin
   actor:=TActor(FList.Items[x]);
   steer:=getVector(0,0,0);
   //actor.Look:=TLevel(actor.owner).theActorToFollow.Look;
   if (actor.ControlProc<>nil) then
    steer:=actor.ControlProc.Steer(actor,deltaTime);
   if actor.physics.jumping then
    actor.physics.floorSteerPower:=0.3;
   steer:=mul(steer,actor.physics.steerPower);
   actor.physics.jumping:=true;
   dBodySetForce(actor.body,0,0,0);
//   odeimport.dBodySetAngularVel(actor.body,steer[2],steer[1],steer[0]);
//   odeimport.dBodyAddTorque(actor.body,steer[0],steer[1],steer[2]);
   dBodyAddForce(actor.body,steer[0]*0.5*actor.physics.floorSteerPower,steer[1]*0.5*actor.physics.floorSteerPower,steer[2]*0.5*actor.physics.floorSteerPower);
//   odeimport.dBo
//   odeimport.dBodyAddRelForce((actor.body,steer[0],steer[1],steer[2]);
  odeimport.dBodySetTorque(actor.body,steer[2]*0.2*0.5,steer[1]*0.5,-steer[0]*0.2*0.5);
//      t:=dBodyGetAngularVel(actor.body);
//      if len(toVector3d(dBodyGetLinearVel(body)),1,1,1)<0.01 then
//       dBodySetAngularVel(actor.body,t[0]+steer[0],0,steer[2]);
//   odeimport.dBodySetForce(actor.body,0,0,0);
   if TLevel(actor.owner).Physics.ForceFields.Count>0 then
    for i:=0 to TLevel(actor.owner).Physics.ForceFields.Count-1 do begin
     forceField:=PForceField(TLevel(actor.owner).Physics.ForceFields[i])^;
     if @forceField.getForce=nil then Continue;
     steer:=getVector(forceField.cx-actor.x,forcefield.cy-actor.y,forcefield.cz-actor.z);
     steer:=norm(steer);
     steer:=mul(steer,
                forceField.getForce(
                GetVector(
                forceField.cx-actor.x,
                forcefield.cy-actor.y,
                forcefield.cz-actor.z),
                forcefield.Data,
                actor
                ));
//     odeimport.dBodyGetQuaternion()
//     dBodySetForce(actor.body,0,0,0);
     dBodyAddForce(actor.body,steer[0],steer[1],steer[2]);
    end;
   if actor.ECharge<>0 then
    if TLevel(actor.owner).Physics.ElectricCharged.FList.Count>0 then
     for i:=0 to TLevel(actor.owner).Physics.ElectricCharged.FList.Count-1 do begin
      chargedObject:=TLevel(actor.owner).Physics.ElectricCharged.FList[i];
      if chargedObject=nil then continue;
      if chargedObject=actor then continue;
      steer:=getVector(chargedObject.cx-actor.x,chargedObject.cy-actor.y,chargedObject.cz-actor.z);
      l:=len(steer,1,1,1);
      steer:=norm(steer);
      steer:=mul(steer,
                 ((-actor.ECharge/100)*
                 (chargedObject.ECharge/100))/
                 (l));
      dBodyAddForce(actor.body,steer[0],steer[1],steer[2]);
     end;

  end;
  Level.physics.step(deltaTime);
  for x:=0 to FList.Count-1 do begin
   if FList.Items[x]<>nil then begin
    if FList.Items[x]<>nil then
     with TActor(FList.Items[x]) do begin
      t:=dBodyGetPosition(body);
      cx:=t[0];
      cy:=t[1];
      cz:=t[2];
      t:=dBodyGetAngularVel(body);
//      if len(toVector3d(dBodyGetLinearVel(body)),1,1,1)<0.01 then
      l2:=t[0]*t[0]+t[2]*t[2];
      if l2>625 then begin
       t[0]:=(t[0]/sqrt(l2))*25;
       t[2]:=(t[2]/sqrt(l2))*25;
      end;
      dBodySetAngularVel(body,t[0],0,t[2]);
     end;
   end;
  end;
 end;
end;

constructor TActorList.Create;
begin
 inherited Create;
 level:=lev;
 FList := TList.Create;
end;

destructor TActorList.Destroy;
var x:integer;
begin
 TOzoneMainClass(level.owner).O3Log.LogStatus('freeing TActorList, count='+Inttostr(Flist.count),'TActorList.Free');
 if FList.Count>0 then begin
  for x:=0 to FList.Count-1 do begin
   if (FList.Items[x]<>nil)and(not TActor(FList[x]).freed) then begin
    TOzoneMainClass(level.owner).O3Log.LogStatus('calling actor.free','TActorList.Free');
    TActor(FList.Items[x]).Free;
   end;
  end;
 end;
 FList.Clear;
 FList.Free;
 TOzoneMainClass(level.owner).O3Log.LogStatus('inherited!','TActorList.Free');
 inherited Destroy;
 TOzoneMainClass(level.owner).O3Log.LogStatus('freed TActorList','TActorList.Free');
end;

constructor TCell.Create;
begin
 Owner:=own;
 if stone <> 'nil' then FStone := CreateStone(stone,owner);
 if item  <> 'nil' then FItem  := CreateItem(item,owner);
end;

procedure TCell.Render;
begin
 if FItem<>nil then begin
  FItem.Render(delta);
 end;
 if FStone<>nil then begin
  FStone.Render(delta);
 end;
 if grid and (FItem=nil) and (FStone=nil) then
  TOzoneMainClass(TLevel(owner).owner).gfxSys.paintGrid;
end;

function TLevel.UpdateAllOxydsOpen:boolean;
var x,y,z:integer;
begin
 Result:=true;
 if (FX>0)and(FY>0)and(FZ>0) then
  for x:=0 to FX-1 do
   for y:=0 to FY-1 do
    for z:=0 to FZ-1 do
     if (FCells[x,y,z]<>nil)and(FCells[x,y,z].FStone<>nil) then begin
      with FCells[x,y,z].FStone do begin
       if Info['oxyd']='yes' then
        Result:=Result and (Info['open']='yes');
       if not Result then Exit;
      end;
     end;
 if Result then begin
  state:=LevWon;
 end;
end;


procedure TLevel.Render;
var x,y,z,xb,zb:integer;
begin
 if paused then
  delta:=0;
 if ex<-1E100 then begin
  if (FX>0)and(FY>0)and(FZ>0) then begin
   for y:=0 to FY-1 do begin
    if plane<>-1 then
     if y<>plane then continue;
    for x:=0 to FX-1 do
     for z:=0 to FZ-1 do
      if getCell(x,y,z)<>nil then begin
       TOzoneMainClass(owner).gfxSys.translate(x,y,z);
       FCells[x,y,z].Render(delta,grid);
       TOzoneMainClass(owner).gfxSys.translate(-x,-y,-z);
      end;
   end;
  end;
 end
 else begin
 if (FX>0)and(FY>0)and(FZ>0) then begin
  xb:=floor(ex);
  zb:=floor(ez);
  xb:=max(xb,0);
  xb:=min(xb,FX-1);
  zb:=max(zb,0);
  zb:=min(zb,FZ-1);
  for y:=0 to FY-1 do begin
   if plane<>-1 then
    if y<>plane then continue;
   if xb>0 then
    for x:=0 to xb-1 do begin
     if zb>0 then
      for z:=0 to zb-1 do
       if getCell(x,y,z)<>nil then begin
        TOzoneMainClass(owner).gfxSys.translate(x,y,z);
        FCells[x,y,z].Render(delta,grid);
        TOzoneMainClass(owner).gfxSys.translate(-x,-y,-z);
       end;
     if zb<FZ then
      for z:=FZ-1 downto zb do
       if getCell(x,y,z)<>nil then begin
        TOzoneMainClass(owner).gfxSys.translate(x,y,z);
        FCells[x,y,z].Render(delta,grid);
        TOzoneMainClass(owner).gfxSys.translate(-x,-y,-z);
       end;
    end;
   if xb<FX then
    for x:=FX-1 downto xb do begin
     if zb>0 then
      for z:=0 to zb-1 do
       if getCell(x,y,z)<>nil then begin
        TOzoneMainClass(owner).gfxSys.translate(x,y,z);
        FCells[x,y,z].Render(delta,grid);
        TOzoneMainClass(owner).gfxSys.translate(-x,-y,-z);
       end;
     if zb<FZ then
      for z:=FZ-1 downto zb do
       if getCell(x,y,z)<>nil then begin
        TOzoneMainClass(owner).gfxSys.translate(x,y,z);
        FCells[x,y,z].Render(delta,grid);
        TOzoneMainClass(owner).gfxSys.translate(-x,-y,-z);
       end;
    end;
   end;
  end;
 end;
end;

function sideToString(t:TEnvBoxSide):string;
begin
 Result:='';
 case t of
  ebLeft: Result:='left';
  ebRight: Result:='right';
  ebTop: Result:='top';
  ebBottom: Result:='bottom';
  ebFront: Result:='front';
  ebBack: Result:='back';
 end;
end;

procedure TLevel.saveToFile;
begin
 if @loadingLib.SaveLevel<>nil then
  loadingLib.SaveLevel(filename+','+args,self);
end;

constructor TLevel.Create;
var t:TEnvBoxSide;
begin
 inherited Create;
 editorMode:=false;
 owner:=own;
 envBoxTheme:='normal';
 signals:=TSignals.Create;
 SendSignal(nil,0);
 state:=nochange;
 paused:=false;
 Physics:=TO3Physics.create(self);
 disposeList:=TList.Create;
 CallBackBuffer:=TCallBackList.Create;
 FActors:=TActorList.Create(self);
 theActorToFollow:=nil;
 theActorToFollow2:=nil;
 MakeStone:=st;
 MakeItem:=it;
 MakeActor:=ac;
 if loader<>'nil' then begin
  loadingLib:=TLevelLib.Create(loader);
  if @loadingLib.LoadLevel<>nil then
   loadingLib.LoadLevel(filename+','+args,self);
 end;
 for t:=ebLeft to ebBottom do begin
  TOzoneMainClass(owner).gfxSys.LoadTexture(envBoxTex[t],'sky_'+envBoxTheme+'_'+sideToString(t)+'.bmp',BitmapSubDir);
 end;
end;

destructor TLevel.Destroy;
var x,y,z:integer;
    t:TEnvBoxSide;
begin
 for t:=ebLeft to ebBottom do begin
  if envBoxTex[t]<>0 then TOzoneMainClass(owner).gfxSys.FreeTexture(envBoxTex[t]);
 end;
 TOzoneMainClass(owner).O3Log.LogStatus('freeing level','TLevel.Destroy');
 if (FX>0)and(FY>0)and(FZ>0) then
  for x:=0 to FX-1 do
   for y:=0 to FY-1 do
    for z:=0 to FZ-1 do
     if FCells[x,y,z]<>nil then
      FCells[x,y,z].Free;
 FActors.Free;
 TOzoneMainClass(owner).O3Log.LogStatus('freeing physics','TLevel.Destroy');
 Physics.Free;
 TOzoneMainClass(owner).O3Log.LogStatus('freeing signals','TLevel.Destroy');
 Signals.Free;
 TOzoneMainClass(owner).O3Log.LogStatus('freeing callback-buffer','TLevel.Destroy');
 CallBackBuffer.Free;
 if loadingLib<>nil then begin
  TOzoneMainClass(owner).O3Log.LogStatus('unloading level','TLevel.Destroy');
  if @loadingLib.UnLoadLevel<>nil then
   loadingLib.UnLoadLevel;
 end;
 if DisposeList.Count>0 then
  for x:=0 to DisposeList.Count-1 do begin
   Dispose(DisposeList[x]);
//   DisposeList.Delete(x);
  end;
 DisposeList.Free;
 TOzoneMainClass(owner).O3Log.LogStatus('free level successful','TLevel.Destroy');
 inherited;
end;

procedure TLevel.Update;
begin
 if not Paused then begin
  if state=LevWon then
   Physics.noclip:=true;
  CallBackBuffer.CallBackAll(deltaTime);
  if deltaTime<>0 then 
   FActors.UpdateAll(deltaTime);
 end;
end;




end.
